/***************************************************************************
 *   Copyright (C) 2005 by Vukasin Toroman                                 *
 *   vtoroman@vtoroman05                                                   *
 *                                                                         *
 *  For more licensing information consult COPYING                         *
 ***************************************************************************/
#ifndef WOOLQUEUE_H
#define WOOLQUEUE_H
#include <list>
#include <iostream>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>
#include "../base/exception.h"
#include "../base/debug.h"

namespace wool {

  /**
     @author Vukasin Toroman
  */

  template < typename unitT >
  class Queue {
    public:
    typedef Queue<unitT> QueueType;
    typedef unitT Unit;
    typedef typename boost::shared_ptr<Unit> UnitPtr;
    typedef typename std::list< UnitPtr > UnitPtrList;
    
    private:
    boost::mutex m_guard;
    boost::condition m_unitAvailable;
    boost::condition m_threadFinished;
    boost::condition m_unitFinished;

    UnitPtrList m_workLoad;
  
    size_t m_workerCount;
    bool m_shutdown;
    size_t m_minWorkerCount;
    size_t m_maxWorkerCount;
    size_t m_maxQueueSize;
    size_t m_timeout;

    public:
    
    virtual void handle( UnitPtr unit ) = 0;
    virtual void onError( const std::exception & e, UnitPtr unit ) = 0;
    virtual inline bool canStartWorker() { return true; };

    const size_t &  getMinWorkerCount() const {    return m_minWorkerCount;  }
    const size_t &  getMaxWorkerCount() const {    return m_maxWorkerCount;  }  
    const size_t &  getMaxQueueSize() const {    return m_maxQueueSize;  }  
    const size_t &  getTimeout() const {    return m_timeout;  }  
    void setMinWorkerCount( const size_t & s ) {    m_minWorkerCount = s;  }
    void setMaxWorkerCount( const size_t & s ) {    m_maxWorkerCount = s;  }  
    void setMaxQueueSize( const size_t & s ) {    m_maxQueueSize = s;  }  
    void setTimeout( const size_t & s ) {    m_timeout = std::min<size_t>(1000,s);  }  // minimal timeout is 1s to prevent busy looping
  

    Queue( ) 
      :m_workerCount(0)
      ,m_shutdown(false)
      ,m_minWorkerCount(0)
      ,m_maxWorkerCount(10)
      ,m_maxQueueSize(1)
      ,m_timeout(1000)
      {}
  
    void shutdown () {
      boost::mutex::scoped_lock l(m_guard);
      m_shutdown = true;
      m_unitAvailable.notify_all();
      m_unitFinished.notify_all();
      while(m_workerCount>0) {
	m_threadFinished.wait(l);
      }
    }
    
    virtual ~Queue() {
      shutdown();
    }

    void add(UnitPtr p) {
      boost::mutex::scoped_lock l(m_guard);
      if(m_shutdown)
	throw wool::exception("Queue is shutting down");
      if( (m_workerCount == 0) ||
	  (getMinWorkerCount()>m_workerCount) ||
	  (
	   (m_workerCount<=m_workLoad.size()) &&
	   (m_workerCount<getMaxWorkerCount()) &&
	   (canStartWorker())
	   )
	  ) {
	startWorker();
      }
        
      while(getMaxQueueSize()<m_workLoad.size()) {
	boost::xtime t;
	xtime_get(&t, boost::TIME_UTC);
	t.nsec += 500000000;
	
	m_unitFinished.timed_wait(l, t);
      }
    
      if(m_workLoad.empty())
	m_unitAvailable.notify_all();
      m_workLoad.push_back(p);
    }

    void wait( const size_t & limit = 0 ) {
      boost::mutex::scoped_lock l(m_guard);
      while(limit<m_workLoad.size()) {
        boost::xtime t;
        xtime_get(&t, boost::TIME_UTC);
        t.nsec += 100000000;
        m_unitFinished.timed_wait(l, t);
      }
    }

    void add(Unit * p) {
      UnitPtr ps(p);
      add(ps);    	
    }

    protected:
    void startWorker() {
      m_workerCount++;
      delete new boost::thread(boost::bind(&QueueType::run,this));
      WOOL_DEBUG("Worker count:" << m_workerCount);
    }

    void run() {
      while (true) {
	UnitPtr currentTask;
	{
	  boost::mutex::scoped_lock l(m_guard);
	  while(m_workLoad.empty()) {
	    boost::xtime t;
	    xtime_get(&t, boost::TIME_UTC);
	    t.sec += getTimeout();
	    if( (m_shutdown) || 
		(
		 (!m_unitAvailable.timed_wait(l,t)) 
		 && (m_workerCount>getMinWorkerCount())
		 )
		) {
	      m_workerCount--;
	      std::cout << "worker done: Worker count:" << m_workerCount<<std::endl;
	      m_threadFinished.notify_one();
	      return;
	    }
	  }
                
	  currentTask = m_workLoad.front();
	  m_workLoad.pop_front();
	}
	
	try {
	  handle(currentTask);
	} catch ( const std::exception & e) {
	  boost::mutex::scoped_lock l2(m_guard);
	  m_workerCount--;
	  std::cout << "worker error: '" << e.what() <<"' Worker count:" << m_workerCount<<std::endl;
	  m_threadFinished.notify_one();
	  onError(e,currentTask);
	  return;
	} catch ( ... ) {
	  boost::mutex::scoped_lock l2(m_guard);
	  m_workerCount--;
	  std::cout << "worker error: 'unknown' Worker count:" << m_workerCount<<std::endl;
	  m_threadFinished.notify_one();
	  onError(wool::exception("unknown error"), currentTask);
	  return;
	}
	m_unitFinished.notify_all();
      }
    }
  };
};

#endif
