/**************************************************************************
*   Copyright (C) 2004 by Vukasin Toroman                                 *
*   vukasin@toroman.org                                                   *
**************************************************************************/
/************************************************************************** 
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifndef WOOL_UNSORTED_MULTIMAP_H
#define WOOL_UNSORTED_MULTIMAP_H


#include "../base/cvs_info.h"
#include <string>
#include <map>
#include <list> 

#ifndef GCC_VERSION
#define CREATE_VERSION(major,minor,patch) (major * 1000000 + minor * 1000 + patch)
#define GCC_VERSION CREATE_VERSION(__GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__)
#endif


namespace wool
{
/**
 *  Map interface which <b>not</b> sort the objects.
 */
template <typename Key,typename Data,typename Compare>
class unsorted_multimap
{

public:
    /**
     *  cvs info
     */
    VERSION_INFO_INLINE("$Id: unsorted_multimap.h,v 1.4 2007/01/11 15:18:24 vtoroman Exp $");
    /**
     *  value definition
     */
    typedef typename std::pair<Key,Data> value_type ;
    /**
     *  reference definition
     */
    typedef value_type & reference ;
    /**
     *  pointer definition
     */
    typedef value_type * pointer ;

    /**
     * const value definition
     */
    typedef const value_type const_value_type ;
    /**
     *  const reference definition
     */
    typedef const_value_type & const_reference ;
    /**
     *  const pointer definition
     */
    typedef const_value_type * const_pointer ;
    /**
     * list definition
     */
    typedef typename std::list<value_type> list_type ;
    /**
     * key definition
     */
    typedef Key key_type ;

    /**
     *  iterator definition
     */
    typedef typename list_type::iterator iterator;
    /**
     * const iterator definition
     */
    typedef typename list_type::const_iterator const_iterator;
    /**
     *  reverse iterator definition
     */
    typedef typename list_type::reverse_iterator reverse_iterator;
    /**
     *  const reverse iterator definition
     */
    typedef typename list_type::const_reverse_iterator const_reverse_iterator;
    /**
     *  size_type definition
     */
    typedef typename list_type::size_type size_type;
    /**
     *  difference_type definition
     */
    typedef typename list_type::difference_type difference_type;
    /**
     *  compare definition
     */
    typedef Compare key_compare;
    /**
     *  struct to compare keys
     */
    struct eq_key_value
    {
        /**
         *  reference to the key
         */
        const key_type & key;
        /**
         *  reference to the compare function
         */
        key_compare compare_function;
        /**
         *  @param __key  key
         */
        eq_key_value(const key_type & __key):key(__key){ }
        /**
         *  () operator
         *  @param subject value
         *  @return true if
         */
        bool operator () (const_value_type & subject) const
        {return ((!compare_function(subject.first,key))&&(!compare_function(key,subject.first)));}
    };
    /**
     *  Returns a const iterator that points to the first element.
     *  @return const iterator that points to the first element
     */
    inline const_iterator begin() const
    {
        return m_value.begin();
    };
    /**
     *  Returns a const iterator that points one past the last element.
     *  @return const iterator that points one past the last element
     */
    inline const_iterator end() const
    {
        return m_value.end();
    };
    /**
     *  Returns a const reverse iterator that points to the last element.
     *  @return const reverse iterator that points to the last element
     */
    inline const_reverse_iterator rbegin() const
    {
        return m_value.rbegin();
    };
    /**
     *  Returns a const reverse iterator that points to one before the first element.
     *  @return const reverse iterator that points to one before the first element
     */
    inline const_reverse_iterator rend() const
    {
        return m_value.rend();
    };
    /**
     *  Returns an iterator that points to the first element.
     *  @return iterator that points to the first element
     */
    inline iterator begin()
    {
        return m_value.begin();
    };
    /**
     *  Returns an iterator that points one past the last element.
     *  @return iterator that points one past the last element
     */
    inline iterator end()
    {
        return m_value.end();
    };
    /**
     *  Returns a reverse iterator that points to the last element.
     *  @return reverse iterator that points to the last element
     */
    inline reverse_iterator rbegin()
    {
        return m_value.rbegin();
    };
    /**
     *  Returns a reverse iterator that points to one before the first element.
     *  @return reverse iterator that points to one before the first element
     */
    inline reverse_iterator rend()
    {
        return m_value.rend();
    };
    /**
     *  Returns the number of elements.
     *  @return number of elements
     */
    inline size_type size() const
    {
        return  m_value.size();
    };
    /**
     *  Returns the size() of the largest possible unsorted_multimap.
     *  @return max size
     */
    inline size_type max_size() const
    {
        return  m_value.max_size();
    };
    /**
     *  Returns true if the unsorted_multimap is empty.
     *  @return true if the unsorted_multimap is empty
     */
    inline bool empty() const
    {
        return m_value.empty();
    };
    /**
     *  Returns the compare method for keys.
     *  @return key_compare
     */
    inline key_compare key_comp() const
    {
        return key_compare();
    }
    /**
     *  Insert an object into the unsorted_multimap.
     *  @param pos An iterator that serves as a hint as to where the element should be inserted.
     *  @param x element to be inserted
     *  @return An iterator that points to the element.
     */
    inline iterator insert(iterator pos, const value_type& x)
    {
        return m_value.insert(pos,x);
    }
    
    /**
     *  Insert an object into the unsorted_multimap.
     *  @param pos An iterator that serves as a hint as to where the element should be inserted.
     *  @param x key to be inserted
     *  @param y value to be inserted
     *  @return An iterator that points to the element.
     */
    inline iterator insert(iterator pos, const key_type& x, const Data & y )
    {
        return m_value.insert(pos,value_type(x,y));
    }
    
    /**
     *  Insert an object into the unsorted_multimap (at the end).
     *  @param x key to be inserted
     *  @param y value to be inserted
     *  @return An iterator that points to the element.
     */
    inline iterator insert(const key_type& x, const Data & y )
    {
        return m_value.insert(end(),value_type(x,y));
    }
    /**
     *  Insert an object at the end of the unsorted_multimap.
     *  @param x element
     *  @return An iterator that points to the element.
     */
    inline iterator insert(const value_type& x)
    {
        return m_value.insert(m_value.end(),x);
    }

    inline void push_back(const value_type& x)
    {
        m_value.push_back(x);
    }

    inline void push_front(const value_type& x)
    {
        m_value.push_front(x);
    }

    inline void pop_back()
    {
        m_value.pop_back();
    }

    inline void pop_front()
    {
        m_value.pop_front();
    }

    /**
     *  A template function that attemps to insert a range of elements.
     *  @param pos An iterator that serves as a hint as to where the range should be inserted.
     *  @param __begin Iterator pointing to the start of the range to be inserted.
     *  @param __end Iterator pointing to the end of the range to be inserted.
     */
    template <class InputIterator>
    void insert( iterator pos, InputIterator __begin, InputIterator __end)
    {
        m_value.insert(pos,__begin,__end);
    }

    /**
     *  Remove element at given position.
     *  @param pos position of the element which should be erased
     *  @return An iterator pointing to the next element or end().
     */
    inline iterator erase(iterator pos)
    {
        return m_value.erase(pos);
    }

    /**
     *  Remove element with this key.
     *  @param pos key
     *  @return number of erased elements
     */
    inline size_type erase(const key_type & pos)
    {
        size_type count = 0;
        count = this->count(pos);
        m_value.remove_if(eq_key_value(pos));
        return count;
    }
    /**
     *  Remove a range of elements.
     *  @param first Iterator pointing to the first element to be erased.
     *  @param last Iterator pointing to one past the last element to be erased.
     *  @return An iterator pointing to the element pointed to by last prior to erasing (or end()).
     */
    inline iterator erase(iterator first, iterator last)
    {
         return m_value.erase(first, last);
    }
    /**
     *  Erases all the elements.
     */
    inline void clear()
    {
        m_value.clear();
    }
    /**
     *  Tries to locate the first element with this key.
     *  @param k key
     *  @return Iterator pointing to sought-after element, or end() if not found.
     */
    inline iterator find(const key_type& k )
    {
      return find(k,begin());
    }
    /**
     *  Tries to locate the first element with this key.
     *  @param k key
     *  @return Cosnt iterator pointing to sought-after element, or end() if not found.
     */
    inline const_iterator find(const key_type& k ) const
    {
      return find(k,begin());
    }

    /**
     *  Tries to locate the first element with this key after the current iterator position.
     *  @param k key
     *  @param current iterator to begin the search
     *  @return Iterator pointing to sought-after element, or end() if not found.
     */
    inline iterator find(const key_type& k , iterator current)
    {
        iterator last = m_value.end();
        eq_key_value compare_function(k);
        for(;current!=last;current++)
            if(compare_function(*current))
                return current;
        return current;
    };
    /**
     *  Tries to locate the first element with this key after the current iterator position.
     *  @param k key
     *  @param current const iterator to begin the search
     *  @return Const iterator pointing to sought-after element, or end() if not found.
     */
    inline const_iterator find(const key_type& k, const_iterator current) const
    {
        const_iterator last = m_value.end();
        eq_key_value compare_function(k);
        for(;current!=last;current++)
            if(compare_function(*current))
                return current;
        return current;
    };


    /**
     *  Counting the elements with this key.
     *  @param k key
     *  @return number of elements with this key
     */
    inline size_type count(const key_type& k) const
    {
        size_type result=0;
        const_iterator current = m_value.begin();
        const_iterator last = m_value.end();
        eq_key_value compare_function(k);
        for(;current!=last;current++)
            if(compare_function(*current))
                result++;
        return result;
    }

    /**
     *  Equal operator
     *  @param other_map map to compare to
     *  @return true if unsorted_multimaps are equal else false
     */
    inline bool operator == ( const unsorted_multimap<Key,Data,Compare> & other_map) const
    {
        return m_value== other_map.m_value;
    }
    /**
     *  Lower operator
     *  @param other_map map to compare to
     *  @return true if other_map is greater else false
     */
    inline bool operator < ( const unsorted_multimap<Key,Data,Compare> & other_map) const
    {
        return m_value<other_map.m_value;
    }

    inline unsorted_multimap( )
    { }
    inline ~unsorted_multimap( )
    { }

    /**
     *  Copy constructor
     *  @param other_map map to copy to
     */
    inline unsorted_multimap ( const unsorted_multimap<Key,Data,Compare> & other_map )
    {
      m_value = other_map.m_value;
    }

    /**
     *  Swap values 
     *  @param other_map map to swap with
     */
    inline void swap ( unsorted_multimap<Key,Data,Compare> & other_map )
    {
      m_value.swap(other_map.m_value);
    }


    /**
     *  Constructor which inserts a range of elements into the unsorted_multimap.
     *  @param __begin Iterator pointing to the start of the range to be inserted.
     *  @param __end Iterator pointing to the end of the range to be inserted.
     */
    template <typename InputIterator>
    unsorted_multimap( InputIterator __begin, InputIterator __end)
    {
        insert(end(),__begin,__end);
    }
     /**
     *  Constructor which copies the value from the given map
     *  @param other_map map to copy the data from into the new unsorted_multimap
     */
    template <typename Key2,typename Data2, typename Compare2>
    unsorted_multimap( const unsorted_multimap<Key2,Data2,Compare2> & other_map)
    {
        insert(end(),other_map.begin(),other_map.end());
    }
    /**
     *  Assignment operator
     *  @param other_map original map
     *  @return unsorted_multimap
     */
    template <typename Key2,typename Data2, typename Compare2>
    unsorted_multimap<Key,Data,Compare> & operator = ( const unsorted_multimap<Key2,Data2,Compare2> & other_map)
    {
        clear();
        insert(end(),other_map.begin(),other_map.end());
        return (*this);
    }
private:
    list_type m_value;
};

};

#endif
