//
// C++ Implementation: config
//
// Description: 
//
//
// Author: Vukasin Toroman <vtoroman@astaro.de>, (C) 2006
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "config/config.h"
#include "base/exception.h"
#include "rfc3986/url.h"

#include <boost/thread.hpp>
#include <stdexcept>

#include <map>
#include <xmlwrapp/init.h>
#include <sys/stat.h>
namespace wool {

  Configuration::Configuration()
    :m_documentParser(NULL){
  }

  Configuration::Configuration(const Configuration & c)
    :m_config_path(c.m_config_path),
     m_documentParser(NULL){
  }

  void Configuration::load(const std::string & config_path) {
    static boost::mutex init_guard;
    static bool inited = false;
    static xml::init i;
    boost::mutex::scoped_lock l(init_guard);
    if(!inited) {
      i.indent_output(false);
      i.remove_whitespace(false);
      i.substitute_entities(true);
      inited = true;
    }
    l.unlock();
    boost::mutex::scoped_lock lm(m_guard);
    if(config_path == m_config_path) {
      struct stat file_stat;
      if ( stat(config_path.c_str(), &file_stat) != 0 ) 
	throw wool::exception ("cannot stat config file: " + config_path );
      if( file_stat.st_mtime <= m_create_time )
	return;
      std::cout << "File '"<<config_path<<"' needs reload" << std::endl;
    }
    WOOL_DEBUG("load " << config_path);
    delete m_documentParser;
    m_documentParser = NULL;
    m_config_path = config_path;
    m_documentParser = new xml::tree_parser(m_config_path.c_str(),true);
    time(&m_create_time);
  }

  Configuration::~Configuration(){
    delete m_documentParser;
  }

  const Configuration & Configuration::GetInstance( const std::string & s) {    
    static boost::mutex guard;
    static std::map<std::string, Configuration> instances;

    boost::mutex::scoped_lock l(guard);
    Configuration & r = instances[s];
    r.load(s);
    return r;
  }

  void Configuration::extract( const std::string & path, std::string & s ) const {
    boost::mutex::scoped_lock lm(m_guard);
    WOOL_DEBUG("extract " << path );
    wool::url u(path);
    wool::url::path_type & p = u.path;
    s.clear();
    if(m_documentParser) {
    
      if( (p.empty()) || (m_documentParser->get_document().get_root_node().get_name() != p.front()))
	return;
      wool::url::path_type::iterator i = p.begin();
      i++;
      xml::node::const_iterator curr = m_documentParser->get_document().get_root_node().self();
      for(;i!=p.end();i++) {
	if(!i->empty()) {
	  int e_name = i->find('%');
	  std::string name (*i,0,e_name);
	  int number = 0;

	  if(e_name!=std::string::npos) {
	    std::istringstream s( std::string (*i,e_name + 1,i->length()) );
	    s >> number;
	  }

	  if(curr->get_type()!=xml::node::type_element) {
	    WOOL_DEBUG("found a non element node "<< *i);
	    return;
	  }
	  xml::node::const_iterator next = curr->begin();
	  for( int j =0; j<=number; j++ ) {
	    if(j!=0)
	      next++;
	    next = curr->find(name.c_str(), next);
	    if(next==curr->end()) {
	      WOOL_DEBUG("did not find node " << name );
	      return;
	    }
	  }
	  curr = next;
	}
      }
      s = curr->get_content();
    }
  }
  
  void Configuration::list( const std::string & path, std::list<std::string> & s ) const {
    boost::mutex::scoped_lock lm(m_guard);
    WOOL_DEBUG("extract " << path );
    wool::url u(path);
    wool::url::path_type & p = u.path;
    s.clear();
    if(m_documentParser) {
    
      if( (p.empty()) || (m_documentParser->get_document().get_root_node().get_name() != p.front()))
	return;
      wool::url::path_type::iterator i = p.begin();
      i++;
      xml::node::const_iterator curr = m_documentParser->get_document().get_root_node().self();
      for(;i!=p.end();i++) {
	if(!i->empty()) {
	  int e_name = i->find('%');
	  std::string name (*i,0,e_name);
	  int number = 0;

	  if(e_name!=std::string::npos) {
	    std::istringstream ss( std::string (*i,e_name + 1,i->length()) );
	    ss >> number;
	  }

	  if(curr->get_type()!=xml::node::type_element) {
	    WOOL_DEBUG("found a non element node "<< *i);
	    return;
	  }
	  xml::node::const_iterator next = curr->begin();
	  for( int j =0; j<=number; j++ ) {
	    if(j!=0)
	      next++;
	    next = curr->find(name.c_str(), next);
	    if(next==curr->end()) {
	      WOOL_DEBUG("did not find node " << name );
	      return;
	    }
	  }
	  curr = next;
	}
      }
      std::map<std::string, int> known_names;
      for(xml::node::const_iterator child = curr->begin(); child!=curr->end(); child++) {
	if(child->get_type()==xml::node::type_element) {
	  std::ostringstream val_stream;
	  std::string name = child->get_name();
	  
	  val_stream << name << '%' << known_names[name];
	  s.push_back(val_stream.str());
	  known_names[name] ++;
	}
      }
    }
  }

  void Configuration::extract( const std::string & path, std::map<std::string,std::string> & s ) const {
    boost::mutex::scoped_lock lm(m_guard);
    WOOL_DEBUG("extract " << path );
    wool::url u(path);
    wool::url::path_type & p = u.path;
    s.clear();
    if(m_documentParser) {
    
      if( (p.empty()) || (m_documentParser->get_document().get_root_node().get_name() != p.front()))
	return;
      wool::url::path_type::iterator i = p.begin();
      i++;
      xml::node::const_iterator curr = m_documentParser->get_document().get_root_node().self();
      for(;i!=p.end();i++) {
	if(!i->empty()) {
	  int e_name = i->find('%');
	  std::string name (*i,0,e_name);
	  int number = 0;

	  if(e_name!=std::string::npos) {
	    std::istringstream s( std::string (*i,e_name+1,i->length()) );
	    s >> number;
	  }

	  if(curr->get_type()!=xml::node::type_element) {
	    WOOL_DEBUG("found a non element node "<< *i);
	    return;
	  }
	  xml::node::const_iterator next = curr->begin();
	  for( int j =0; j<=number; j++ ) {
	    if(j!=0)
	      next++;
	    next = curr->find(name.c_str(), next);
	    if(next==curr->end()) {
	      WOOL_DEBUG("did not find node " << name );
	      return;
	    }
	  }
	  curr = next;
	}
      }
      s["_self_"] = curr->get_content();
      for (xml::node::const_iterator m = curr->begin(); m!=curr->end(); ++m)
	if(m->get_type()==xml::node::type_element)
	  s[m->get_name()] = m->get_content();   
    }
  }
  
  void Configuration::extract( const std::string & path, std::list<std::string> & s ) const {
    boost::mutex::scoped_lock lm(m_guard);
    wool::url u(path);
    wool::url::path_type & p = u.path;
    s.clear();
    if(m_documentParser) {
      if( (p.empty()) || (m_documentParser->get_document().get_root_node().get_name() != p.front()))
	return;
      wool::url::path_type::iterator i = p.begin();
      i++;
      xml::node::const_iterator curr = m_documentParser->get_document().get_root_node().self();
      xml::node::const_iterator parent = curr;
      for(;i!=p.end();i++) {
	if(!i->empty()) {
	  int e_name = i->find('%');
	  std::string name (*i,0,e_name);
	  int number = 0;

	  if(e_name!=std::string::npos) {
	    std::istringstream s( std::string (*i,e_name+1,i->length()) );
	    s >> number;
	  }

	  if(curr->get_type()!=xml::node::type_element) {
	    WOOL_DEBUG("found a non element node "<< *name);
	    return;
	  }
	  xml::node::const_iterator next = curr->begin();
	  for( int j =0; j<=number; j++ ) {
	    if(j!=0)
	      next++;
	    next = curr->find(name.c_str(), next);
	    if(next==curr->end()) {
	      WOOL_DEBUG("did not find node " << name );
	      return;
	    }
	  }
	  parent = curr;
	  curr = next;
	}
      }
      do {
	std::string val(curr->get_content());
	s.push_back(val);
      }while((parent != curr++)&&((curr= parent->find(p.back().c_str(),curr))!=parent->end()));
    
    }
  }
}
