/**************************************************************************
*   Copyright (C) 2004 by Vukasin Toroman                                 *
*   vukasin@toroman.org                                                   *
**************************************************************************/
/************************************************************************** 
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifndef WOOLDLMODULE_H
#define WOOLDLMODULE_H

#include "../base/exception.h"
#include "../base/cvs_info.h"

namespace wool
{

  /**
   * dlmodule wraps the OS's ability to load module (dynamic libraries) and
   * access symbols inside them.
   * @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
   */
  class dlmodule
  {
  public:
    /**
     *  cvs info
     */
    VERSION_INFO_INLINE("$Id: dlmodule.h,v 1.4 2007/01/11 15:18:26 vtoroman Exp $");
    /**
     *  cvs info
     */
    VERSION_INFO_DECL;
    /**
     *  definition of symbol_name_type
     */
    typedef std::string symbol_name_type;
    /**
     * definition of path_type
     */
    typedef std::string path_type;
    /**
     *  definition of a pointer to a symbol_type
     */
    typedef void * symbol_type;
    /**
     * definition of a pointer to an os_module_handle_type
     */
    typedef void * os_module_handle_type;
    /**
     *  definition load_flag_type
     */
    typedef int load_flag_type;
    /**
     * symbol values are first resolved when needed
     */
    static const load_flag_type lazy;
    /**
     *  all undefined symbols in the library are resolved before the constructor returns.
     */
    static const load_flag_type now;
    /**
     * external symbols defined in the library will be made available for
     * symbol  resolution  of  subsequently  loaded  libraries.
     */
    static const load_flag_type global;
    /**
     * The converse of the global flag.
     */
    static const load_flag_type local;


    /**
     * load_failure indicates a failure while loading a module from disc
     * @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
     */
    class load_failure
    :public wool::exception
    {
      public:
        /**
         *  load_failure indicates a failure while loading a module from disc
         */
        load_failure (wool::exception::string_type) throw ();
        virtual ~load_failure () throw ();
    };

    /**
     * symbol_lookup_failure indicates a failure while loading a symbol from a module
     * @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
     */
    class symbol_lookup_failure
    :public wool::exception
    {
      public:
        /**
         *  symbol_lookup_failure indicates a failure while loading a symbol from a module
         */
        symbol_lookup_failure  (wool::exception::string_type) throw ();
        virtual ~symbol_lookup_failure  () throw ();
    };

    /**
     * Creates an empty module wrapper. Loads no modules and throws no exceptions
     */
    dlmodule () throw ();

    /**
     * Create a copy of src. The module used by src is re-opened.
     * @param src the original instance
     */
    dlmodule (const dlmodule & src) throw ();

    /**
     * Create a module wrapper and load a module from fs.
     * @param path location of the module in fs
     * @param flags for loading the module
     */
    dlmodule (const path_type & path, load_flag_type flags = global|now  ) throw();

    /**
     * Destroy the reference to the loaded module (if any)
     * @see close
     */
    virtual ~dlmodule () throw();


    /**
     * Get the currently loaded module. If no module is loaded return an empty string.
     * @return the path of the currently loaded module
     */
    inline path_type path() const throw() { return m_path; }

    /**
     * Load a module from path. In case of an error an exception of type dlmodule::failure
     * is thrown.
     * @param path the path of the module on fs
     * @param flags flags for loading the module
     * @throw exception dlmodule::load_failure
     */
    void path(const path_type & path, load_flag_type flags = global|now ) throw( dlmodule::load_failure);

    /**
     *  If a module is currently loaded, the references to it are cleaned up.
     *  No error is thrown on failure (some OS's do not support dl unload).
     */
    void close() throw ();


    /**
     *  Unload the current module and load the module from src
     *  @param src module
     *  @return reference to loaded module
     */
    dlmodule & operator &=(const dlmodule & src) throw ();

    /**
     * Returns true if a library is currently loaded else false
     * @return true if a library is currently loaded else false
     */
    inline operator bool () const { return (m_handle!=null_handle);}

  protected:
    /**
     * Load the symbol "name" and return a reference to it
     * @param name symbol name
     * @return the reference to name
     * @throws exception dlmodule::symbol_lookup_failure if the symbol cannot be found
     * or no module is loaded
     */
    symbol_type symbol (const symbol_name_type &name) const throw( dlmodule::symbol_lookup_failure);
  private:
    path_type m_path;
    os_module_handle_type m_handle;
    load_flag_type m_flags;

    static const os_module_handle_type null_handle;
  };

};

#endif
