/***************************************************************************
 *   Copyright (C) 2005 by Vukasin Toroman                                 *
 *   vtoroman@astaro.de                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "../jackal/xmlrpc.h"
#include "../swool/streambuf_bio.h"

#include <boost/bind.hpp>
#include "../socket/unixdomain.h"
#include <iconv.h>

namespace jackal
{

  namespace xmlrpc
  {

    static const std::string name_b64 ( "base64" );
    static const std::string name_empty  ( "" );

    static const std::string name_i4     ( "i4" );
    static const std::string name_int    ( "int" );
    static const std::string name_bool   ( "boolean" );
    static const std::string name_double ( "double" );

    static const std::string name_string ( "string" );
    static const std::string name_date   ( "dateTime.iso8601" );

    static const std::string name_array  ( "array" );
    static const std::string name_struct ( "struct" );
    static const std::string name_param  ( "param" );
    static const std::string name_value  ( "value" );
    static const std::string name_member ( "member" );
    static const std::string name_name   ( "name" );
    static const std::string name_fault  ( "fault" );

    static const std::string name_method ( "methodName" );


    struct char_encoder
    {
      std::ostream & out;
      bool b64_enabled;
      char_encoder( std::ostream & io, bool b )
	: out(io)
	, b64_enabled(b)
      {}

      void operator () (const std::string & c )
      {
        if(c.length()==0) {
          out << "<string/>";
          return;
        }
        if(this->b64_enabled)
	  {
	    static const char * head = "<text>";
	    static const char * tail = "</text>";
	    bool ok = true;

	    if(c.length()>0) {
	      XML_Parser parser = XML_ParserCreate("UTF-8");
	      XML_Parse(parser, head, 6,0);
	      if(XML_Parse(parser, c.c_str(), c.length(),0)==XML_STATUS_ERROR) {
	  	  ok = false;
  	      } else {
	     	  if(XML_Parse(parser, tail, 7,1)==XML_STATUS_ERROR)
		    ok = false;
	      }
	      XML_ParserFree(parser); 
	    }

            if(ok) {
	      out << "<string>";
	      std::for_each(c.begin(), c.end(), *this);
	      out << "</string>";
	    } else {
	      out << "<base64>";
	      {
		wool::swool::streambuf_bio output_bio (out.rdbuf());
		wool::swool::bio encoder(wool::swool::bio::f_base64);
		wool::swool::bio_stream stream(encoder);

		encoder.set_flags(BIO_FLAGS_BASE64_NO_NL);
		encoder.push(output_bio);
		stream << c << std::flush;
	      }
	      out << "</base64>";
	    }
	  } else {
	    std::for_each(c.begin(), c.end(), *this);
	  }
      }
      
      void operator () (const char & c )
      {
        static const char * lt = "&lt;";
        static const char * gt = "&gt;";
        static const char * amp =  "&amp;";
        static const char * apos = "&apos;";
        static const char * quot = "&quot;";

        switch (c)
	  {
	  case '<':
            out.write(lt,4);
            break;
	  case '>':
            out.write(gt,4);
            break;
	  case '&':
            out.write(amp,5);
            break;
	  case '\'':
            out.write(apos,6);
            break;
	  case '"':
            out.write(quot,6);
            break;
	  default:
            out.put(c);
	  }
      }
    };

    struct member_encoder
    {
      std::ostream & out;
      void operator () (const jackal::value::map_type::value_type & c );
    };

    struct value_encoder
    {
      std::ostream & out;
      void operator () (const jackal::value & c )
      {
        out << "<value>";
        switch(c.get_type())
	  {
	  case jackal::value::string:
            {
	      char_encoder enc (out,true);
	      enc(c.get_string());
	      break;
            }
	  case jackal::value::number:
            {
	      out << "<double>";
	      out <<c.get_number();
	      out << "</double>\n";
	      break;
            }
	  case jackal::value::map:
            {
	      out << "<struct>\n";
	      member_encoder enc = {out};
	      const jackal::value::map_type & m = c.get_map();
	      std::for_each(m.begin(), m.end(), enc);
	      out << "</struct>\n";
	      break;
            }
	  case jackal::value::array:
            {
	      out << "<array>\n<data>\n";
	      value_encoder enc = {out};
	      const jackal::value::array_type & a = c.get_array();
	      std::for_each(a.begin(), a.end(), enc);
	      out << "</data>\n</array>\n";
	      break;
            }
	  }
        out << "</value>\n";
      }
    };
    int func0 ()
    {
      int a = 0;
      int b = 34;
      return a+b;
    }
    void member_encoder::operator () (const jackal::value::map_type::value_type & c )
    {
      value_encoder v_enc = {out};
      char_encoder n_enc (out, false);

      out << "<member>\n<name>";

      n_enc(c.first);
      out << "</name>\n";
      v_enc(c.second);
      out << "</member>\n";
    }

    struct param_encoder
    {
      std::ostream & out;
      void operator () (const jackal::value & c )
      {
        value_encoder enc = {out};
        out << "<param>\n";
        enc(c);
        out << "</param>\n";
      }
    };

    struct filter_base64
    {
      std::string & res;
      void operator () (const char & c)
      {
        if(((c>='0')&&(c<='9'))||((c>='A')&&(c<='Z'))||((c>='a')&&(c<='z'))||(c=='=') )
	  {
            res.push_back(c);
	  }
        else
	  {}
      }
    }
    ;

    void Saver::save(const jackal::Request & r, std::ostream & out ) const
    {
      char_encoder enc_n (out, false);
      out << "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<methodCall>\n<methodName>";
      enc_n(r.method);
      out << "</methodName>\n<params>\n";
      param_encoder enc = {out};
      std::for_each(r.parameters.begin(), r.parameters.end(), enc);
      out << "</params>\n</methodCall>\n";

    }

    void Saver::save(const jackal::Response & r, std::ostream & out ) const
    {
      out << "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<methodResponse>\n<params>\n";
      param_encoder enc = {out};
      std::for_each(r.parameters.begin(), r.parameters.end(), enc);
      out << "</params>\n</methodResponse>\n";
    }

    void Saver::save(const jackal::exception & r, std::ostream & out ) const
    {
      char_encoder enc_n (out, true);
      out   << "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<methodResponse>\n<fault>\n<value>\n"
	    << "<struct>\n<member>\n<name>faultCode</name>\n<value><int>"
	    << r.id()
	    << "</int></value></member>\n<member>\n<name>faultString</name>\n<value>";

      enc_n(r.what());

      out << "</value>\n</member>\n</struct>\n</value>\n</fault>\n</methodResponse>\n";
    }


    Loader::Loader( jackal::Request & r )
      : method_name(&r.method), parameters(r.parameters)
    {
      parser = XML_ParserCreate(NULL);
      XML_SetUserData(parser, this);
      XML_SetElementHandler(parser, Loader::startElement, Loader::endElement);
      XML_SetCharacterDataHandler(parser, Loader::charHandler);
    }

    Loader::Loader( jackal::Response & r )
      : method_name(NULL), parameters(r.parameters)
    {
      parser = XML_ParserCreate(NULL);
      XML_SetUserData(parser, this);
      XML_SetElementHandler(parser, Loader::startElement, Loader::endElement);
      XML_SetCharacterDataHandler(parser, Loader::charHandler);
    }

    Loader::~Loader()
    {
      XML_ParserFree(parser);
    }

    void Loader::load (std::istream & i)
    {
      char buffer[128];
      while(i.good())
	{
	  i.read(buffer,128);
	  int j = i.gcount();
	  if(j>0)
	    {
	      if(XML_Parse(parser, buffer, j,0)==XML_STATUS_ERROR)
                throw jackal::exception(0, std::string("Expat error: ") + XML_ErrorString(XML_GetErrorCode(parser))) ;
	    }
	}
      if(XML_Parse(parser, buffer, 0,1)==XML_STATUS_ERROR)
        throw jackal::exception(0, std::string("Expat error: ") + XML_ErrorString(XML_GetErrorCode(parser))) ;
    }

    void Loader::startElement(void *userData, const char *name, const char **atts)
    {
      Loader * r = reinterpret_cast<Loader *>(userData);
      r->start_element(name);
    }

    void Loader::charHandler(void *userData, const XML_Char *s, int len)
    {
      Loader * r = reinterpret_cast<Loader *>(userData);
      r->text(std::string(s,len));
    }

    void Loader::endElement(void *userData, const char *name)
    {
      Loader * r = reinterpret_cast<Loader *>(userData);
      r->end_element( name);
    }

    void Loader::start_element(const std::string &name )
    {
      if(name==name_value)
	{
	  if(   (stage.back()!=Member)
                &&  (stage.back()!=Array)
                &&  (stage.back()!=Parameter)
                &&  (stage.back()!=Fault) )
	    {
	      throw jackal::exception(0,"Parse Error"); //error
	    }
	  if(stage.back()!=Array)
	    {
	      if(current_value.empty())
		{
		  throw jackal::exception(0,"Parse Error");
		}
	    }
	  else
	    {
	      current_value.back()->set_type(jackal::value::array);
	      current_value.push_back(&(*current_value.back()->get_array().insert(current_value.back()->get_array().end(),jackal::value::EmptyString)));
	    }
	  stage.push_back(Value);
	  implicit_string = true;
	  value_text.clear();
	}
      else if(name==name_param)
	{
	  if(!current_value.empty())
            throw jackal::exception(0,"Parse Error"); // error
	  if(!stage.empty())
            throw jackal::exception(0,"Parse Error"); // error
	  parameters.insert(  parameters.end(), jackal::value::EmptyString);
	  current_value.push_back(&(*parameters.rbegin()));
	  stage.push_back(Parameter);
	}
      else if(name==name_name)
	{
	  if(stage.back()!=Member)
            throw jackal::exception(0,"Parse Error"); //error
	  stage.push_back(Name);
	}
      else if((name==name_string) ||
	      (name==name_date))
	{
	  if(stage.back()!=Value)
            throw jackal::exception(0,"Parse Error");
	  implicit_string = false;
	  value_text.clear();
	  stage.push_back(String);
	}
      else if((name==name_double) ||
	      (name==name_bool) ||
	      (name==name_i4) ||
	      (name==name_int))
	{
	  if(stage.back()!=Value)
            throw jackal::exception(0,"Parse Error");
	  implicit_string = false;
	  value_text.clear();
	  stage.push_back(Number);
	}
      else if(name==name_b64)
	{
	  if(stage.back()!=Value)
            throw jackal::exception(0,"Parse Error");
	  implicit_string = false;
	  value_text.clear();
	  stage.push_back(Base64);
	}
      else if(name==name_array)
	{
	  if(stage.back()!=Value)
            throw jackal::exception(0,"Parse Error");
	  stage.push_back(Array);
	  implicit_string = false;
	  value_text.clear();
	}
      else if(name==name_struct)
	{
	  if(stage.back()!=Value)
            throw jackal::exception(0,"Parse Error");
	  stage.push_back(Struct);
	}
      else if(name==name_array)
	{
	  if(stage.back()!=Value)
            throw jackal::exception(0,"Parse Error");
	  stage.push_back(Array);
	}
      else if(name==name_member)
	{
	  if(stage.back()!=Struct)
	    {
	      throw jackal::exception(0,"Parse Error"); //error
	    }
	  current_name.clear();
	  stage.push_back(Member);
	}
      else if(name==name_method)
	{
	  if(!stage.empty())
	    {
	      throw jackal::exception(0,"Parse error"); //error
	    }
	  stage.push_back(MethodName);
	}
      else if(name==name_fault)
	{
	  if(!stage.empty())
	    {
	      throw jackal::exception(0,"Parse error"); //error
	    }
	  current_value.push_back(&fault_val);
	  stage.push_back(Fault);
	}
    }

    void Loader::end_element (const std::string &name)
    {
      if(name==name_param)
	{
	  if(stage.empty())
            throw jackal::exception(0,"Parse Error"); // error
	  if(stage.back()!=Parameter)
            throw jackal::exception(0,"Parse Error");
	  stage.pop_back();
	}
      else if(name==name_name)
	{
	  if(stage.empty())
            throw jackal::exception(0,"Parse Error"); // error 
	  if(stage.back()!=Name)
            throw jackal::exception(0,"Parse Error"); //error

	  (*current_value.back())[current_name] = jackal::value::EmptyString;
	  current_value.push_back(&((*current_value.back())[current_name]));
	  current_name.clear();
	  stage.pop_back();
	}
      else if(name==name_member)
	{
	  if(stage.empty())
            throw jackal::exception(0,"Parse Error"); // error
	  if(stage.back()!=Member)
            throw jackal::exception(0,"Parse Error"); //error
	  stage.pop_back();
	}
      else if(name==name_value)
	{
	  if(stage.empty())
            throw jackal::exception(0,"Parse Error"); // error
	  if(current_value.empty())
            throw jackal::exception(0,"Parse Error"); // error
	  if(stage.back()!=Value)
            throw jackal::exception(0,"Parse Error"); // error
	  stage.pop_back();
	  if(implicit_string) {
            switch( current_value.back()->get_type()) {
                  case jackal::value::map: 
                        //(*current_value.back())[current_name].set_string(value_text);
                        break;
                  case jackal::value::array:
                        //current_value.back()->get_array().push_back(value::string_type(value_text));
                        break;
                  default:
                        current_value.back()->set_string(value_text);
            }
        }
	  current_value.pop_back();
	  value_text.clear();
	}
      else if(name==name_struct)
	{
	  if(stage.empty())
            throw jackal::exception(0,"Parse Error"); // error
	  if(stage.back()!=Struct)
            throw jackal::exception(0,"Parse Error"); // error
	  stage.pop_back();
	}
      else if(name==name_array)
	{
	  if(stage.empty())
            throw jackal::exception(0,"Parse Error"); // error
	  if(stage.back()!=Array)
            throw jackal::exception(0,"Parse Error"); // error
	  stage.pop_back();
	}
      else if((name==name_string) ||
	      (name==name_date))
	{
	  if(stage.empty())
            throw jackal::exception(0,"Parse Error"); // error
	  if(stage.back()!=String)
            throw jackal::exception(0,"Parse Error"); // error
	  stage.pop_back();
	  current_value.back()->set_string(value_text);
	}
      else if((name==name_double) ||
	      (name==name_bool) ||
	      (name==name_i4) ||
	      (name==name_int))
	{
	  if(stage.empty())
            throw jackal::exception(0,"Parse Error"); // error
	  if(stage.back()!=Number)
            throw jackal::exception(0,"Parse Error"); // error
	  stage.pop_back();
	  current_value.back()->set_string(value_text);
	}
      else if(name==name_b64)
	{
	  if(stage.empty())
            throw jackal::exception(0,"Parse Error"); // error
	  if(stage.back()!=Base64)
            throw jackal::exception(0,"Parse Error"); // error
	  if(current_value.empty())
            throw jackal::exception(0,"Parse Error");
	  else
	    {
	      std::string res;
	      {
                wool::swool::bio input_bio(BIO_new_mem_buf(const_cast<char *>
							   (value_text.data()),value_text.size()));
                wool::swool::bio decoder(wool::swool::bio::f_base64);

                decoder.set_flags(BIO_FLAGS_BASE64_NO_NL);
                input_bio.set_close(BIO_NOCLOSE);
                decoder.push(input_bio);
                char buffer[64];
                while(!decoder.eof())
		  {
                    res.append(buffer, decoder.read(buffer,64));
		  }
	      }
	      current_value.back()->set_string(res);
	    }
	  stage.pop_back();
	}
      else if(name==name_method)
	{
	  if(stage.empty())
            throw jackal::exception(0,"Parse Error"); // error
	  if(stage.back()!=MethodName)
            throw jackal::exception(0,"Parse Error"); // error
	  stage.pop_back();
	}
      else if(name==name_fault)
	{
	  if(stage.empty())
            throw jackal::exception(0,"Parse Error"); // error
	  if(stage.back()!=Fault)
            throw jackal::exception(0,"Parse Error"); // error
	  throw jackal::exception((int)fault_val["faultCode"],(std::string)fault_val["faultString"]);
	}

    }



    void Loader::text (const std::string &contents)
    {

      if(stage.empty())
	{
	  return; // loose text
	}


      switch(stage.back())
	{
	case Value:
	  if(current_value.empty())
            throw jackal::exception(0,"Parse Error");
	  if(implicit_string)
            value_text.append(contents);
	  break;
	case Number:
	case String:
	  if(current_value.empty())
            throw jackal::exception(0,"Parse Error");
	  value_text.append(contents);
	  break;
	case Base64:
	  if(current_value.empty())
            throw jackal::exception(0,"Parse Error");
	  {
            filter_base64 fltr = {value_text};
            std::for_each(contents.begin(), contents.end(), fltr);
	  }
	  break;
	case Name:
	  if(current_value.empty())
            throw jackal::exception(0,"Parse Error");
	  current_name.append(contents);
	  break;
	case MethodName:
	  if(method_name != NULL)
            method_name->append(contents);
	  break;
	default:
	  ;
	};
    }


    Adapter::Adapter( MethodRequestHandler & h ) : m_handler(h)
    {}

    Adapter::~Adapter()
    {}

    void Adapter::handle ( const jackal::Request & req, jackal::Response & res )
    {
      this->m_handler.handle(req, res);
    }

    void Adapter::handle ( const wool::http::request & http_req, wool::http::response & http_res )
    {

      jackal::xmlrpc::Saver svr;
      jackal::Request req;
      jackal::Response res;
      res.request = &req;
      jackal::xmlrpc::Loader ldr(req);

      http_res.message = "OK";
      http_res.status_code = 200;
      http_res.protocol = "HTTP/1.0";
      http_res.body.clear();
      http_res.header.clear();
      http_res.header.insert("Content-type","text/xml");
      http_res.header.insert("Connection","close");

      wool::swappable_data::ostream out_res;
      wool::swappable_data::istream in_req;

      in_req.open(http_req.body);
      out_res.open(http_res.body);

      try
	{
	  ldr.load(in_req);
	  handle(req,res);
	  svr.save(res, out_res);
	}
      catch (const jackal::exception & e)
	{
	  svr.save(e, out_res);
	}

      out_res.close();
      in_req.close();

      std::ostringstream os;

      os << http_res.body.size();
      http_res.header.insert("Content-Length",os.str());

    }

    bool Adapter::shouldHandle( const http::Request & req ) const
    {
      static const std::string post = "POST";
      return (req.method == post) && (req.body.size()>0);
    }

    Proxy::Proxy( const std::string & unix_path, const std::string & object_path ) : m_socket(NULL), m_stream(NULL), m_url(object_path), m_address(wool::local::address(unix_path))
    {
      for(wool::url::path_type::const_iterator i=m_url.path.begin();i!=m_url.path.end();++i)
	{
	  m_path += "/";
	  m_path += *i;
	}
    }

    Proxy::Proxy( const std::string & url ) : m_socket(NULL), m_stream(NULL), m_url(url), m_address(wool::inet::address(m_url.host))
    {
      for(wool::url::path_type::const_iterator i=m_url.path.begin();i!=m_url.path.end();++i)
	{
	  m_path += "/";
	  m_path += *i;
	}
    }

    Proxy::~Proxy()
    {
      delete m_stream;
      delete m_socket;
    }

    void Proxy::connect()
    {
      if((m_stream==NULL)||(m_socket==NULL))
	{
	  std::cerr << "reconnect " << std::endl;
	  delete m_stream;
	  delete m_socket;
	  m_socket = new wool::socket((wool::socket::domain)m_address.get_family(),wool::socket::stream_socket, wool::socket::unspec);
	  m_socket->set_timeout(300);
	  m_socket->connect( m_address);
	  m_stream = new  wool::iostream<wool::socket>( *m_socket );
	  m_stream->exceptions(m_stream->exceptions()|std::iostream::failbit);
	}
    }


    void Proxy::do_io(jackal::http::Request & http_req, jackal::http::Response & http_res )
    {
      connect();
      try
	{

	  *m_stream << http_req;
	  *m_stream >> http_res;

	}
      catch (const std::ios::failure & e)
	{
	  std::cerr << "error " << e.what() << " - reconnecting"<<std::endl;
	  delete m_stream;
	  delete m_socket;
	  //retry
	  m_stream = NULL;
	  m_socket = NULL;
	  connect();
	  *m_stream << http_req;
	  *m_stream >> http_res;
	}
    }

    void Proxy::handle ( const jackal::Request & req, jackal::Response & res )
    {
      jackal::http::Request http_req;
      jackal::http::Response http_res;
      Saver rs;
      Loader rl(res);

      wool::swappable_data::ostream save_stream;
      save_stream.open(http_req.body);
      rs.save(req, save_stream);
      save_stream.close();

      std::ostringstream bs_os;
      bs_os << http_req.body.size();

      http_req.method = "POST";
      http_req.protocol = "HTTP/1.0";
      http_req.path = m_path;

      http_req.header.insert("Content-Type", "text/xml");
      http_req.header.insert("User-Agent", "libwool");
      http_req.header.insert("Content-Length", bs_os.str());
      http_req.header.insert("Connection", "close");

      do_io(http_req, http_res);

      wool::swappable_data::istream load_stream;
      load_stream.open(http_res.body);
      rl.load(load_stream);
      load_stream.close();
    }


  };
};

