 
/**************************************************************************
*   Copyright (C) 2006 by Vukasin Toroman                                 *
*   vukasin@toroman.org                                                   *
**************************************************************************/
/************************************************************************** 
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
#include <syslog.h>

#include "../log/event.h"

namespace wool
{

std::map <int, Entry::system_type> Entry::systems;

const std::string Entry::DEBUG("debug");
const std::string Entry::INFO("info");
const std::string Entry::WARN("warn");
const std::string Entry::ERROR("error");
const std::string Entry::CRIT("crit");

const unsigned int Entry::DEBUG_LEVEL = 4;
const unsigned int Entry::INFO_LEVEL = 3;
const unsigned int Entry::WARN_LEVEL = 2;
const unsigned int Entry::ERROR_LEVEL = 1;
const unsigned int Entry::CRIT_LEVEL = 0;

bool Entry::s_initialized = false;


std::map <std::string, unsigned int>  Entry::s_logLevel;


DocumentParser::DocumentParser(std::map <int, Entry::system_type> & systems):
		m_map(systems),
		current_system(NULL),
		current_subsystem(NULL)
		{
		}
		
long long Entry::parseID(std::string id){
	long long res = 0l;
	for (unsigned i = 0; i< id.length(); i++)
		res = res * 36 + ( (id[i]>'9')?(id[i]-'A'+10):(id[i]-'0'));
	return res;
}

std::string Entry::createID(long long id){
	int first = id/36;
	int second = id%36;
	std::stringstream ss;
	if(first>9)
		ss << (char)(first+'A'-10);
	else
		ss << first;
	if(second>9)
		ss << (char)(second+'A'-10);
	else
		ss << second;
	return ss.str();
}

bool DocumentParser::start_element(const std::string &name, const attrs_type &__attrs)
{
	attrs_type attrs(__attrs);
	if (name == "system")
	{
		int id = atoi(attrs["id"].c_str());
		WOOL_DEBUG("system id: " << id);
		Entry::system_type & val = m_map[id];
		val.name = attrs["name"];
		WOOL_DEBUG("system name: " << val.name);
		current_system = & val;
		return true;
	} 
	if (name == "subsystem")
	{
		if(current_system == NULL)
			return false;
		int id = attrs["id"][0];
		id = (id>'9')?(id-'A'+10):(id-'0');
		WOOL_DEBUG("subsystem id: " << id);
		Entry::subsystem_type & val = current_system->subsystems[id];
		val.name = attrs["name"];
		WOOL_DEBUG("subsystem name: " << val.name);
		current_subsystem = & val;
		return true;

	}
	if (name == "event")
	{
		if(current_subsystem == NULL)
			return false;
		long long id = Entry::parseID(attrs["id"]);
		WOOL_DEBUG("event id: " << id);
		Entry::event_type & val = current_subsystem->events[id];
		val.name = attrs["name"];
		WOOL_DEBUG("event name: " << val.name);
		val.level = attrs["level"];
		WOOL_DEBUG("event level: " << val.level);
		return true;
	}
	return true;

}

bool DocumentParser::end_element(const std::string &name )
{
	if (name == "system")
	{
		current_system = NULL;
		return true;
	}
	if (name == "subsystem")
	{
		current_subsystem = NULL;
		return true;

	}
	
	return true;

}


	Entry::Entry(	const unsigned long long & event_id):
	  m_committed(false) {
		initialize("/etc/woollog.xml");
		long eventId = event_id%(36*36);
		WOOL_DEBUG("eventId: " << eventId);
		
 		long subId = (event_id/(36*36))%(36);
		WOOL_DEBUG("subsystemID: " << subId);		
		
		long systemId = event_id/(36*36*36);
		WOOL_DEBUG("systemId: " << systemId);		

		system_type system = systems[systemId];
		subsystem_type subsystem = system.subsystems[subId];
		event_type event = subsystem.events[eventId];
		
		std::stringstream ss;
		ss << systemId;
		subId = subId%36;
		if(subId>9)
			ss << (char)(subId+'A'-10);
		else
			ss << subId;
		
		ss << createID(eventId);
		WOOL_DEBUG("id: "  << ss.str());
		add( "id", ss.str());		
		add( "severity", event.level);
		add( "sys", system.name);
		add( "sub", subsystem.name);
		add( "name", event.name);
		m_level = s_logLevel[event.level];
	}

	void Entry::initialize(const std::string & path) {
		if(s_initialized) 
			return;
		
		DocumentParser parser(systems);
		parser.parse_file(path.c_str());
		
		s_logLevel[DEBUG] = DEBUG_LEVEL;
		WOOL_DEBUG(DEBUG << " " << DEBUG_LEVEL);
		s_logLevel[INFO] 	= INFO_LEVEL;
		WOOL_DEBUG(INFO << " " << INFO_LEVEL);
		s_logLevel[WARN] 	= WARN_LEVEL;
		WOOL_DEBUG(WARN << " " << WARN_LEVEL);
		s_logLevel[ERROR] = ERROR_LEVEL;
		WOOL_DEBUG(ERROR << " " << ERROR_LEVEL);
		s_logLevel[CRIT] 	= CRIT_LEVEL;
		WOOL_DEBUG(CRIT << " " << CRIT_LEVEL);
		s_initialized = true;
	}

	Entry::~Entry()	{}


	Entry::Entry(const Entry & e):
		m_params(e.m_params),
		m_committed(e.m_committed) {
	}
	
	Entry::ParamList & Entry::get_params() {
		return m_params;
	}
	
	Entry & Entry::add( const std::string & key, const std::string & value ) {
		boost::mutex::scoped_lock lock(m_mut);	
		value_pair pair;
		pair.first = key;
		char buff[6];
		for(std::string::const_iterator it = value.begin();
		    it!=value.end(); 
		    it++) {
		  if((*it <  ' ') ||
		     (*it == '"') ||
		     (*it == '%') ||
		     (*it > '~') ) {
		    sprintf(buff, "%%%02X", *it);
		    pair.second.append(buff);
		  } else {
		    pair.second.push_back(*it);
		  }
		}
		m_params.push_back(pair);	
		return *this;
	}

	bool Entry::is_committed () const {
		return m_committed;
	}
	
  void Entry::commit() {
    boost::mutex::scoped_lock lock(m_mut);
    if(m_params.size() < 5){
      throw event_error("wrong number of parameters");
    } else {
      std::stringstream ss;
      ParamList::const_iterator it;
      for(it = m_params.begin(); it != m_params.end(); ++it) {
	ss << (*it).first;
	ss << "=\"";
	ss << (*it).second;
	ss << "\" ";
      }
      WOOL_DEBUG("commit: " << ss.str());
      WOOL_DEBUG("loglevel: " << (  m_level));
      Log::log( LOG_INFO|LOG_DAEMON, m_level, ss.str().c_str());
      m_committed = true;
    }
  }
  
	
};

