/**************************************************************************
*   Copyright (C) 2004 by Vukasin Toroman                                 *
*   vukasin@toroman.org                                                   *
**************************************************************************/
/************************************************************************** 
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifndef WOOLPCRE_H
#define WOOLPCRE_H


#include <pcre.h>
#include "../base/wool_stl_includes.h"
#include "../log/log.h"
#include "../base/exception.h"


#define WOOL_PCRE_MAX_VECTOR_SIZE 172

namespace wool
{

/**
 * This is the exption class for the pcre class. It is thrown if an error occur at compile or study a wool::pcre.
 */
class pcre_error
            :public wool::exception
{
public:
    /**
     * @param s message
     * @throw exception wool::exception
     */
    inline pcre_error(wool::exception::string_type s) throw ():wool::exception(s)
    {}
    inline ~pcre_error() throw ()
    {}
}
;


/**
 * This class implements the Perl Compatible Regular Expressions.
 * @author Vukasin Toroman <vtoroman@astaro.com> (C) Astaro AG 2005
 */

template <typename charT, typename traitsT=std::char_traits<charT> >
class pcre
{
private:
    wool::pcre<charT,traitsT> & operator = (const wool::pcre<charT,traitsT> & src)
    {}
    ;
    pcre(const pcre<charT,traitsT> & src)
    {}
    ;

    //Result __match(const std::string & s,bool _all) const;
    //std::string __replace(const std::string & s,const std::string & d,bool _all) const;
    //void compile();

    std::string m_expression;
    int m_flags;
    ::pcre *m_pcre;
    ::pcre_extra *m_extra;

public:
    /**
     *  cvs info
     */
    VERSION_INFO_INLINE("$Id: pcre.h,v 1.4 2007/01/11 15:18:32 vtoroman Exp $");
    /**
     *  string typedef
     */
    typedef std::basic_string<charT,traitsT> string_type;

    /**
     * @param str expression pattern
     * @param flags compile options
     * @see pcre::parse_options
     * @throw pcre_error wool::pcre_error While the compalation an error occur this exception will be thrown.
     */
    explicit pcre(const std::string & str ,int flags=0)
            :m_expression(str)
            ,m_flags(flags)
            ,m_pcre(NULL)
            ,m_extra(NULL)
    {
        const char *error = NULL;
        int erroffset;

        m_pcre = pcre_compile(  m_expression.c_str(),
                                m_flags,
                                &error,
                                &erroffset,
                                NULL);
        if (m_pcre == NULL)
        {
            std::ostringstream  errmsg;
            errmsg << "Error at '"
            << "'" << std::string(m_expression,erroffset,10) << "' "
            << "while compiling '" << m_expression << "': "
            << "'" << error<< "'" << std::endl;
            throw wool::pcre_error(errmsg.str());
        }
    }

    /**
     * free the compiled pcre pattern
     */
    ~pcre()
    {
        WOOL_DEBUG("deleting pattern: "<<m_expression);
        if(m_extra!=NULL)
        {
            WOOL_DEBUG("pattern is studied - removing m_extra");
            pcre_free(m_extra);
        }

        if(m_pcre!=NULL)
        {
            WOOL_DEBUG("pattern is compiled - removing m_pcre");
            pcre_free(m_pcre);
        }
        WOOL_DEBUG("deleting pattern done");
    };

    /**
     * Studying the pattern produces additional information  that  will  help speed up matching.
     * @throw pcre_error wool::pcre_error While the studying an error occur this exception will be thrown.
     */
    void study ()
    {
        const char *error = NULL;
        if( m_extra != NULL )
            pcre_free(m_extra);
        m_extra = pcre_study(m_pcre, m_flags, &error);
        if(error!=NULL)
        {
            std::ostringstream errmsg;
            errmsg << "Error while studying "
            << "'" << m_expression << "':" << std::endl
            << error;
            throw wool::pcre_error(errmsg.str());
        }
    }

    /**
     * String operator
     * @return compiled pcre expression as string
     */
    operator std::string () const
    {
        return m_expression;
    }

    /**
     * @return compile options
     * @see pcre::parse_options
     */
    int flags() const
    {
        return m_flags;
    }

    /**
     * Checks the string s against the compiled pcre expressions and writes the result into dst.
     * @param dst destination insert_iteratorT for result
     * @param s string to check against the compiled pcre expressions
     */
    template <typename insert_iteratorT>
    inline void match (insert_iteratorT dst, const std::string & s) const
    {
        WOOL_DEBUG("matching '"<<s<<"' against '"<<m_expression<<"'");
        int vector[WOOL_PCRE_MAX_VECTOR_SIZE];
        int last_match_end=0;
        int match_count = pcre_exec(
                              m_pcre,                 // the compiled pattern
                              m_extra,                // study result
                              s.data(),               // the subject string
                              s.length(),             // the length of the subject
                              last_match_end,         // start at last match end
                              0,                      // options
                              vector,               // vector for substring information
                              WOOL_PCRE_MAX_VECTOR_SIZE);  // max number of elements in the vector
        for(int i=0;i<match_count;i++)
        {
            long loc = vector[i*2];
            long len = vector[i*2+1]-vector[i*2];
            *dst=((loc>=0)&&(len>0))?std::string(s,vector[i*2],vector[i*2+1]-vector[i*2]):std::string();
            last_match_end=std::max(vector[i*2+1],last_match_end);
        }
        WOOL_DEBUG("done matching '"<<s<<"' against '"<<m_expression<<"'");
    }

    /**
     * @param dst destination stream
     * @param s string to check against the compiled pcre expressions
     * @param d If an expression will be found in the string s it will be replaced with this.
     * @return count of the matches
     */
    int replace (std::ostream & dst, const std::string & s,const std::string & d) const
    {
        int match_count;
        int total_matches = 0;
        unsigned last_match_end=0;
        int vector[WOOL_PCRE_MAX_VECTOR_SIZE];

        match_count = pcre_exec(
                          m_pcre,                // the compiled pattern
                          m_extra,               // study result
                          s.data(),              // the subject string
                          s.length(),            // the length of the subject
                          last_match_end,        // start at last match end
                          0,                     // options
                          vector,              // vector for substring information
                          WOOL_PCRE_MAX_VECTOR_SIZE); // max number of elements in the vector

        for(int i=0;i<match_count;i++)
        {
            long loc = vector[i*2];
            long len = vector[i*2+1]-vector[i*2];

            if((loc>0)&&(len>0)) {
                dst.write( s.data()+last_match_end,  vector[i*2]-last_match_end );
                dst.write( d.data(), d.length());
                last_match_end = vector[i*2+1];
                total_matches ++;
            }
        }

        if(last_match_end<s.length())
            dst.write( s.data()+last_match_end, s.length()-last_match_end);

        return total_matches;
    }
    /**
     * @param dst destination stream
     * @param s string to check against the compiled pcre expressions
     * @param d If an expression will be found in the string s it will be replaced with this.
     * @return count of the matches
     */
    int replace_all (std::ostream & dst, const std::string & s,const std::string & d) const
    {
        int match_count;
        int total_matches = 0;

        std::string subject(s);
        std::ostringstream res;
        while ((match_count=replace(res,subject,d))>0) {
            total_matches += match_count;
            subject = res.str();
            res.str("");
        }
        dst << res.str();
        return total_matches;
    }

    /**
     * Parse the given options to set the compile options.
     * Following options are supported: <br>
     * m - PCRE_MULTILINE <br>
     * s - PCRE_DOTALL <br>
     * x - PCRE_EXTENDED <br>
     * i - PCRE_CASELESS <br>
     * @see \ref pcre_options
     * @return options
     */
    static int parse_options (const std::string & options_str)
    {
        int options = 0;
        if(options_str.find('m')<options_str.length())
            options = options | PCRE_MULTILINE;
        if(options_str.find('s')<options_str.length())
            options = options | PCRE_DOTALL;
        if(options_str.find('x')<options_str.length())
            options = options | PCRE_EXTENDED;
        if(options_str.find('i')<options_str.length())
            options = options |  PCRE_CASELESS;
        return options;
    }

};
/**
        \page  pcre_options Compile options for Perl-compatible regular expressions
        @see http://www.pcre.org/pcre.txt<p>
        The following names for option bits are defined:<p>

        \section PCRE_ANCHORED

       If this bit is set, the pattern is forced to be "anchored", that is, it
       is constrained to match only at the first matching point in the  string
       that  is being searched (the "subject string"). This effect can also be
       achieved by appropriate constructs in the pattern itself, which is  the
       only way to do it in Perl.

        \section PCRE_AUTO_CALLOUT

       If this bit is set, pcre_compile() automatically inserts callout items,
       all with number 255, before each pattern item. For  discussion  of  the
       callout facility, see the pcrecallout documentation.

        \section PCRE_CASELESS

       If  this  bit is set, letters in the pattern match both upper and lower
       case letters. It is equivalent to Perl's  /i  option,  and  it  can  be
       changed  within  a  pattern  by  a (?i) option setting. When running in
       UTF-8 mode, case support for high-valued characters is  available  only
       when PCRE is built with Unicode character property support.

        \section PCRE_DOLLAR_ENDONLY

       If  this bit is set, a dollar metacharacter in the pattern matches only
       at the end of the subject string. Without this option,  a  dollar  also
       matches  immediately before the final character if it is a newline (but
       not before any  other  newlines).  The  PCRE_DOLLAR_ENDONLY  option  is
       ignored if PCRE_MULTILINE is set. There is no equivalent to this option
       in Perl, and no way to set it within a pattern.

        \section PCRE_DOTALL

       If this bit is set, a dot metacharater in the pattern matches all char-
       acters,  including  newlines.  Without  it, newlines are excluded. This
       option is equivalent to Perl's /s option, and it can be changed  within
       a  pattern  by  a  (?s)  option  setting. A negative class such as [^a]
       always matches a newline character, independent of the setting of  this
       option.

        \section PCRE_EXTENDED

       If  this  bit  is  set,  whitespace  data characters in the pattern are
       totally ignored except  when  escaped  or  inside  a  character  class.
       Whitespace  does  not  include the VT character (code 11). In addition,
       characters between an unescaped # outside a  character  class  and  the
       next newline character, inclusive, are also ignored. This is equivalent
       to Perl's /x option, and it can be changed within a pattern by  a  (?x)
       option setting.

       This  option  makes  it possible to include comments inside complicated
       patterns.  Note, however, that this applies only  to  data  characters.
       Whitespace   characters  may  never  appear  within  special  character
       sequences in a pattern, for  example  within  the  sequence  (?(  which
       introduces a conditional subpattern.

       \section PCRE_EXTRA

       This  option  was invented in order to turn on additional functionality
       of PCRE that is incompatible with Perl, but it  is  currently  of  very
       little  use. When set, any backslash in a pattern that is followed by a
       letter that has no special meaning  causes  an  error,  thus  reserving
       these  combinations  for  future  expansion.  By default, as in Perl, a
       backslash followed by a letter with no special meaning is treated as  a
       literal.  There  are  at  present  no other features controlled by this
       option. It can also be set by a (?X) option setting within a pattern.

       \section PCRE_MULTILINE

       By default, PCRE treats the subject string as consisting  of  a  single
       line  of characters (even if it actually contains newlines). The "start
       of line" metacharacter (^) matches only at the  start  of  the  string,
       while  the  "end  of line" metacharacter ($) matches only at the end of
       the string, or before a terminating newline (unless PCRE_DOLLAR_ENDONLY
       is set). This is the same as Perl.

       When  PCRE_MULTILINE  it  is set, the "start of line" and "end of line"
       constructs match immediately following or immediately before  any  new-
       line  in the subject string, respectively, as well as at the very start
       and end. This is equivalent to Perl's /m option, and it can be  changed
       within a pattern by a (?m) option setting. If there are no "\n" charac-
       ters in a subject string, or no occurrences of ^ or  $  in  a  pattern,
       setting PCRE_MULTILINE has no effect.

        \section PCRE_NO_AUTO_CAPTURE

       If this option is set, it disables the use of numbered capturing paren-
       theses in the pattern. Any opening parenthesis that is not followed  by
       ?  behaves as if it were followed by ?: but named parentheses can still
       be used for capturing (and they acquire  numbers  in  the  usual  way).
       There is no equivalent of this option in Perl.

        \section PCRE_UNGREEDY

       This  option  inverts  the "greediness" of the quantifiers so that they
       are not greedy by default, but become greedy if followed by "?". It  is
       not  compatible  with Perl. It can also be set by a (?U) option setting
       within the pattern.

        \section PCRE_UTF8

       This option causes PCRE to regard both the pattern and the  subject  as
       strings  of  UTF-8 characters instead of single-byte character strings.
       However, it is available only when PCRE is built to include UTF-8  sup-
       port.  If not, the use of this option provokes an error. Details of how
       this option changes the behaviour of PCRE are given in the  section  on
       UTF-8 support in the main pcre page.

        \section PCRE_NO_UTF8_CHECK

       When PCRE_UTF8 is set, the validity of the pattern as a UTF-8 string is
       automatically checked. If an invalid UTF-8 sequence of bytes is  found,
       pcre_compile()  returns an error. If you already know that your pattern
       is valid, and you want to skip this check for performance reasons,  you
       can  set  the  PCRE_NO_UTF8_CHECK option. When it is set, the effect of
       passing an invalid UTF-8 string as a pattern is undefined. It may cause
       your  program  to  crash.   Note that this option can also be passed to
       pcre_exec(),  to  suppress  the  UTF-8  validity  checking  of  subject
       strings.
*/

};

#endif
