/**************************************************************************
*   Copyright (C) 2004 by Vukasin Toroman                                 *
*   vukasin@toroman.org                                                   *
**************************************************************************/
/************************************************************************** 
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifndef WOOL_CONTENT_TYPE_H
#define WOOL_CONTENT_TYPE_H

#include "../base/wool_stl_includes.h"
#include "../base/cvs_info.h"
#include "../base/debug.h"

namespace wool
{
  /**
   * This class holds the information about the content type in form "type/subtype", e.g. "text/html".
   * @author Vukasin Toroman <vtoroman@astaro.com> (C) Astaro AG 2005
   */
  class content_type
  {
    public:
      /**
       * Construct a content_type with "unknown" type and subtype.
       */
      content_type()
      :first("unknown"),
      second("unknown")
      {}

      /**
       *  cvs info
       */
      VERSION_INFO_INLINE("$Id: content_type.h,v 1.4 2007/01/11 15:18:35 vtoroman Exp $");

      /**
       * Construct a  content_type with the given value. It separates the type and the subtype by locating the slash ("/"). If no slash was found only the type is known. In these case the subtype is set to "unknown".
       * @param str content type
       */
      explicit content_type(std::string str)
      :first("unknown"),
      second("unknown")
      {
        std::istringstream istr(str);
        load(istr);
      }
      void load (std::istream & str) {
        char c = 0;
        first.clear();
	str >> std::ws;
        while ((str.good())&&(c!='/')) {
          c= str.get();
          if(c!='/') {
            if( ((c<'A')||(c>'Z'))
              &&((c<'a')||(c>'z'))
              &&((c<'0')||(c>'9'))
              &&((c!='*')&&(c!='-')&&(c!='_')) )
              return;
            first.push_back(std::tolower(c));
          }
        }

        second.clear();
        while (str.good()) {
          c= str.get();
          if( ((c<'A')||(c>'Z'))
            &&((c<'a')||(c>'z'))
            &&((c<'0')||(c>'9'))
            &&((c!='*')&&(c!='-')&&(c!='_')) ) {
            WOOL_DEBUG("parsed: '"<<first<<"'/'"<<second<<"'");
            return;
          }
          second.push_back(std::tolower(c));
        }

      }

      /**
       * Assignment operator
       * @param ct content_type to assign
       */
      content_type & operator = (const content_type & ct)
      {
        first = ct.first;
	second = ct.second;
        WOOL_DEBUG("assigned: '"<<first<<"'/'"<<second<<"'");
        return *this;
      }

      /**
       * Copy constructor
       * @param ct content_type to copy
       */
      content_type(const content_type & ct)
      :first(ct.first),
      second(ct.second)
      {
      	toLower(first);
	toLower(second);
        WOOL_DEBUG("copied: '"<<first<<"'/'"<<second<<"'");
      }

      /**
       * How often contains this content_type the given content_type.
       * @param ct content_type to check if included
       * @return 0 if not contains or 1 if contains.
       */
      inline int count (const content_type & ct) const
      {
        return (((first==ct.first)||(first=="*"))
              &&((second==ct.second)||(second=="*")))?1:0;
      }

      /**
       * Lower operator
       * @param ct content_type to compare
       * @return true if this type is lower as type of ct else false
       */
      bool operator < (const content_type & ct) const { return first < ct.first;  }
      /**
       * Equal operator
       * @param ct content_type to compare
       * @return true if this type is equal as type of ct else false
       */
      bool operator == (const content_type & ct) const { return (first == ct.first) && (second == ct.second); }
      /**
       * Greater operator
       * @param ct content_type to compare
       * @return true if this type is greater as type of ct else false
       */
      bool operator > (const content_type & ct) const { return first > ct.first;  }
      /**
       * String operator
       * @return content_type as string
       */
      operator std::string () const { return first+"/"+second;  }
      /**
       *  Destructor
       */
      ~content_type() {}

      std::string first;

      std::string second;

private:

	void toLower (std::string & s)
	{
	  	for(unsigned int i = 0; i < s.length(); i++)
		{
			s[i] = std::tolower(s[i]);
		}
	}

  };

};

namespace std {
  inline std::istream & operator >> (std::istream & stream, wool::content_type & ct) {
      ct.load(stream);
      return stream;
  }

  inline std::ostream & operator << (std::ostream & stream, const wool::content_type & ct) {
      stream << ct.first << '/' << ct.second;
      return stream;
  }

};
#endif


