#ifndef HEADER_RFC2822_H_
#define HEADER_RFC2822_H_
namespace wool {
namespace rfc2822 {



template <typename _Header>
struct header_loader_template {
    //definitions
    typedef _Header header_type;
    typedef typename header_type::pointer pointer_type;
    typedef typename header_type::value_type value_type;
    typedef enum { linebegin, key, value, done } state_type;
    typedef enum { ws, separator, nl, normal, special } charclass;
    typedef char char_type;
	typedef wool::exception header_exception;

    //members
    header_type & header;
    pointer_type current_entry;
    state_type current_state;

    static const unsigned max_entry_length = 65536;
    static const unsigned max_header_count = 1024;

    header_loader_template(header_type & h):header(h),current_entry(NULL),current_state(linebegin) { }
    inline bool check_entry() {
        return ((!current_entry->first.empty()) && (!current_entry->second.empty()));
    }

    inline void finish_entry() {
        if(current_entry == NULL)
            return;
        if(check_entry()) {
            if((current_entry!=NULL) && (*current_entry->second.rbegin()==' '))
                current_entry->second.erase(current_entry->second.size()-1,1);
            current_entry = NULL;
        } else {
            current_entry->first.clear();
            current_entry->second.clear();
        }
    }

    inline void start_entry() {
        if(current_entry == NULL) {
            if(max_header_count<=header.size())
                throw header_exception( "too many header entries" );
            current_entry = & (*header.insert(header.end(), typename header_type::value_type()));
        }
    }

    inline void finish() {
        finish_entry();
        if(current_entry != NULL)
            header.pop_back();
        current_state = done;
    }

    inline charclass  getclass( char_type c) {
        switch(c) {
          case ' ':
          case '\t':
            return ws;
          case ':':
            return separator;
          case '\n':
            return nl;
          default:
            return ( (c>32) && (c<127) )?normal:special;
        }
    }

    inline void append_to_value( char_type c, charclass cc  ) {
        start_entry();
        if(current_entry->second.size()>max_entry_length)
            throw header_exception("header entry value too long");
        if( (cc!=ws) || ((!current_entry->second.empty()) && (getclass(*current_entry->second.rbegin())!=ws)))
            current_entry->second.insert(current_entry->second.end(), c);
    }

    inline void append_to_key( char_type c, charclass cc  ) {
        start_entry();
        if(current_entry->second.size()>max_entry_length)
            throw header_exception("header entry key too long");
        if(cc!=ws)
            current_entry->first.insert(current_entry->first.end(), c);
    }


    inline void parse_char( char_type c ) {
        charclass cc = getclass(c);
        switch ( current_state ) {
        case linebegin:
            switch (cc) {
            case ws:
                append_to_value(' ',cc);
                current_state = value;
                break;
            case separator:
                finish_entry();
                current_state = value;
                break;
            case normal:
                finish_entry();
                append_to_key(c,cc);
                current_state = key;
                break;
            case nl:
                finish();
                break;
            default:
                break;
            }
            break;
        case key:
            switch (cc) {
            case ws:
            case special:
                append_to_key(' ',ws); //noop
                break;
            case separator:
                current_state = value;
                break;
            case normal:
                append_to_key(c,cc);
                break;
            case nl:
                current_state = linebegin;
                break;
            }
            break;
        case value:
            switch (cc) {
            case ws:
            case special:
                append_to_value(' ',ws);
                break;
            case separator:
            case normal:
                append_to_value(c,cc);
                break;
            case nl:
                current_state = linebegin;
                break;
            }
            break;
        case done:
            break;
        }
    }

    template <typename InputIterator>
    inline InputIterator parse( InputIterator in, InputIterator end ) {
        while((current_state!=done)&&(in!=end)){
            parse_char(*in);
            in++;
        }
        return in;
    }
};
}
};
#endif /*HEADER_RFC2822_H_*/
