/**************************************************************************
*   Copyright (C) 2004 by Vukasin Toroman                                 *
*   vukasin@toroman.org                                                   *
**************************************************************************/
/************************************************************************** 
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/


#ifndef MESSAGE_BODY_H_
#define MESSAGE_BODY_H_

#include <string>
#include "../storage/swappable_data.h"

namespace wool {
  namespace rfc2822 {
  /**
   * body traits for std::string
   * @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
   */
  struct string_body_traits {
    /**
     *  body definition
     */
    typedef std::string body_type;
    
    /**
     *  size definition
     */
    typedef unsigned int size_type;

    /**
     *  Returns the size of the body.
     *  @return body size
     */
    inline size_type get_size( body_type & b ) {
    	return b.size();
    }
    
    /**
     *  Appends a chunk to the body
     *  @param b body
     *  @param start chunk start
     *  @param end chunk end
     */
    template <typename iteratorT>
    static void append( body_type & b, 
    					iteratorT start, 
    					iteratorT end ) {
		b.append(start,end);
	}
  };

  /**
   * body traits for wool::swappable_data
   * @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
   */
  struct sdata_body_traits {
    /**
     *  body definition
     */
    typedef wool::swappable_data body_type;
    
    /**
     *  size definition
     */
    typedef unsigned int size_type;

    
    /**
     *  Returns the size of the body.
     *  @return body size
     */
    inline size_type get_size( body_type & b ) {
    	return b.size();
    }
    
    /**
     *  Appends a chunk to the body
     *  @param b body
     *  @param start chunk start
     *  @param end chunk end
     */
    template <typename iteratorT>
    static void append( body_type & b, 
    					iteratorT start, 
    					iteratorT end ) {
		b.append(start,end);
	}
  };

	template <typename bodyT, typename traitsT>
	struct body_loader{
		bodyT & body;
		traitsT traits;
		long expected_size;
		
		body_loader( bodyT & b ) : body(b), expected_size(-1) {
		}

		

		void finish() {
			if(get_size()<expected_size)
				throw wool::exception("premature end of data while loading body");
		}
		
		template <typename iteratorT>
		iteratorT parse (iteratorT b, iteratorT e) {
			if(expected_size<0) {
				append(b,e);
				b = e;
			} else {
				while((get_size()<expected_size) && (b!=e))
					append(b,++b);
			}
			return b;
		}
		
		typename traitsT::size_type get_size() {
			return traits.get_size(body);
		}

		template <typename iteratorT>
		void append (iteratorT b, iteratorT e) {
			traits.append(body, b, e);
		}
	};

};
};
#endif /*MESSAGE_BODY_H_*/
