/**************************************************************************
*  Copyright (C) 2006 by Jens Finkhaeuser <unwesen@users.sourceforge.net> *
**************************************************************************/
/**************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
**************************************************************************/

// boost includes
#include <boost/bind.hpp>

#include <boost/spirit/core.hpp>
#include <boost/spirit/dynamic.hpp>

// own includes
#include "rfc3986/detail/uri_genlocimpl.h"

namespace sp = boost::spirit;

namespace wool {
namespace detail {

// ===========================================================================
// === class URI::ImplBase methods

GenericLocationURIImpl::GenericLocationURIImpl()
{
}



GenericLocationURIImpl::GenericLocationURIImpl(URI * uri, std::string const & uristring)
    : wool::URI::ImplBase(uri, uristring)
{
    parseUserinfo();
    parsePath();
}



GenericLocationURIImpl::~GenericLocationURIImpl()
{
}



boost::shared_ptr<wool::URI::ImplBase>
GenericLocationURIImpl::clone(URI * uri, std::string const & uristring) const
{
    return boost::shared_ptr<wool::URI::ImplBase>(
            new GenericLocationURIImpl(uri, uristring));
}



bool
GenericLocationURIImpl::isURL() const
{
    return true;
}



bool
GenericLocationURIImpl::isURN() const
{
    return false;
}



std::vector<std::string>
GenericLocationURIImpl::getPathParts() const
{
    return m_path_parts;
}



std::string
GenericLocationURIImpl::getUserName() const
{
    return m_username;
}



std::string
GenericLocationURIImpl::getPassword() const
{
    return m_password;
}



uint16_t
GenericLocationURIImpl::getPort() const
{
    if (m_port) {
        return m_port;
    }

    // if no port is set, we assume the default port for a number of known
    // schemes
    std::string scheme = m_uri->getScheme();
    if (scheme == "ftp") {
        return 21;
    } else if (scheme == "ssh") {
        return 22;
    } else if (scheme == "telnet") {
        return 23;
    } else if (scheme == "sftp") {
        return 115;
    }

    // return zero (m_port is zero here)
    return m_port;
}



void
GenericLocationURIImpl::parseUserinfo()
{
    // thanks to ImplBase's fairly complete parser, we know that userinfo
    // contains only valid characters at this point. we assume that the first
    // colon ':' separates the username from the password part of the userinfo.
    // once username and password are separate, percent-encoded characters can
    // be decoded again.
    std::string::size_type pos = m_userinfo.find(':');
    if (pos == std::numeric_limits<std::string::size_type>::max()) {
        // found no password part!
        m_username = normalize(m_userinfo.begin(), m_userinfo.end(), all);
    } else {
        // username and password!
        std::string tmp = m_userinfo.substr(0, pos);
        m_username = normalize(tmp.begin(), tmp.end(), all);

        tmp = m_userinfo.substr(pos + 1);
        m_password = normalize(tmp.begin(), tmp.end(), all);
    }
}



void
GenericLocationURIImpl::parsePath()
{
    // similar to parseUserinfo() above, we know that the path component only
    // contains valid characters at this point. we further assume that -
    // in accordance with the rfc - the path segments are divided by the slash
    // character '/'. the parser defined here can be very simple, therefore.

    // as the path component contains only valid characters, any char that is
    // not a slash is part of a segment.
    sp::rule<> pchar    = ~sp::ch_p('/');
    sp::rule<> segment  = *pchar;
    sp::rule<> path     = *sp::ch_p('/')
                          >> !(
                                segment[boost::bind(&GenericLocationURIImpl::addPathSegment,
                                    boost::ref(*this), _1, _2)]
                                >> *(
                                    +sp::ch_p('/')
                                    >> segment[boost::bind(&GenericLocationURIImpl::addPathSegment,
                                        boost::ref(*this), _1, _2)]
                                )
                          )
                          >> *sp::ch_p('/');

    if (!sp::parse(m_path.c_str(), path).full) {
        throw parse_error(parse_error::uri_internal_parsing_error,
            "Could not parse path component. Weird.");
    }
}




void
GenericLocationURIImpl::addPathSegment(char const * begin, char const * end)
{
    std::string tmp(begin, end);
    std::string tmp2 = normalize(tmp.begin(), tmp.end(), all);
    m_path_parts.push_back(tmp2);
}




bool
GenericLocationURIImpl::equals(boost::shared_ptr<URI::ImplBase> other) const
{
    if (!this->URI::ImplBase::equals(other)) {
        return false;
    }

    // these are the only components we can safely compare because they're
    // fully normalized if they exist.
    try {
        return (
                    (getUserName() == other->getUserName()) &&
                    (getPassword() == other->getPassword()) &&
                    (getPathParts() == other->getPathParts())
               );
    } catch (parse_error const & ex) {
        // can be thrown if other's uri scheme does not support parsing
        // of path parts, in which case the comparison returns false.
        return false;
    }
}



}} // namespaces
