/**************************************************************************
*  Copyright (C) 2006 by Jens Finkhaeuser <unwesen@users.sourceforge.net> *
**************************************************************************/
/**************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
**************************************************************************/

// boost includes
#include <boost/bind.hpp>

#include <boost/spirit/core.hpp>
#include <boost/spirit/dynamic.hpp>

// own includes
#include "rfc3986/detail/uri_httpimpl.h"

namespace sp = boost::spirit;

namespace wool {
namespace detail {

// ===========================================================================
// === class URI::ImplBase methods

HTTPURIImpl::HTTPURIImpl()
{
}



HTTPURIImpl::HTTPURIImpl(URI * uri, std::string const & uristring)
    : wool::detail::GenericLocationURIImpl(uri, uristring)
    , m_last_part(m_query_parts.end())
{
    parseQuery();

    // we also know at this point that the fragment is opaque, so we can normalize it
    // as well.
    std::string tmp = normalize(m_fragment.begin(), m_fragment.end(), all);
    m_fragment = tmp;
}



HTTPURIImpl::~HTTPURIImpl()
{
}



boost::shared_ptr<wool::URI::ImplBase>
HTTPURIImpl::clone(URI * uri, std::string const & uristring) const
{
    return boost::shared_ptr<wool::URI::ImplBase>(
            new HTTPURIImpl(uri, uristring));
}



std::multimap<std::string, std::string>
HTTPURIImpl::getQueryParts() const
{
    return m_query_parts;
}




uint16_t
HTTPURIImpl::getPort() const
{
    if (m_port) {
        return m_port;
    }

    // if no port is set, we assume the default port for a number of known
    // schemes
    std::string scheme = m_uri->getScheme();
    if (scheme == "webdav" || scheme == "http") {
        return 80;
    } else if (scheme == "webdavs" || scheme == "https") {
        return 443;
    }

    // return zero (m_port is zero here)
    return m_port;
}




void
HTTPURIImpl::parseQuery()
{
    // parsing the query isn't necessarily very difficult. we know that the query
    // string contains only valid characters. we further know that individual
    // parameters are subdivided by the ampersand character '&' and may contain
    // a value separated from the name by an equals character '='.

    sp::rule<>  path_delim  = sp::ch_p('&');
    sp::rule<>  kv_delim    = sp::ch_p('=');
    sp::rule<>  kv          = +( sp::anychar_p - sp::ch_p('&') - sp::ch_p('=') );
    sp::rule<>  param       = kv[boost::bind(&HTTPURIImpl::addPart, boost::ref(*this), _1, _2)]
                              >> !(
                                      kv_delim
                                      >> kv[boost::bind(&HTTPURIImpl::addPartValue, boost::ref(*this), _1, _2)]
                              );
    sp::rule<>  query       = !( param >> *( +path_delim >> param ) );

    if (!sp::parse(m_query.c_str(), query).full) {
        throw parse_error(parse_error::uri_internal_parsing_error,
                "Could not parse query component. Weird.");
    }
}




void
HTTPURIImpl::addPart(char const * begin, char const * end)
{
    std::string tmp(begin, end);
    std::string tmp2 = normalize(tmp.begin(), tmp.end(), all);
    m_last_part = m_query_parts.insert(
            std::pair<std::string, std::string>(tmp2, ""));
}



void
HTTPURIImpl::addPartValue(char const * begin, char const * end)
{
    std::string tmp(begin, end);
    std::string tmp2 = normalize(tmp.begin(), tmp.end(), all);
    if (m_last_part == m_query_parts.end()) {
        throw parse_error(parse_error::uri_internal_parsing_error,
                "Found query parameter value without key. Weird.");
    }
    m_last_part->second = tmp2;
}



bool
HTTPURIImpl::equals(boost::shared_ptr<URI::ImplBase> other) const
{
    if (!this->GenericLocationURIImpl::equals(other)) {
        return false;
    }

    // these are the only components we can safely compare because they're
    // fully normalized if they exist.
    try {
        return (
                    (getQueryParts() == other->getQueryParts()) &&
                    (getFragment() == other->getFragment())
               );
    } catch (parse_error const & ex) {
        // can be thrown if other's uri scheme does not support parsing
        // of query parts, in which case the comparison returns false.
        return false;
    }
}




}} // namespaces
