/**************************************************************************
*  Copyright (C) 2006 by Jens Finkhaeuser <unwesen@users.sourceforge.net> *
**************************************************************************/
/**************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
**************************************************************************/

// std::includes
#include <sstream>

// boost includes
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>

// XXX #define BOOST_SPIRIT_DEBUG 1

#include <boost/spirit/core.hpp>
#include <boost/spirit/dynamic.hpp>
#include <boost/spirit/error_handling/exceptions.hpp>

// own includes
#include "rfc3986/detail/uri_implbase.h"



namespace sp = boost::spirit;

namespace wool {

// ===========================================================================
// === class URI::ImplBase::Solicitor

/**
 * The following two macros are used to define accessors in the Solicitor class
 * below.
 **/
#define ALLOW_TYPE(member_name, type)                                                           \
    static inline void set_##member_name(URI::ImplBase * uri, std::string const & member_name)  \
    {                                                                                           \
        if (uri && !member_name.empty()) {                                                      \
            try {                                                                               \
                type tmp = boost::lexical_cast<type>(member_name);                              \
                uri->m_##member_name = tmp;                                                     \
            } catch (boost::bad_lexical_cast const & ex) {                                      \
                std::stringstream s;                                                            \
                s << "Could not cast '" << member_name << "' to type '"                         \
                        << typeid(type).name() << "'";                                          \
                throw parse_error(parse_error::uri_internal_parsing_error, s.str());            \
            }                                                                                   \
        }                                                                                       \
    }

#define ALLOW(member_name)                                                                             \
    static inline void set_##member_name(URI::ImplBase * uri, std::string const & member_name)  \
    {                                                                                           \
        if (uri) {                                                                              \
            uri->m_##member_name = member_name;                                                 \
        }                                                                                       \
    }


class URI::ImplBase::Solicitor
{
public:
    /**
     * Allow access to the following members of URI::ImplBase
     **/
    ALLOW(authority);
    ALLOW(userinfo);
    ALLOW_TYPE(port, uint32_t);
    ALLOW(path);
    ALLOW(query);
    ALLOW(fragment);

    /**
     * Granting access to the host component is slightly more complex, as the
     * we already know that when setting this component it should already be
     * fully normalized.
     **/
    static inline void set_host(URI::ImplBase * uri, std::string const & host)
    {
        if (uri) {
            uri->m_host = uri->normalize(host.begin(), host.end(), URI::ImplBase::all);
        }
    }

};

// ===========================================================================
// === URI Parsing
namespace {

/**
 * Parser rules are taken with as little adaption as possible directly from
 * RFC 3986 at http://www.ietf.org/rfc/rfc3986.txt
 *
 * The names of the rules have also been taken from that RFC, except that '-'
 * was converted to '_'.
 **/

/**
 * The set of unreserved characters is used in several places, and therefore
 * gets a definition all of its own. It differes from the other subrules in
 * that '_grammar' is appended to its name.
 **/
struct unreserved_grammar
    : public sp::grammar<unreserved_grammar>
{
    template <typename scannerT>
    struct definition
    {
        definition(unreserved_grammar const & self)
        {
            first = sp::as_lower_d[sp::alpha_p | sp::digit_p | '-' | '.' | '_' | '~'];
        }

        sp::rule<scannerT> const & start() const
        {
            return first;
        }

        sp::rule<scannerT>  first;
    };
};


/**
 * SETTER() defines a small wrapper for URI::ImplBase's members that
 * calls URI::ImplBase::Solicitor for setting a member. The double abstraction
 * is used to Solicitor does not have to know about string iterators
 * and can contain only static members.
 **/
#define SETTER(name)                                                    \
    inline void set_##name(char const * begin, char const * end) const  \
    {                                                                   \
        std::string tmp(begin, end);                                    \
        URI::ImplBase::Solicitor::set_##name(m_uri, tmp);               \
    }

/**
 * ACTION() is just defined to shorten the binding of setter methods
 * in the grammar definitions below.
 **/
#define ACTION(name)                                                    \
    boost::bind(&uri_grammar::set_##name, boost::ref(self), _1, _2)

/**
 * Full URI grammar definition, uses unreserved_grammar in some places
 * (see above).
 **/
struct uri_grammar
    : public sp::grammar<uri_grammar>
{
    // =======================================================================
    // === uri_grammar ctor/dtor
    uri_grammar(URI::ImplBase * uri)
        : m_uri(uri)
    {
    }


    // =======================================================================
    // === uri_grammar setter methods
    SETTER(authority);
    SETTER(host);
    SETTER(port);
    SETTER(path);
    SETTER(fragment);
    SETTER(query);

    /**
     * Setting the userinfo string requires a slightly different setter from
     * the ones above. Due to the exact time that spirit calls semantic actions
     * for a parse match, userinfo would contain the host information if
     * the authority rule (below) would be defined thus:
     *
     *   authority = !( userinfo[ACTION(userinfo)] >> '@' )
     *               >> host[ACTION(host)]
     *               >> !( ':' >> port[ACTION(port)] )
     *
     * That is, first the userinfo rule matches the characters for the host,
     * then the parser notices that it's the host part and that the userinfo
     * is missing from the URI. At that point, the userinfo setter has already
     * been called. Therefore we define authority thus:
     *
     *   authority = !( userinfo >> '@' )[ACTION(userinfo)]
     *               >> host[ACTION(host)]
     *               >> !( ':' >> port[ACTION(port)] )
     *
     * Here userinfo is only set if the URI contains such a string, but now it
     * will include the '@' sign at the end. That needs to be stripped away,
     * which is what this specialized setter function does.
     **/
    inline void set_userinfo(char const * begin, char const * end) const
    {
        if (*(end - 1) == '@') {
            if (end <= begin) {
                throw parse_error(parse_error::uri_internal_parsing_error,
                        "Invalid iterators when attempting to parse userinfo!");
            }

            // skip the '@' sign at the end, it's not part of the userinfo
            --end;
        }
        std::string tmp(begin, end);
        URI::ImplBase::Solicitor::set_userinfo(m_uri, tmp);
    }


    // =======================================================================
    // === struct definition
    /**
     * The definition subclass defines the actual grammar.
     **/
    template <typename scannerT>
    struct definition
    {
        // ===================================================================
        // === subrule identifiers
        /**
         * The grammar contains many subrules, each of which needs a unique
         * identifier. The identifier not only is used by boost::spirit to
         * keep the subrules apart, it's also reported to the error_handler
         * when one of the subrules fails to match.
         **/
        enum subrule_ids
        {
            pct_encoded_id      =  0,
            gen_delims_id       =  1,
            sub_delims_id       =  2,
            reserved_id         =  3,
            unreserved_id       =  4,
            dec_octet_id        =  5,
            pchar_id            =  6,
            userinfo_id         =  7,
            ipv4address_id      =  8,
            reg_name_id         =  9,
            host_id             = 10,
            port_id             = 11,
            authority_id        = 12,
            segment_nz_nc_id    = 13,
            segment_nz_id       = 14,
            segment_id          = 15,
            path_rootless_id    = 16,
            path_noscheme_id    = 17,
            path_absolute_id    = 18,
            path_abempty_id     = 19,
            path_empty_id       = 20,
            scheme_id           = 21,
            hier_part_id        = 22,
            query_id            = 23,
            fragment_id         = 24,
            uri_id              = 25,
        };


        // ===================================================================
        // === struct error_handler
        /**
         * We handle all errors by formatting the error nicely and then throwing
         * a parse_error.
         **/
        struct error_handler
        {
            template <typename errorT>
            sp::error_status<>
            operator()(scannerT const & scanner, errorT error) const
            {
                std::stringstream s;
                s << "Unexpected character (subrule #" << error.descriptor
                    << ") at input '" << error.where << "'!";
                throw parse_error(parse_error::uri_unexpected_character, s.str());

                return sp::error_status<>(sp::error_status<>::fail);
            }
        };


        // ===================================================================
        // === struct definition methods
        definition(uri_grammar const & self)
            : expect_scheme(scheme_id)
            , expect_hier_part(hier_part_id)
            , expect_query(query_id)
            , expect_fragment(fragment_id)
            , expect_authority(authority_id)
            , expect_path_abempty(path_abempty_id)
        {
            first =
            (
                /**
                 * URI definition
                 **/
                uri         = guard(
                                    expect_scheme(scheme)
                                    >> ':'
                                    >> expect_hier_part(hier_part)
                                    >> !( '?' >> expect_query(query)[ACTION(query)] )
                                    >> !( '#' >> expect_fragment(fragment)[ACTION(fragment)] )
                              )[error_handler()],

                /**
                 * Primary URI components
                 **/
                scheme      = sp::alpha_p
                              >> *(
                                  sp::alpha_p | sp::digit_p | '+' | '-' | '.'
                              ),

                hier_part   = (
                                    sp::str_p("//")
                                    >> expect_authority(authority)[ACTION(authority)]
                                    >> expect_path_abempty(path_abempty)[ACTION(path)]
                              )
                              | path_absolute[ACTION(path)]
                              | path_rootless[ACTION(path)]
                              | path_empty,

                query       = *( pchar | '/' | '?' ),

                fragment    = *( pchar | '/' | '?' ),

                /**
                 * Authority components
                 **/
                userinfo    = *( unreserved | pct_encoded | sub_delims | ':' ),

                ipv4address = dec_octet >> '.' >> dec_octet >> '.' >> dec_octet
                              >> '.' >> dec_octet,

                reg_name    = *( unreserved | pct_encoded | sub_delims ),

                host        = ipv4address | reg_name,

                port        = !( sp::limit_d(1, 65535)[dec_max_5_digit] ),

                authority   = !( userinfo >> '@' )[ACTION(userinfo)]
                              >> host[ACTION(host)]
                              >> !( ':' >> port[ACTION(port)] ),

                /**
                 * Path components
                 **/
                segment_nz_nc   = +( unreserved | pct_encoded | sub_delims | '@' ),

                segment_nz      = +pchar,

                segment         = *pchar,

                path_abempty    = *( '/' >> segment ),

                path_absolute   = sp::ch_p('/')
                                  >> !(
                                        segment_nz >> *( '/' >> segment )
                                  ),

                path_noscheme   = segment_nz_nc >> *( '/' >> segment ),

                path_rootless   = segment_nz >> *( '/' >> segment ),

                path_empty      = sp::epsilon_p, // no characters

                /**
                 * Character sets
                 **/
                pct_encoded = sp::ch_p('%') >> sp::xdigit_p >> sp::xdigit_p,

                gen_delims  = sp::ch_p(':') | '/' | '?' | '#' | '[' | ']' | '@',

                sub_delims  = sp::ch_p('!') | '$' | '&' | '\'' | '(' | ')' | '*' | '+' | ',' | ';' | '=',

                reserved    = gen_delims | sub_delims,

                dec_octet   = sp::limit_d(0, 255)[dec_max_3_digit],

                pchar       = unreserved | pct_encoded | sub_delims | ':' | '@'
            );
        }

        sp::rule<scannerT> const & start() const
        {
            return first;
        }


        // ===================================================================
        // === struct definition members
        /**
         * Subrule declarations. You can ignore these to a large part, focus on
         * the definitions in the ctor above instead.
         **/
        sp::subrule<pct_encoded_id>     pct_encoded;
        sp::subrule<gen_delims_id>      gen_delims;
        sp::subrule<sub_delims_id>      sub_delims;
        sp::subrule<reserved_id>        reserved;
        sp::subrule<dec_octet_id>       dec_octet;
        sp::subrule<pchar_id>           pchar;

        sp::subrule<scheme_id>          scheme;

        sp::subrule<userinfo_id>        userinfo;
        sp::subrule<ipv4address_id>     ipv4address;
        sp::subrule<reg_name_id>        reg_name;
        sp::subrule<host_id>            host;
        sp::subrule<port_id>            port;
        sp::subrule<authority_id>       authority;

        sp::subrule<segment_nz_nc_id>   segment_nz_nc;
        sp::subrule<segment_nz_id>      segment_nz;
        sp::subrule<segment_id>         segment;
        sp::subrule<path_rootless_id>   path_rootless;
        sp::subrule<path_noscheme_id>   path_noscheme;
        sp::subrule<path_absolute_id>   path_absolute;
        sp::subrule<path_abempty_id>    path_abempty;
        sp::subrule<path_empty_id>      path_empty;
        sp::subrule<hier_part_id>       hier_part;

        sp::subrule<query_id>           query;
        sp::subrule<fragment_id>        fragment;

        sp::subrule<uri_id>             uri;

        /**
         * Other parsers used besides the above subrules
         **/
        unreserved_grammar              unreserved;

        sp::int_parser<int, 10, 1, 3>   dec_max_3_digit;
        sp::int_parser<int, 10, 1, 5>   dec_max_5_digit;


        sp::rule<scannerT>  first;

        /**
         * Assertion parser declarations; used to assert each mandatory subrule
         **/
        sp::assertion<subrule_ids>  expect_scheme;
        sp::assertion<subrule_ids>  expect_hier_part;
        sp::assertion<subrule_ids>  expect_query;
        sp::assertion<subrule_ids>  expect_fragment;
        sp::assertion<subrule_ids>  expect_authority;
        sp::assertion<subrule_ids>  expect_path_abempty;

        sp::guard<subrule_ids>      guard;
    };

    // =======================================================================
    // === struct uri_grammar members
    //! Pointer to URI implementation
    URI::ImplBase * m_uri;
};

} // anonymous namespace



// ===========================================================================
// === Helper functions

bool check_uri_impl(std::string const & uri, bool throw_on_error /* = false */)
{
    try {
        uri_grammar grammar(0);
        return sp::parse(uri.c_str(), grammar).full;
    } catch (parse_error const & ex) {
        if (throw_on_error) {
            throw;
        } else {
            return false;
        }
    }
}




// ===========================================================================
// === class URI::ImplBase methods

URI::ImplBase::ImplBase()
{
}



URI::ImplBase::ImplBase(URI * uri, std::string const & uristring)
    : m_uri(uri)
    , m_port(0)
{
    m_normalized_string = normalize(uristring.begin(), uristring.end());

    uri_grammar grammar(this);
    sp::parse_info<> info = sp::parse(m_normalized_string.c_str(), grammar);
    if (!info.full) {
        throw parse_error(parse_error::uri_internal_parsing_error,
                "URI could not be fully parsed, yet no error was reported. Weird.");
    }
}




URI::ImplBase::~ImplBase()
{
}



std::string
URI::ImplBase::getAuthority() const
{
    return m_authority;
}



std::string
URI::ImplBase::getUserInfo() const
{
    return m_userinfo;
}



std::string
URI::ImplBase::getHost() const
{
    return m_host;
}



uint16_t
URI::ImplBase::getPort() const
{
    return m_port;
}



std::string
URI::ImplBase::getPath() const
{
    return m_path;
}




std::string
URI::ImplBase::getQuery() const
{
    return m_query;
}




std::string
URI::ImplBase::getFragment() const
{
    return m_fragment;
}



std::vector<std::string>
URI::ImplBase::getPathParts() const
{
    throw parse_error(parse_error::uri_not_implemented_by_scheme,
            "This URI scheme does not support parsing of the path component!");
    return std::vector<std::string>();
}



std::multimap<std::string, std::string>
URI::ImplBase::getQueryParts() const
{
    throw parse_error(parse_error::uri_not_implemented_by_scheme,
            "This URI scheme does not support parsing of the query component!");
    return std::multimap<std::string, std::string>();
}



std::string
URI::ImplBase::getUserName() const
{
    throw parse_error(parse_error::uri_not_implemented_by_scheme,
            "This URI scheme does not support parsing of the userinfo component!");
    return std::string();
}



std::string
URI::ImplBase::getPassword() const
{
    throw parse_error(parse_error::uri_not_implemented_by_scheme,
            "This URI scheme does not support parsing of the userinfo component!");
    return std::string();
}



std::string
URI::ImplBase::getNormalized() const
{
    return m_normalized_string;
}



bool
URI::ImplBase::isURL() const
{
    throw parse_error(parse_error::uri_not_implemented_by_scheme,
            "The base parser does not know whether this URI is also a URL!");
    return false;
}



bool
URI::ImplBase::isURN() const
{
    throw parse_error(parse_error::uri_not_implemented_by_scheme,
            "The base parser does not know whether this URI is also a URN!");
    return false;
}



boost::shared_ptr<URI::ImplBase>
URI::ImplBase::clone(URI * uri, std::string const & uristring) const
{
    throw parse_error(parse_error::uri_not_implemented_by_scheme,
            "The base parser cannot be cloned!");
    return boost::shared_ptr<URI::ImplBase>();
}



std::string::value_type
URI::ImplBase::hexDecode(std::string::value_type c) const
{
    std::string::value_type ret = c;
    if (ret >= 'a' && ret <= 'f') {
        ret = (ret - 'a') + 10;
    } else if (ret >= 'A' && ret <= 'F') {
        ret = (ret - 'A') + 10;
    } else {
        ret = ret - '0';
    }
    return ret;
}



std::string
URI::ImplBase::normalize(std::string::const_iterator begin,
        std::string::const_iterator end,
        normalize_type chars_to_normalize /* = unreserved */) const
{
    std::string retval;

    std::string::const_iterator iter = begin;
    for ( ; iter != end ; ++iter) {
        if (*iter != '%') {
            if (chars_to_normalize == all && *iter == '+') {
                retval += ' ';
            } else {
                retval += *iter;
            }
            continue;
        }

        // decode percent-encoded characters, if possible
        std::string::const_iterator tmp     = iter;
        std::string::value_type first_raw   = *(++tmp);
        std::string::value_type first       = hexDecode(first_raw);
        std::string::value_type second_raw  = *(++tmp);
        std::string::value_type second      = hexDecode(second_raw);

        if ((first >= 0 && first <= 15) && (second >= 0 && second <= 15)) {
            std::string::value_type decoded = (first * 16) + second;

            if (chars_to_normalize == unreserved) {
                // only normalize chars that are within the unreserved ranges
                std::string::value_type buf[2];
                buf[0] = decoded;
                buf[1] = 0;
                unreserved_grammar grammar;
                sp::parse_info<> info = sp::parse(buf, grammar);
                if (!info.full) {
                    // the decoded char is not in the unreserved range
                    retval += '%';
                    retval += first_raw;
                    retval += second_raw;
                    iter = tmp;
                    continue;
                }
            }

            // add decoded char to retval and skip to the following character
            // for processing
            retval += decoded;
            iter = tmp;

        } else {
            std::stringstream s;
            s << "Could not decode percent-encoded character '%" <<
                first_raw << second_raw << "'!";
            throw parse_error(parse_error::uri_invalid_percent_encoding, s.str());
        }
    }

    return retval;
}




bool
URI::ImplBase::equals(boost::shared_ptr<URI::ImplBase> other) const
{
    if (!other) {
        return false;
    }

    // these are the only components we can safely compare because they're
    // fully normalized if they exist.
    return (
                (getHost() == other->getHost()) &&
                (getPort() == other->getPort())
           );
}

} // namespaces
