/**************************************************************************
*  Copyright (C) 2006 by Jens Finkhaeuser <unwesen@users.sourceforge.net> *
**************************************************************************/
/**************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
**************************************************************************/

#ifndef WOOL_URI_DETAIL_IMPLBASE_H
#define WOOL_URI_DETAIL_IMPLBASE_H

// std includes
#include <vector>
#include <map>

// own includes
#include "rfc3986/uri.h"

namespace wool {

// ===========================================================================
// === Helper functions
/**
 * Simple helper function for checking an URI syntax.
 **/
bool check_uri_impl(std::string const & uri, bool throw_on_error = false);



// ===========================================================================
// === class URI::ImplBase
/**
 * The URI::ImplBase class defines those parsing rules common to all well-formed
 * URIs as defined by RFC 3986 (http://www.ietf.org/rfc/rfc3986.txt), with the
 * exception of the "IP-literal" rule defined in the document (which includes
 * IPv6 addresses and raw IPv4 addresses).
 *
 * Implementations of a specific scheme must derive from this class, and register
 * an object for their respective scheme with the URI class (internal to the
 * URI class, see uri.cpp).
 **/
class URI::ImplBase
{
public:
    // =======================================================================
    // === public ctor/dtor
    ImplBase(URI * uri, std::string const & uristring);
    virtual ~ImplBase();

    // =======================================================================
    // === public methods
    /**
     * The clone() method is used by the URI class to construct a fully
     * initialized implementation object from a default-constructed prototype.
     * Scheme implementations need to implement it somewhat like this:
     *
     *     return boost::shared_ptr<ImplBase>(new MyImpl(uri, uristring));
     **/
    virtual boost::shared_ptr<ImplBase>
    clone(URI * uri, std::string const & uristring) const;

    /**
     * The following getters return specific URI parts common to all URIs.
     * Not all of these URIs need to be set. Note that all of these members
     * are virtual so Scheme implementations may override them. For example,
     * a HTTP implementation might implement getPort() to return the number
     * 80 if no port has been set in the URI string.
     **/
    virtual std::string getAuthority() const;
    virtual std::string getUserInfo() const;
    virtual std::string getHost() const;
    virtual uint16_t    getPort() const;

    virtual std::string getPath() const;
    virtual std::string getQuery() const;
    virtual std::string getFragment() const;

    /**
     * These two getters are similar to the one above, but different in
     * that they have to be defined by the Scheme implementation. The defaults
     * provided here throw a parse_error.
     **/
    virtual std::vector<std::string> getPathParts() const;
    virtual std::multimap<std::string, std::string> getQueryParts() const;
    virtual std::string getUserName() const;
    virtual std::string getPassword() const;

    /**
     * The following two methods are implemented as getPathParts() etc.
     * above, but return true if the URI is also an URL or URN respectively.
     * Only Scheme implementations can know that, however, so these methods
     * must be overridden.
     **/
    virtual bool isURL() const;
    virtual bool isURN() const;

    /**
     * getNormalized() returns a normalized version of the URI that has those
     * percent-encoded characters decoded that the RFC defines as safe to
     * decode before parsing of the URI components is attempted.
     **/
    std::string getNormalized() const;


    /**
     * Compares the other URI implementation to our own state. Derived
     * classes need to overload this method and may benefit from calling
     * the base type's method for some of the comparison.
     **/
    virtual bool equals(boost::shared_ptr<ImplBase> other) const;

    // =======================================================================
    // === class Solicitor
    /**
     * The solicitor class provides access to the protected members to
     * the (internally encapsulated) parsing grammar.
     **/
    class Solicitor;
    friend class Solicitor;

protected:
    // =======================================================================
    // === protected types/typedefs
    /**
     * You can either decode all percent-encoded characters ore only those
     * from the 'unreserved' character set as defined by the RFC. The latter
     * are safe to decode _before_ parsing URI components, the former might
     * confuse the URI parser if they are decoded before parsing.
     **/
    enum normalize_type
    {
        unreserved  = 0,
        all         = 1
    };


    // =======================================================================
    // === protected ctor
    /**
     * The protected ctor may only be called by derived classes.
     **/
    ImplBase();

    // =======================================================================
    // === protected methods
    /**
     * Convert a hexadecimal _digit_ into a decimal value.
     **/
    inline std::string::value_type hexDecode(std::string::value_type c) const;

    /**
     * Normalize the character range described by the begin and end iterators
     * by decoding percent-encoded characters of the given type.
     **/
    std::string normalize(std::string::const_iterator begin,
            std::string::const_iterator end,
            normalize_type chars_to_normalize = unreserved) const;

    // =======================================================================
    // === protected members
    /** pointer to enclosing URI object. **/
    URI *       m_uri;

    /**
     * normalized URI string, with all unreserved percent-encoded characters
     * decoded.
     **/
    std::string m_normalized_string;

    /** members holding URI components of the same name. **/
    std::string m_authority;
    std::string m_userinfo;
    std::string m_host;
    uint16_t    m_port;
    std::string m_path;
    std::string m_query;
    std::string m_fragment;
};

} // namespaces

#endif // guard
