/**************************************************************************
*  Copyright (C) 2006 by Jens Finkhaeuser <unwesen@users.sourceforge.net> *
**************************************************************************/
/**************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
**************************************************************************/

// std includes
#include <sstream>
#include <limits>
#include <map>

// boost includes
#include <boost/algorithm/string/case_conv.hpp>

// own includes
#include "../rfc3986/uri.h"
#include "../rfc3986/detail/uri_implbase.h"

#include "../rfc3986/detail/uri_genlocimpl.h"
#include "../rfc3986/detail/uri_httpimpl.h"

namespace wool {


// ===========================================================================
// === class parse_error methods

parse_error::parse_error(error_code_type error_code, std::string const & message)
    : exception(message)
    , m_code(error_code)
{
}


parse_error::error_code_type
parse_error::code() const throw()
{
    return m_code;
}

// ===========================================================================
// === class static data

// prefixed with the rfc number, from 00 to 99
parse_error::error_code_type const parse_error::uri_no_scheme                   = 398600;
parse_error::error_code_type const parse_error::uri_empty                       = 398601;
parse_error::error_code_type const parse_error::uri_unknown_scheme              = 398602;
parse_error::error_code_type const parse_error::uri_invalid_percent_encoding    = 398603;
parse_error::error_code_type const parse_error::uri_not_url                     = 398604;
parse_error::error_code_type const parse_error::uri_internal_parsing_error      = 398605;
parse_error::error_code_type const parse_error::uri_unexpected_character        = 398606;
parse_error::error_code_type const parse_error::uri_not_implemented_by_scheme   = 398607;



// ===========================================================================
// === Helper code

namespace {

/**
 * This is a miniature extension to std::set that is used instead of std::map
 * for the known scheme list because C++ cannot yet properly initialize static
 * STL containers.
 **/
struct known_schemes_type
    : public std::map<std::string, boost::shared_ptr<URI::ImplBase> >
{
public:
    known_schemes_type()
        : std::map<std::string, boost::shared_ptr<URI::ImplBase> >()
    {
        // these schemes all use a HTTP-like syntax
        (*this)["http"]     = boost::shared_ptr<URI::ImplBase>(new detail::HTTPURIImpl());
        (*this)["https"]    = boost::shared_ptr<URI::ImplBase>(new detail::HTTPURIImpl());
        (*this)["webdav"]   = boost::shared_ptr<URI::ImplBase>(new detail::HTTPURIImpl());
        (*this)["webdavs"]  = boost::shared_ptr<URI::ImplBase>(new detail::HTTPURIImpl());

        // the following (known) schemes all use a syntax like
        // scheme://username:passwd@host/path/to/resource
        // i.e. like HTTP without the query or fragment
        (*this)["ftp"]      = boost::shared_ptr<URI::ImplBase>(new detail::GenericLocationURIImpl());
        (*this)["ssh"]      = boost::shared_ptr<URI::ImplBase>(new detail::GenericLocationURIImpl());
        (*this)["sftp"]     = boost::shared_ptr<URI::ImplBase>(new detail::GenericLocationURIImpl());
        (*this)["telnet"]   = boost::shared_ptr<URI::ImplBase>(new detail::GenericLocationURIImpl());
    }
};


static known_schemes_type const sg_known_schemes;

} // anonymous namespace




bool check_uri(std::string const & uri, bool throw_on_error /* = false */)
{
    return check_uri_impl(uri, throw_on_error);
}




// ===========================================================================
// === class URI methods

URI::URI(std::string const & uristring, std::string const & default_scheme /* = "" */,
        bool has_authority /* = false */)
    : m_original_string(uristring)
    , m_default_scheme(default_scheme)
    , m_has_authority(has_authority)
{
    init();
}




URI::URI(URI const & other)
    : m_original_string(other.m_original_string)
    , m_scheme(other.m_scheme)
    , m_default_scheme(other.m_default_scheme)
    , m_has_authority(other.m_has_authority)
{
    // XXX currently there is no way to modify a URI object once it's constructed.
    // so all copies of that object may just as well share the internal representation.
    // m_impl = other.m_impl->clone(this, m_original_string);
    m_impl = other.m_impl;
}




URI &
URI::operator=(std::string const & uristring)
{
    m_original_string = uristring;
    init();

    return *this;
}




URI &
URI::operator=(URI const & other)
{
    m_original_string = other.m_original_string;
    m_scheme = other.m_scheme;
    m_default_scheme = other.m_default_scheme;
    m_has_authority = other.m_has_authority;

    // XXX currently there is no way to modify a URI object once it's constructed.
    // so all copies of that object may just as well share the internal representation.
    // m_impl = other.m_impl->clone(this, m_original_string);
    m_impl = other.m_impl;

    return *this;
}




bool
URI::operator==(std::string const & uristring) const
{
    URI tmp_uri(uristring);
    return operator==(tmp_uri);
}




bool
URI::operator==(URI const & other) const
{
    return m_impl->equals(other.m_impl);
}




bool
URI::operator!=(std::string const & uristring) const
{
    return !operator==(uristring);
}




bool
URI::operator!=(URI const & other) const
{
    return !operator==(other);
}




void
URI::init()
{
    std::string uristring = m_original_string;

    if (uristring.empty()) {
        throw parse_error(parse_error::uri_empty, "Empty URI-string!");
    }

    std::string scheme_delimiter = ":";
    if (m_has_authority) {
        scheme_delimiter = "://";
    }

    std::string::size_type pos = m_original_string.find(scheme_delimiter);
    if (pos == std::numeric_limits<std::string::size_type>::max()) {
        if (m_default_scheme.empty()) {
            // if no default scheme is provided, we can't do anything useful.
            std::stringstream s;
            s << "Could not find scheme in URI-string '" << m_original_string << "'!";
            throw parse_error(parse_error::uri_no_scheme, s.str());
        } else {
            // prepend the default scheme, and try again.
            m_original_string = m_default_scheme + scheme_delimiter + m_original_string;
            init();
            return;
        }
    } else if (pos == (m_original_string.size() - 3)) {
        std::stringstream s;
        s << "Empty URI-string consists only of scheme part: " << m_original_string;
        throw parse_error(parse_error::uri_empty, s.str());
    }

    // remember scheme separately
    m_scheme = m_original_string.substr(0, pos);
    boost::to_lower(m_scheme);

    // find out if we support the scheme
    known_schemes_type::const_iterator scheme_iter = sg_known_schemes.find(m_scheme);
    if (scheme_iter == sg_known_schemes.end()) {
        // the base parser should be able to handle most schemes, so we'll instanciate
        // it as the implementation. note that this is just a fallback that might well
        // be inappropriate for your scheme, in which case you just have to implement
        // a scheme implementation.
        m_impl = boost::shared_ptr<ImplBase>(new ImplBase(this, uristring));
    } else {
        // now that we know we support the scheme, we need to instanciate a proper
        // implementation
        m_impl = scheme_iter->second->clone(this, uristring);
    }
}


std::string
URI::getOriginal() const
{
    return m_original_string;
}



std::string
URI::getScheme() const
{
    return m_scheme;
}



std::string
URI::getAuthority() const
{
    return m_impl->getAuthority();
}



std::string
URI::getUserInfo() const
{
    return m_impl->getUserInfo();
}



std::string
URI::getHost() const
{
    return m_impl->getHost();
}



uint16_t
URI::getPort() const
{
    return m_impl->getPort();
}



std::string
URI::getPath() const
{
    return m_impl->getPath();
}



std::string
URI::getFragment() const
{
    return m_impl->getFragment();
}



std::string
URI::getQuery() const
{
    return m_impl->getQuery();
}



std::string
URI::getNormalized() const
{
    return m_impl->getNormalized();
}




bool
URI::isURL() const
{
    return m_impl->isURL();
}



bool
URI::isURN() const
{
    return m_impl->isURN();
}



std::string
URI::getUserName() const
{
    return m_impl->getUserName();
}



std::string
URI::getPassword() const
{
    return m_impl->getPassword();
}



std::vector<std::string>
URI::getPathParts() const
{
    return m_impl->getPathParts();
}




std::multimap<std::string, std::string>
URI::getQueryParts() const
{
    return m_impl->getQueryParts();
}


} // namespaces
