/**************************************************************************
*  Copyright (C) 2006 by Jens Finkhaeuser <unwesen@users.sourceforge.net> *
**************************************************************************/
/**************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
**************************************************************************/

#ifndef WOOL_URI_H
#define WOOL_URI_H

// c includes
#include <stdint.h>

// std includes
#include <string>
#include <vector>
#include <map>

// boost includes
#include <boost/shared_ptr.hpp>

// own includes
#include "../base/exception.h"
#include "../base/cvs_info.h"

namespace wool {

// ===========================================================================
// === free functions

/**
 * Simple helper function for checking an URI syntax.
 * @return true if the URI has correct syntax, else false
 **/
bool check_uri(std::string const & uri, bool throw_on_error = false);

// ===========================================================================
// === class parse_error
/**
 * Subclass of wool::exception for parsing errors, either when parsing URIs
 * or RPC requests/responses.
 **/
class parse_error
    : public exception
{
public:
    typedef uint32_t    error_code_type;

    parse_error(error_code_type error_code, std::string const & message);

    /**
     * Retrieve the error code with which the exception was instanciated.
     **/
    error_code_type code() const throw();

    static error_code_type const uri_no_scheme;
    static error_code_type const uri_empty;
    static error_code_type const uri_unknown_scheme;
    static error_code_type const uri_invalid_percent_encoding;
    static error_code_type const uri_not_url;
    static error_code_type const uri_internal_parsing_error;
    static error_code_type const uri_unexpected_character;
    static error_code_type const uri_not_implemented_by_scheme;

private:
    error_code_type m_code;
};


// ===========================================================================
// === class URI
/**
 * The URI class attempts to encapsulate parsing and validation of all URI
 * schemes.
 *
 * The class supports the generic URI syntax as described in *RFC 3986
 * (http://www.ietf.org/rfc/rfc3986.txt), except for the IP-literal subrule
 * used, among other things, to define IPv6 addresses.
 *
 * It further supports (at this time) HTTP/S syntax, that is parsing of
 * path and query components, and username/password splitting.
 **/
class URI
{
public:
    /**
     *  cvs info
     */
    VERSION_INFO_INLINE("$Id: uri.h,v 1.4 2007/01/11 15:18:38 vtoroman Exp $");

    // =======================================================================
    // === semi-public types
    /**
     * The ImplBase class is the base class for all URI Scheme implementation
     * classes. It is public so that implementation classes may be added
     * freely. However the definition of this class is hidden in a header that
     * is not installed with libwool, making it essentially unusable to users.
     **/
    class ImplBase;

    // =======================================================================
    // === public ctor/dtor
    /**
     * URI instances may be initialized with a string or another URI.
     *
     * If the URI is constructed from a std::string not containing a scheme
     * part, and the default_scheme parameter is provided, the default_scheme
     * is prepended to the uristring, and parsing is tried again.
     *
     * Unfortunately it is not easy to distinguish between a valid scheme
     * and a valid hostname. A URI string such as "first:second" is a valid
     * URI, where "first" is (correctly) detected as a scheme, and "second"
     * is (correctly) detected as the hierarchy part of the URI.
     *
     * In a lot of cases, you have a URI string of the form "hostname:port",
     * which would be parsed differently from what you might expect. If that
     * is the case, you may pass a third boolean parameter (set to true),
     * which indicates to the constructor that you _know_ that your URI
     * string contains an authority. If a URI has an authority, the ":" that
     * delimits the scheme from the hierarchy part is followed by double
     * slashes ("//"). We can therefore safely look for the sequence "://" to
     * determine which part of the URI string is the scheme, and which
     * the hierarchy part.
     **/
    URI(std::string const & uristring,
            std::string const & default_scheme = std::string(),
            bool has_authority = false);

    /** copy ctor **/
    URI(URI const & other);

    // =======================================================================
    // === public operators
    /**
     * Assignment again works from strings or other URIs. Note that assignment
     * always implies parsing of the URI, is therefore moderatly expensive
     * and should not be done excessively often.
     **/
    URI & operator=(std::string const & uristring);
    URI & operator=(URI const & other);

    /**
     * Comparison again works with std::strings and URIs. Note that parsed URIs
     * are compared internally, which causes URIs to compare as equal that
     * differ in which unreserved characters are percent-encoded.
     **/
    bool operator==(std::string const & uristring) const;
    bool operator==(URI const & other) const;

    bool operator!=(std::string const & uristring) const;
    bool operator!=(URI const & other) const;

    // =======================================================================
    // === public methods
    /**
     * The following methods returns the original unprocessed URI string and
     * a partially normalized version, respectively. That is, characters from
     * the unreserved character set that appeared in percent-encoded form are
     * decoded.
     **/
    std::string getOriginal() const;
    std::string getNormalized() const;

    /**
     * The following accessors are common for all URI implementations and return
     * the major URI components. Note that the result value may differ slightly
     * for different URI schemes, e.g. the HTTP URI scheme will return 80
     * for getPort() if no port is specified, etc.
     **/
    std::string getScheme() const;
    std::string getAuthority() const;
    std::string getUserInfo() const;
    std::string getHost() const;
    uint16_t getPort() const;
    std::string getPath() const;
    std::string getQuery() const;
    std::string getFragment() const;

    /**
     * The following methods are only implemented by some URI schemes and may
     * throw parse_error exceptions when called. However, you can always use
     * one of the above methods to find out in advance whether it's necessary
     * to call one of these accessors or not. For example, if getUserInfo()
     * returns an empty string, getUserName() and getPassword() will as well.
     **/
    std::vector<std::string> getPathParts() const;
    std::multimap<std::string, std::string> getQueryParts() const;
    std::string getUserName() const;
    std::string getPassword() const;

    /**
     * Implemented by specific URI scheme implementations, these methods
     * return a boolean value determining if a given URI is also a URL or URN.
     * XXX Note that not all scheme implementations may implement these
     * methods, so don't rely too heavily on them.
     **/
    bool isURL() const;
    bool isURN() const;

private:
    // =======================================================================
    // === private methods
    /**
     * Initialize the URI by parsing the uristring, etc.
     **/
    void init();

    // =======================================================================
    // === private data
    /** pointer to the internal implementation object. **/
    boost::shared_ptr<ImplBase>  m_impl;

    /**
     * @{
     *
     * The URI base class knows the original string passed to the URI and the
     * scheme and nothing else.
     **/
    std::string m_original_string;
    std::string m_scheme;

    /**
     * The default scheme may or may not be provided in the constructor.
     **/
    std::string m_default_scheme;

    /**
     * Set to true if the caller knowns that the original URI string contains
     * an authority.
     **/
    bool        m_has_authority;
    /** @} **/
};


} // namespaces

#endif // guard
