/**************************************************************************
*   Copyright (C) 2004 by Vukasin Toroman                                 *
*   vukasin@toroman.org                                                   *
**************************************************************************/
/**************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/


#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/poll.h>

#include "socket/socket.h"
#include "base/debug.h"

namespace wool {
      int socket::fcntl(int cmd) { return ::fcntl(fd(),cmd); }
      int socket::fcntl(int cmd, long arg) { return ::fcntl(fd(),cmd,arg); }
      int socket::fcntl(int cmd, struct flock *lock) { return ::fcntl(fd(), cmd, lock);}

      socket::flagset socket::flags () const { return ::fcntl(m_fd,F_GETFL); }
      void socket::flags (socket::flagset f) {  ::fcntl(m_fd,F_SETFL,f); }

      bool socket::test_flag (socket::flag f) const { return ((flags()&f)==f);} ;
      void socket::set_flag (socket::flag f) { flags(flags()|f);} ;
      void socket::reset_flag (socket::flag f) { flags(flags()&(~f));} ;

  socket::socket ( const socket & s ) : m_peer_reset(false){
	m_timeout = s.m_timeout;
      	m_fd = ::fcntl(s.m_fd, F_DUPFD);
      }
      
		address::address( const address::family & f ) { 
			sa_family =f ;
		}
		
		address::address( const ::sockaddr & f ) {
			::memmove(&sa_family, &f, sizeof(sockaddr));
		}

		void address::set_family( const family & f ) {
			sa_family = f;
		}
		
		const address::family & address::get_family() const {
			return sa_family;
		}
		void * address::get_data() {
			return sa_data;
		}
		const void * address::get_data() const {
			return sa_data;
		}

//      domain get_domain() const;
//      type get_type() const;
      
      socket::socket ( const socket::type_record & t ) 
	:m_fd(::socket (t.domain, t.type, t.protocol))
	,m_peer_reset(false){
	m_timeout = -1;
        if(m_fd<0)
            throw wool::socket::exception();
      }
      
      socket::socket (socket::domain d , socket::type t , socket::protocol p) 
	:m_fd(::socket (d, t, p)) 
	,m_peer_reset(false){
	m_timeout = -1;
        if(m_fd<0)
            throw wool::socket::exception();
      }

      socket::~socket() {
        close();
        assert(m_fd < 0);
      }

      bool socket::operator ! (  ) const {
        return (m_fd < 0);
      }

      socket & socket::operator = ( const socket & s  ) {
        this->fd(::fcntl(s.m_fd, F_DUPFD));
	m_peer_reset = s.m_peer_reset;
        return *this;
      }

      socket::eventmask socket::poll ( socket::eventmask accepted_events) {
        struct pollfd fdstr;  		
        fdstr.fd = fd();
        fdstr.events = accepted_events;
	long timeout = get_timeout();
        long time_now ;
        ::time(&time_now);
        long time_end = time_now + timeout;
        int res = 0;
	
        do {
          res = ::poll(&fdstr,1,1000);
          if (res>0) {
            WOOL_DEBUG("poll events avail... " );
            return fdstr.revents;
          }
          if((res<0)&&(errno!=EINTR)) {
            WOOL_DEBUG("error/invalid fd? ... " << ::strerror(errno));
            return accepted_events;
          }
          ::time(&time_now);
	  WOOL_DEBUG("check timeout: "<<time_now<<":"<<time_end<<"("<<timeout<<")");
        } while((timeout<0)||(time_now<time_end));
        WOOL_DEBUG("timed out ... " );
        return 0;
      }

      void socket::close () {
        if(m_fd>=0){
          do {
            ::close(m_fd);
          } while(errno==EINTR);
        }
        m_fd = -1;
      }

      socket::size_type socket::__recv_from (wool::address * a, unsigned char * buffer, const unsigned int & total, const flagset & fs) {	
        static socklen_t addrlen = sizeof(::sockaddr);
        static socklen_t addrlen_zero = 0;
	static const int events = 	socket::poll_invalid_fd|
	  socket::poll_error|
	  socket::poll_hup|
	  socket::poll_xnormal_read|
	  socket::poll_input;
	int pres;
	if((fs & MSG_DONTWAIT)==0)
	  pres = poll(events);
	WOOL_DEBUG("poll done "<<pres<< " "<<wool::exception::c_error());
	int j = ::recvfrom (fd(), buffer, total, fs | MSG_DONTWAIT,
			    a, (a==NULL)?&addrlen_zero:&addrlen);	
	WOOL_DEBUG("recv done "<<j<<" ("<<total<<")"<<wool::exception::c_error());
	if(j==0)
	  m_peer_reset = true;
	return std::max<int>(0,j);		
      }

      socket::size_type socket::__send_to (const wool::address * a, const unsigned char * buffer, const unsigned int & total, const flagset & fs) {
        size_type s = total;
        int i;
        static socklen_t addrlen = sizeof(::sockaddr);
        static socklen_t addrlen_zero = 0;
	bool to = ((fs & MSG_DONTWAIT)==0);
	flagset f = to?(fs|MSG_DONTWAIT):fs;
	static const int events = 	socket::poll_invalid_fd|
	  socket::poll_error|
	  socket::poll_hup|
	  socket::poll_xnormal_write|
	  socket::poll_output;
        do {
	  int pres;
	  if(to)
	    pres = poll(events);
	  WOOL_DEBUG("poll done "<<pres<< " "<<wool::exception::c_error());
	  int j = ::sendto (fd(), buffer, s, f,a, (a==NULL)?addrlen_zero:addrlen);
	  if(j==0)
	    m_peer_reset = true;
	  i = std::max<int>(0,j);
	  WOOL_DEBUG("send done "<<j<<" ("<<s<<")"<<wool::exception::c_error());
	  buffer +=i;
	  s-=i;
        } while ((s>0) && (i>0));
        return total - s;
      }

  bool socket::peer_reset() const {
    return m_peer_reset;
  }


      socket::size_type socket::send  (const unsigned char * buffer, const size_type & s, const flagset & f) {
      	return __send_to(NULL, buffer,s,f);
      }
      
      socket::size_type socket::send_to  (const address & a, const unsigned char * buffer, const size_type & s, const flagset & f) {
      	return __send_to(&a, buffer,s,f);
      }
      
      socket::size_type socket::recv  (unsigned char * buffer, const size_type & s, const flagset & f) {
      	return __recv_from(NULL, buffer,s,f);
      }
      
      socket::size_type socket::recv_from  (address & a, unsigned char * buffer, const size_type & s, const flagset & f) {
      	return __recv_from(&a, buffer,s,f);
      }
      
      void socket::set_timeout ( const int & _t) {
      	WOOL_DEBUG("Set timeout to "<<_t);
        m_timeout = _t;
      }

      const int & socket::get_timeout ( ) const {
        return m_timeout;
      }

      socket::fd_type socket::release() {
        int __old_fd = m_fd;
        m_fd = -1;
        return __old_fd;
      }

      socket::fd_type socket::fd() const {
        return m_fd;
      };

      void socket::fd( socket::fd_type __fd ){
        close();
	m_peer_reset = false;
        m_fd = __fd;
      };
      
      
   	  void socket::connect ( const wool::address & a) {
	    int oldflags = flags();
	    set_flag((flag)O_NONBLOCK);
	    int i = ::connect(m_fd, &a, sizeof(wool::address));
	    if (i!=0) {
	    	if (errno!=EINPROGRESS) {
	  	      flags(oldflags);
		      throw wool::socket::exception();
		}
	     }
	     flags(oldflags);
   	  }
     
      void socket::bind (const wool::address & t) {
	
        setsockopt(SOL_SOCKET, SO_REUSEADDR,1);
      	const sockaddr* addr = &t;
        if(::bind (fd(), (struct sockaddr *) addr, sizeof (struct sockaddr))<0)
        	throw socket::exception(errno);     	
      }
      
      void socket::listen (const int & backtrace) {
        
        //int res = ::setsockopt (fd(), SOL_SOCKET, SO_REUSEADDR, &i, sizeof (int));        
        if(::listen (fd(),backtrace)<0)
        	throw  socket::exception (errno);
      }

      void socket::accept ( socket & s ) {
        static socklen_t l;
        static const int events = socket::poll_xnormal_read|socket::poll_input;
        socket::fd_type __fd = -1;
		
        if ( poll(events) ==0 )
          throw socket::exception(errno);

	      do {
	        __fd=::accept(fd(),NULL,&l);
	        WOOL_DEBUG("accept returned");
	        //retry if accept failed due to an interruption
	      } while((__fd<0)&&(errno==EINTR));

        if(__fd<0)
          throw socket::exception(errno);

        s.fd(__fd);
      }   
      
      address socket::get_socket_address () const {
      	address a;
      	socklen_t l = sizeof(a);
      	if(::getsockname(fd(), &a, &l)!=0)
      		throw socket::exception(errno);
      	return a;
      }
      
	  address socket::get_peer_address () const {
	  	address a;
      	socklen_t l = sizeof(a);
      	if(::getpeername(fd(), &a, &l)!=0)
      		throw socket::exception(errno);
      	return a;
	  }

};

