/**************************************************************************
*   Copyright (C) 2004 by Vukasin Toroman                                 *
*   vukasin@toroman.org                                                   *
**************************************************************************/
/**************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifndef WOOL_SOCKET_H
#define WOOL_SOCKET_H

#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <string.h>
#include <iostream>
#include <fcntl.h>
#include <assert.h>
#include <poll.h>

#include <linux/netfilter_ipv4.h>

#include "../base/exception.h"
#include "../stream/streambuf.h"


namespace wool {
	struct address : public ::sockaddr {
		public:
			typedef sa_family_t family;
			typedef enum {
				unspec = AF_UNSPEC,
				local = AF_LOCAL,
				inet = AF_INET,
				ax25 = AF_AX25,
				ipx = AF_IPX,
				appletalk = AF_APPLETALK,
				netrom = AF_NETROM,
				bridge = AF_BRIDGE,
				atmpvc = AF_ATMPVC,
				x25 = AF_X25,
				inet6 = AF_INET6,
				rose = AF_ROSE,
				dec_net = AF_DECnet,
				netbeui = AF_NETBEUI,
				security = AF_SECURITY,
				key = AF_KEY,
				netlink = AF_NETLINK,
				packet = AF_PACKET,
				ash = AF_ASH,
				econet = AF_ECONET,
				atm_svc = AF_ATMSVC,
				sna = AF_SNA,
				irda = AF_IRDA,
				pppox = AF_PPPOX,
				wanpipe = AF_WANPIPE,
				bluetooth = AF_BLUETOOTH,
			} family_specs;
			
			address( const family & f = unspec);
			address( const ::sockaddr & f );

			void set_family( const family & f );
			const family & get_family() const;
		protected:
			void * get_data();
			const void * get_data() const;
	};

	class socket {
		public:
			typedef int fd_type;
			typedef unsigned size_type;
			typedef char char_type;
			
			typedef enum {
				unspec = PF_UNSPEC,
				local = PF_LOCAL,
				inet = PF_INET,
				ax25 = PF_AX25,
				ipx = PF_IPX,
				appletalk = PF_APPLETALK,
				netrom = PF_NETROM,
				bridge = PF_BRIDGE,
				atmpvc = PF_ATMPVC,
				x25 = PF_X25,
				inet6 = PF_INET6,
				rose = PF_ROSE,
				dec_net = PF_DECnet,
				netbeui = PF_NETBEUI,
				security = PF_SECURITY,
				key = PF_KEY,
				netlink = PF_NETLINK,
				packet = PF_PACKET,
				ash = PF_ASH,
				econet = PF_ECONET,
				atm_svc = PF_ATMSVC,
				sna = PF_SNA,
				irda = PF_IRDA,
				pppox = PF_PPPOX,
				wanpipe = PF_WANPIPE,
				bluetooth = PF_BLUETOOTH,
			} domain;
			
			typedef enum {
				stream_socket = SOCK_STREAM,
				dgram_socket = SOCK_DGRAM,
				rdm_socket = SOCK_RDM,
				seqpacket_socket = SOCK_SEQPACKET,
				packet_socket = SOCK_PACKET
			} type;
			
			typedef enum {
			    oob = MSG_OOB, 				/* Process out-of-band data.  */
			    peek = MSG_PEEK, 			/* Peek at incoming messages.  */
			    dontroute = MSG_DONTROUTE, 	/* Don't use local routing.  */
			    ctrunc = MSG_CTRUNC, 		/* Control data lost before delivery.  */
			    proxy = MSG_PROXY,			/* Supply or ask second address.  */
			    trunc = MSG_TRUNC,
			    dontwait = MSG_DONTWAIT,	/* Nonblocking IO.  */
			    eor = MSG_EOR, 				/* End of record.  */
			    waitall = MSG_WAITALL,		/* Wait for a full request.  */
			    fin = MSG_FIN,
			    syn = MSG_SYN,
			    confirm = MSG_CONFIRM,		/* Confirm path validity.  */
			    rst = MSG_RST,
			    errqueue = MSG_ERRQUEUE,	/* Fetch message from error queue.  */
			    nosignal = MSG_NOSIGNAL,	/* Do not generate SIGPIPE.  */
			    more = MSG_MORE  			/* Sender will send more.  */
			} flag;
			
			typedef int flagset;
			typedef int protocol;
			typedef int level;
			typedef short eventmask;
		
     static const flag tcp_nodelay;
     static const flag tcp_cork;

      static const level sol_tcp;
      static const level sol_ip;
      static const level sol_socket;

      //static const eventmask poll_any;
      typedef enum {
      	poll_input = POLLIN,
      	poll_urgent_input = POLLPRI,
      	poll_output = POLLOUT, 
      	poll_error = POLLERR, 
      	poll_hup = POLLHUP,
      	poll_invalid_fd = POLLNVAL,
      	poll_xnormal_read = POLLRDNORM,
      	poll_xpriority_read = POLLRDBAND,
      	poll_xnormal_write = POLLWRNORM,
      	poll_xpriority_write = POLLWRBAND,
      	poll_any = 0xffff
      } poll_event;


      class exception : public wool::exception {
        public:
          exception ( int __errno = errno ) throw () : wool::exception (wool::exception::c_error(__errno)){
            m_error_message.insert(0,"socket error:");
          }
          virtual ~exception () throw () {      }
      };

	  struct type_record {
	  	socket::domain domain;
	  	socket::type type;
	  	socket::protocol protocol;
	  };
	  
	  socket ( const type_record & t );
      socket ( socket::domain d , socket::type t , socket::protocol p );
      socket ( const socket & s );
      virtual ~socket();

      socket & operator = ( const socket & s  );
      bool operator ! (  ) const;    
      void close ();

      int fcntl(int cmd);
      int fcntl(int cmd, long arg);
      int fcntl(int cmd, struct flock *lock);

      eventmask poll ( eventmask accepted_events);

      fd_type release();
      fd_type fd() const;
      void fd( fd_type );

      flagset flags () const;
      void flags (flagset);

      bool test_flag (socket::flag f) const;
      void set_flag (socket::flag f);
      void reset_flag (socket::flag f);
      
		address get_socket_address () const;
		address get_peer_address () const;
		
		template <typename T>
		void setsockopt( int level, int optname, T t ) {
			socklen_t len = sizeof(T);
	    	int res =::setsockopt (m_fd, level, optname, &t, len);
	    	if (res<0) wool::socket::exception();
		}

		template <typename T>
		void getsockopt( int level, int optname, T & t ) {
		/**
	      *  get dst address before a netfilter redirect ::getsockopt (fd(),SOL_IP,SO_ORIGINAL_DST,&a,&l)
		  */
			
			socklen_t len = sizeof(T);
			int res =::getsockopt (m_fd, level, optname, &t, &len);
			if (res<0) wool::socket::exception();
		}
		
		void set_timeout(const int & t);
		const int & get_timeout() const;

	  inline size_type flush   	() { return 0; }
	  inline size_type read    	( char * b, const size_type & sz) { return this->recv((unsigned char *)b,sz); }
	  inline size_type write   	( const char * b, const size_type & sz) { return this->send((const unsigned char *)b,sz); }
	  size_type send    	( const unsigned char * b, const size_type & sz, const flagset & f = 0x00);
	  size_type recv    	( unsigned char * b, const size_type & sz, const flagset & f  = 0x00);
	  size_type send_to 	( const wool::address & a, const unsigned char * b, const size_type & sz, const flagset & f  = 0x00);
	  size_type recv_from   ( wool::address & a, unsigned char * b, const size_type & sz, const flagset & f  = 0x00);
	  
	  template <typename T>
	  void send    	( const T & t, const flagset & f  = 0x00 ) {  	
	  	size_type sz = sizeof(t);
	  	const unsigned char * ve = (const unsigned char *)&t + sz;
	  	do {
	  		sz -= send(ve - sz, sz, f);
	  	} while(sz>0);
	  }

	  template <typename T>
	  void recv  	( T & t, const flagset & f  = 0x00) {
	  	size_type sz = sizeof(t);
	  	unsigned char * ve = (unsigned char *)&t + sz;
	  	do {
	  		sz -= recv(ve - sz, sz, f);
	  	} while(sz>0);
	  }

	  template <typename T>
	  void send_to    	( const wool::address & a, const T & t, const flagset & f  = 0x00) {
	  	size_type sz = sizeof(t);
	  	const unsigned char * ve = (const unsigned char*)&t + sz;
	  	do {
	  		sz -= send_to(a,ve - sz, sz, f);
	  	} while(sz>0);
	  }

	  template <typename T>
	  void recv_from  	( wool::address & a, T & t, const flagset & f  = 0x00) {
		size_type sz = sizeof(t);
	  	unsigned char * ve = (unsigned char *)&t + sz;
	  	do {
	  		sz -= recv_from(a, ve - sz, sz, f);
	  	} while(sz>0);
		
	  }
	  
	  void connect ( const wool::address & a);
	  void bind    ( const wool::address & a);
	  void listen  ( const int & backtrace );
	  void accept  ( socket & s );

	  bool peer_reset() const;

   	private:
   		size_type __send_to 	( const wool::address * a, const unsigned char * b, const size_type & sz, const flagset & f);
	  	size_type __recv_from   ( wool::address * a, unsigned char * b, const size_type & sz, const flagset & f);
		fd_type m_fd;
		int m_timeout;
		bool m_peer_reset;
  };

};

#endif
