/**************************************************************************
*   Copyright (C) 2004 by Vukasin Toroman                                 *
*   vukasin@toroman.org                                                   *
**************************************************************************/
/************************************************************************** 
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifndef SWAPPABLEDATA_H
#define SWAPPABLEDATA_H

#include "../base/cvs_info.h"
#include "../storage/tmp_file.h"

#include <sys/mman.h>
#include <boost/shared_ptr.hpp>
#include <ios>

namespace wool
  {
  /**
    * Class to swap automaticly data out to the hard disc.
    * @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
    * @version $Id: swappable_data.h,v 1.4 2007/01/11 15:18:41 vtoroman Exp $
    */
  class swappable_data
    {
    public:
      /**
       * wrapper class for std::ostrestreambuf
       * @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
       * @version $Id: swappable_data.h,v 1.4 2007/01/11 15:18:41 vtoroman Exp $
       */
    class streambuf:public std::streambuf
        {
        public:
          /**
           *  cvs info of swappable_data.h file
           */
          VERSION_INFO_INLINE("$Id: swappable_data.h,v 1.4 2007/01/11 15:18:41 vtoroman Exp $");

          streambuf                      ();
          virtual ~streambuf             ();
          /**
           *  Returns if the streambuffer is opened
           *  @return true if opened else false
           */
          bool is_open                   ();
          /**
           *  open a streambuf
           *  @param data swappable_data
           *  @param om mode to open
           */
          void open                      (swappable_data & data,
                                          std::ios_base::openmode om=std::ios_base::ate|std::ios_base::out|std::ios_base::app);
          /**
           *  open a streambuf
           *  @param data swappable_data
           *  @param om mode to open
           */
          void open                      (const swappable_data & data,
                                          std::ios_base::openmode om=std::ios_base::in);
          /**
           *  close the streambuffer
           */
          void close                     ();

        protected:
          /**
           *  Synchronizes the buffer arrays with the controlled sequences.
           *  @return -1 on failure
           */
          virtual int sync               ();
          /**
           *  Investigating the data available.
           *  @return An estimate of the number of characters available in the input sequence, or -1.
           */
          virtual int showmanyc          ();
          /**
           *  Tries to back up the input sequence.
           *  @param c The character to be inserted back into the sequence.
           *  @return eof() on failure, "some other value" on success
           */
          virtual int_type pbackfail     ( int_type c = std::filebuf::traits_type::eof() );
          /**
           *  Fetches more data from the controlled sequence.
           *  @return The first character from the pending sequence.
           */
          virtual int_type underflow     ();
          /**
           *  Fetches more data from the controlled sequence.
           *  @return The first character from the pending sequence.
           */
          virtual int_type uflow         ();
          /**
           *  Multiple character extraction.
           *  @param src A buffer area.
           *  @param count Maximum number of characters to assign.
           */
          virtual std::streamsize xsgetn ( char *src, std::streamsize count);
          /**
           *  Multiple character insertion.
           *  @param src A buffer area.
           *  @param count Maximum number of characters to write.
           */
          virtual std::streamsize xsputn ( const char *src, std::streamsize count);
          /**
           *  Consumes data from the buffer; writes to the controlled sequence.
           *  @param i  An additional character to consume.
           *  @return eof() to indicate failure, something else (usually i, or not_eof())
           */
          virtual int_type overflow      ( int_type i );
          /**
           *  Alters the stream positions.
           *  @param sp streampos
           *  @param which mode to open
           *  @return streampos
           */
          virtual std::streampos seekpos ( std::streampos sp,
                                           std::ios_base::openmode which =
                                              std::ios_base::in|std::ios_base::out );
          /**
           *  Alter the stream positions.
           *  @param off  streamoff
           *  @param way direction
           *  @param which mode to open
           *  @return streampos
           */
          virtual std::streampos seekoff ( std::streamoff off,
                                           std::ios_base::seekdir way,
                                           std::ios_base::openmode which =
                                              std::ios_base::in | std::ios_base::out|std::ios_base::app );
        private:
          void check_size( std::streamsize );
          swappable_data * m_data;
          std::streampos m_offset;
          std::ios_base::openmode m_mode;
          std::filebuf m_filebuf;
        };
      /**
       * wrapper class for std::ostream
       * @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
       * @version $Id: swappable_data.h,v 1.4 2007/01/11 15:18:41 vtoroman Exp $
       */
      class ostream:public std::ostream
      {
        public:
          inline ostream():std::ostream(&m_buf) {}
          inline virtual ~ostream() {}
          /**
           *  Accessing the underlying buffer.
           *  @return The current stream buffer.
           */
          inline swappable_data::streambuf * rdbuf() {return &m_buf; }
          /**
           *  Returns if the buffer is opened.
           *  @return True if is opened else false.
           */
          inline bool is_open() { return m_buf.is_open(); }
          /**
           *  Open the stream.
           *  @param data swappable_data to stream to
           *  @param mode mode to open the stream
           */
          inline void open   (swappable_data & data,
                              std::ios::openmode mode=std::ios::out|std::ios::ate|std::ios::app )
          { m_buf.open(data,mode);}
          /**
           *  Close the stream.
           */
          inline void close() {m_buf.close();};
        private:
          swappable_data::streambuf m_buf;
      };

      /**
       * wrapper class for std::istream
       * @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
       * @version $Id: swappable_data.h,v 1.4 2007/01/11 15:18:41 vtoroman Exp $
       */
      class istream:public std::istream
      {
        public:
          inline istream():std::istream(&m_buf) {}
          inline virtual ~istream() {}
           /**
            * Accessing the underlying buffer.
            * @return The current stream buffer.
            */
          inline swappable_data::streambuf * rdbuf() {return &m_buf; }
          /**
           *  Returns if the buffer is opened.
           *  @return True if is opened else false.
           */
          inline bool is_open() { return m_buf.is_open(); }
          /**
           *  Open the stream.
           *  @param data swappable_data to stream to
           *  @param mode mode to open the stream
           */
          inline void open   (const swappable_data & data,
                              std::ios::openmode mode=std::ios::in)
          { m_buf.open(data,mode);}
          /**
           *  Close the stream.
           */
          inline void close() {m_buf.close();};
        private:
          swappable_data::streambuf m_buf;
      };




      friend class streambuf;
      /**
       *  cvs info
       */
      VERSION_INFO_INLINE("$Id: swappable_data.h,v 1.4 2007/01/11 15:18:41 vtoroman Exp $");
      /**
       *  cvs info
       */
      VERSION_INFO_DECL;
      /**
       *  value definition
       */
      typedef char value_type;
      /**
       *  path definition
       */
      typedef std::string path_type;
      /**
       *  data definition
       */
      typedef std::string data_type;
      /**
       *  size definition
       */
      typedef size_t size_type;
      /**
       *  filedescriptor definition
       */
      typedef int os_fd_type;

      /**
       *  Size of data which should hold in memory
       */
      static const size_type resident_size;

      swappable_data () throw();
      /**
       *  Copy constructor
       */
      swappable_data ( const swappable_data & src );

      /**
       *  Assignment operator
       *  @param src data to assign
       *  @return reference to swappable_data
       */
      swappable_data & operator = ( const swappable_data & src );

      ~swappable_data () throw();
      /**
       *  Appends data.
       *  @param data data to append
       */
      template <typename T>
      inline void append ( T data )
      {
        wool::swappable_data::ostream os;
        os.open(*this,std::ios::out|std::ios::ate|std::ios::app);
        os<< data;
      }

      /**
       *  Returns the size of the data
       *  @return size of the data
       */
      size_type size () const;

      /**
       * For CPPUnit-Test only
       * @return filename of tempfile
       * @deprecated
       */
      std::string swap_file();

       /**
       * Swap the content with another objects'
       * @param v
       * @return filename of tempfile
       */
      void swap( wool::swappable_data & v);

      /**
       * Swap out data manually to hard disc.
       */
      void swap_out();

      /**
       *  Deletes all data
       */
      void clear() throw ();

      void load( std::istream & in);
      void save( std::ostream & out) const;
    private:
      path_type m_fname;
      data_type m_data;
    };
};

namespace std
{
  std::ostream & operator << ( std::ostream & os, const wool::swappable_data & d );
  std::istream & operator >> ( std::istream & is, wool::swappable_data & d );
}
#endif

