/**************************************************************************
*   Copyright (C) 2004 by Vukasin Toroman                                 *
*   vukasin@toroman.org                                                   *
**************************************************************************/
/************************************************************************** 
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifndef WOOL_STREAMBUF_H
#define WOOL_STREAMBUF_H

#include <iostream>
#include "../base/debug.h"

namespace wool {
  /**
   */
  template <typename ioT>
  class io_traits {
    public:
      /**
      *  IO type
      */
      typedef ioT io_type;

      inline static std::streamsize write (io_type & io, const char * buffer, std::streamsize size) {
        return io.write ( buffer, size);
      }

      inline static std::streamsize read (io_type & io, char * buffer, std::streamsize size) {
        return io.read ( buffer, size);
      }

      inline static std::streamsize flush (io_type & io) {
        //return 1;
        return io.flush ();
      }
  };

  /**
   * Wrapper class for std::basic_streambuf.
   *  @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
   */
  template <typename ioT, typename io_traitsT=io_traits<ioT> >
  class streambuf
        :public std::streambuf
  {
  private:
    ioT & m_io;
    char * m_buffer;
    std::streamsize m_buffer_size;
  public:
    /**
     *  IO type
     */
    typedef ioT io_type;
    /**
     *  traits type
     */
    typedef io_traitsT io_traits_type;

	inline const io_type & io() const { return m_io; };
	inline io_type & io() { return m_io; };
	
    /**
     *  @param b  reference to the IO
     */
    streambuf (io_type & b):m_io (b)
    {
        m_buffer = new char[2048];
        setbuf(m_buffer,2048);
    }
    /**
     *  Flush the IO and delete the internal buffer.
     */
    virtual ~streambuf ( ) { io_traits_type::flush(m_io); if(m_buffer) delete m_buffer;}


  protected:
    std::streambuf *
    setbuf ( char * s, std::streamsize n ){
        WOOL_DEBUG("set buffer to "<< n <<"bytes");
        if((s==NULL)||(n<2))
          return NULL;
        if((m_buffer!=NULL)&&(s!=m_buffer)) {
          delete m_buffer;
          m_buffer = NULL;
        }
        m_buffer_size = n;
        setg (s,s+n/2,s+n/2);
        setp (s+n/2,s+n);
        return this;
    }

    /**
     *  Returns the character at current get position, or EOF if the current get pointer is at
     *  the end of the input sequence.
     */
    virtual int_type underflow () {
     if(sync()==-1)
     	return traits_type::eof();
     if(in_avail( )<=0 ){
        int_type res = io_traits_type::read(m_io,m_buffer,m_buffer_size/2);
        WOOL_DEBUG("refill res: "<< res);
        if( res <= 0 )
          return traits_type::eof();
        setg(m_buffer, m_buffer, m_buffer+res);
      }
      WOOL_DEBUG("return: "<< traits_type::to_int_type(*gptr()));
      return traits_type::to_int_type(*gptr());
    }

    /**
     *  Syncronizes the content of the stream's internal buffer with the physical source
     *  of characters associated with the stream.
     *  @return Returns 1 for success and 0 or -1 for failure.
     */
     virtual int sync () {
      return(overflow()==-1)?-1:1;
    }
    
    
	/** Writes n characters from the array specified by parameter s to the output sequence of the stream buffer.
  */
	virtual std::streamsize other_xsputn ( const char * s, std::streamsize n ) {
		std::streamsize res = 0;
		int_type r = 0;
		if(overflow()==traits_type::eof())
			return 0;
		while(n>res) {
			r = io_traits_type::write(m_io, s, n-res);
			if(r<=0)
				return res;
			res+=r;
			s+=r;
		}
		return res;
	}
	
	/** WGets a series of characters from in input sequence and stores them in the array pointed by s until n characters have been copied or the input stream has no more characters. 
  */
	virtual std::streamsize other_xsgetn ( char * s, std::streamsize n ) {
		if(sync()==-1)
     		return 0;
		int_type r = 0;
		std::streamsize res = std::min(in_avail(), n);
		memmove( gptr(), s, res);
		s+=res;
		
		while(n>res) {
			r = io_traits_type::read(m_io, s, n-res);
			if(r<=0)
				return res;
			res+=r;
			s+=r;
		}
		return res;
	}
	

    /**
     *   Put character at current put position.
     *   @param i     Character to be put.
     *   @return 1 on success or -1 on failure.
     */
    virtual int_type overflow ( int_type i = traits_type::eof()) {
        WOOL_DEBUG("insert: "<< i);
        int_type buf_filled;
        
        if(i!=traits_type::eof()) {
          if((epptr()-pptr())<=0){
            if(overflow()!=1) {
              WOOL_DEBUG("return (error) : "<< -1);
              return traits_type::eof();
            }
          }
          *pptr() = static_cast<char>(i);
          pbump(1);
        } else {
        		buf_filled = pptr()-pbase();
        		int_type res = 1;
				int_type sent = 0;
				while ((buf_filled>sent)&&(res>0)) {
					res = io_traits_type::write(m_io, pbase()+sent, buf_filled-sent);
					sent+=std::max<int_type>(0,res);
				}
				
				buf_filled-=sent;
				pbump(-sent);
				if(res<=0) {
					memmove(pbase(),pptr()+buf_filled-sent,buf_filled);
					WOOL_DEBUG("return (error) : "<< -1);
					return traits_type::eof();
				}
        }
        WOOL_DEBUG("return : "<< 1);
        return 1;
      }
    };
    
    template <typename ioT, typename io_traitsT=io_traits<ioT> >
  	class iostream
        :public std::iostream {
    	public:
    		wool::streambuf<ioT, io_traitsT> m_streambuf;
    		
    		inline iostream( ioT & s ):std::iostream(NULL), m_streambuf(s) {
    			rdbuf(&m_streambuf);
    		}

    		virtual inline ~iostream() {};
    };

};
#endif
