/**************************************************************************
*   Copyright (C) 2004 by Vukasin Toroman                                 *
*   vukasin@toroman.org                                                   *
**************************************************************************/
/************************************************************************** 
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifndef SSLBIO_H
#define SSLBIO_H

#include <iostream>
#include <string>

#include <openssl/bio.h>
#include <openssl/ssl.h>

#include "../stream/streambuf.h"
#include "../base/wrapper.hh"

namespace wool
{
/**
 *  This namespace contains classes which provides basic functionality for I/O methods. It is used as
 *  additional abstraction layer. If an application or library uses it for its I/O it can transparently
 *  handle SSL connections, unencrypted network connections and file I/O.
 *  @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
 */
namespace swool
{

  /**
   *  Basic wrapper class for openssl bio methods.
   *  @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
   */
  class bio : public Wrapper<BIO>
  {

  public:
    /**
     *  pointer to BIO method definition<p>
     *  There are two type of BIO, a source/sink BIO and a filter BIO.<p>
     *  As its name implies a source/sink BIO is a source and/or sink of data,
     *  examples include a socket BIO and a file BIO.<p>
     *  A filter BIO takes data from one BIO and passes it through to another, or the application.
     *  The data may be left unmodified (for example a message digest BIO) or translated
     *  (for example an encryption BIO). The effect of a filter BIO may change according
     *  (to the I/O operation it is performing: for example an encryption BIO will encrypt data
     *  (if it is being written to and decrypt data if it is being read from.
     */
    typedef BIO_METHOD * method;
    /**
     *  method definition
     */
    typedef method (*method_factory)(void );
    /**
     *  callback definition for BIO methods
     */
    typedef void (*callback)(struct bio_st *, int, const char *, int, long, long);
    /**
     *  callback definition for info methods
     */
    typedef void (*info_cb) (BIO *, int , const char *, int , long , long );

    /**
     * Null source/sink BIO method.<p>
     * Data written to the null sink is discarded, reads return EOF.
     */
    static method_factory s_null;
    /**
     * File source/sink BIO method.<p>
     * As its name implies it is a wrapper round the stdio FILE structure and it is a source/sink BIO.
     */
    static method_factory s_file;
    /**
     *  Memory source/sink BIO method.<p>
     *  Data written to a memory BIO is stored in a BUF_MEM structure which is extended
     *  as appropriate to accommodate the stored data.
     */
    static method_factory s_mem;
    /**
     *  Socket source/sink BIO method.<p>
     *  This is a wrapper round the platform's socket routines.
     */
    static method_factory s_socket;
    /**
     *  Connect source/sink BIO method.<p>
     *  This is a wrapper round the platform's TCP/IP socket connection routines.
     */
    static method_factory s_connect;
    /**
     *  Accept source/sink BIO method.<p>
     *  This is a wrapper round the platform's TCP/IP socket accept routines.
     */
    static method_factory s_accept;
    /**
     *  Filedescriptor source/sink BIO method.<p>
     *  This is a wrapper round the platforms file descriptor routines such as read() and write().
     */
    static method_factory s_fd;
    /**
     *  Bio source/sink BIO method<p>
     *  This method returns a returns a BIO pair. A BIO pair is a pair of source/sink BIOs where
     *  data written to either half of the pair is buffered and can be read from the other half.
     */
    static method_factory s_bio;
    /**
     *  NULL filter BIO method.<p>
     *  This is a filter BIO that does nothing.
     */
    static method_factory f_null;
    /**
     *  Buffering filter BIO method.<p>
     *  Data written to a buffering BIO is buffered
     *  and periodically written to the next BIO in the chain.
     *  Data read from a buffering BIO comes from an internal buffer which is filled
     *  from the next BIO in the chain.
     */
    static method_factory f_buffer;
    /**
     *  SSL filter BIO method.<p>
     *  This is a filter BIO which is a wrapper round the OpenSSL SSL routines adding a BIO ``flavour'' to SSL I/O.
     */
    static method_factory f_ssl;
    /**
     *  Message digest BIO method.<p>
     *  This is a filter BIO that digests any data passed through it.
     */
    static method_factory f_message_digest;
    /**
     *  Base64 filter BIO method.<p>
     *  This is a filter BIO that base64 encodes any data written through it and decodes any data read through it.
     */
    static method_factory f_base64;
    /**
     *  Cipher filter BIOD method.<p>
     *  This is a filter BIO that encrypts any data written through it, and decrypts any data read from it.
     */
    static method_factory f_cipher;
    /**
     *  @param b BIO
     */
    bio( BIO * b, bool owns = true );

    /**
     *  @param m BIO method
     */
    bio( bio::method m );

    /**
     *  @param m BIO method factory
     */
    bio( bio::method_factory m = f_null);

    /**
     *  Free all BIO's
     */
    virtual ~bio();

    /**
     *  Push a new BIO into the chain.
     *  @param b reference to the new BIO
     */
    void push (bio & b);
    
    /**
     *  Pop the last BIO in the chain.
     */
    void pop ();
    /**
     *  Reads 's' bytes from the actual BIO into the buffer 'dst'.
     *  @param dst destination buffer
     *  @param s number of bytes to read
     */
    int read  (void * dst, int s);
    /**
     *  Writes 's' bytes into the actual BIO from the buffer.
     *  @param src  source buffer
     *  @param s number of bytes to write
     */
    int write (const void * src, int s);
    
    void set_flags( int f );
    void reset_flags( int f );

    int get_flags( ) const;
    
    /**
     *  BIO control operation. This function is not normally called directly, various macros are used instead.
     *  @param cmd  command
     *  @param larg number of arguments
     *  @param parg arguments
     *  @return return value of the executed command 'cmd'
     */
    long ctrl(int cmd,long larg,void *parg);
    /**
     *  BIO callback control operation. This function is not normally called directly, various macros are used instead.
     *  @param cmd command
     *  @param f callback function
     *  @return return value of the executed command 'cmd'
     */
    long callback_ctrl (int cmd, callback f);
    /**
     *  BIO pointer control operation. This function is not normally called directly, various macros are used instead.
     *  @param cmd command
     *  @param larg larg
     *  @return return value of the executed command 'cmd'
     */
    char * ptr_ctrl(int cmd,long larg);
    /**
     *  BIO int control method. This function is not normally called directly, various macros are used instead.
     *  @param cmd command
     *  @param larg larg
     *  @param iarg iarg
     *  @return return value of the executed command 'cmd'
     */
    long int_ctrl(int cmd,long larg,int iarg);
    /**
     *  Typically resets a BIO to some initial state, in the case of file related BIOs for example
     *  it rewinds the file pointer to the start of the file.
     *  @return 1 for success else 0 or -1
     */
    int reset();
    /**
     *  Resets a file related BIO's (that is file descriptor and FILE BIOs) file position pointer
     *  to ofs bytes from start of file.
     *  @return current file position on success else -1
     */
    int seek(int ofs);
    /**
     *  Returns the current file position of a file related BIO.
     *  @return current file position on success else -1
     */
    int tell();
    /**
     *  Normally writes out any internally buffered data, in some cases it is used
     *  to signal EOF and that no more data will be written.
     *  @return 1 for success else 0 or -1
     */
    int flush();
    /**
     *  Returns 1 if EOF has been reached 0 otherwise.
     *  @return 1 if EOF has been reached 0 otherwise
     */
    int eof();
    /**
     *  Sets the BIO b close flag to flag. flag can take the value BIO_CLOSE or BIO_NOCLOSE.
     *  Typically BIO_CLOSE is used in a source/sink BIO to indicate that the underlying I/O stream
     *  should be closed when the BIO is freed.
     *  @param flag BIO_CLOSE | BIO_NOCLOSE
     *  @return always 1
     */
    int set_close(long flag);
    /**
     *  Returns the BIOs close flag.
     *  @return BIO_CLOSE | BIO_NOCLOSE
     */
    int get_close();
    /**
     *  Return the number of pending characters in the BIOs read and write buffers.
     *  @return the amount of pending data
     */
    int pending();
    /**
     *  Return the number of pending characters in the BIOs read and write buffers.
     *  @return the amount of pending data
     */
    int wpending();
    /**
    *  Return the number of pending characters in the BIOs read and write buffers.
    *  @return the amount of pending data
    */
    size_t ctrl_pending();
    /**
     *  Return the number of pending characters in the BIOs read and write buffers.
     *  @return the amount of pending data
     */
    size_t ctrl_wpending();

  protected:
    virtual void free( BIO * b );
    virtual BIO * generate( );
    virtual void update();
  };

  /*
   *  buffer size for buffered BIO's
   */
  //const std::streamsize bio_streambuf_buffer_size = 10;

  typedef wool::streambuf<swool::bio> bio_streambuf;

  /**
   *  Wrapper class for std::basic_iostream.
   *  @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
   */
  class bio_stream:public std::iostream
  {
  private:
    bio_streambuf strbf;
  public:
    /**
     *  @param b  BIO
     */
    bio_stream(swool::bio & b):std::iostream(&strbf),strbf(b) { }
    virtual ~bio_stream( ) {}
  };
  /**
   *  Wrapper class for std::basic_ostream.
   *  @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
   */
  class bio_ostream: public std::ostream
  {
  private:
    bio_streambuf strbf;
  public:
    /**
     *  @param b  BIO
     */
    bio_ostream(swool::bio & b):std::ostream(&strbf),strbf(b) { }
    virtual ~bio_ostream( ) {}
  };
  /**
   *  Wrapper class for std::basic_istream.
   *  @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
   */
  class bio_istream:public std::istream
  {
  private:
    bio_streambuf strbf;
  public:
    /**
     *  @param b  BIO
     */
    bio_istream(swool::bio & b):std::istream(&strbf),strbf(b) { }
    virtual ~bio_istream( ) {}
  };
};
};
#endif
