#include "../swool/cert.hh"
#include "../swool/streambuf_bio.h"
#include "../base/exception.h"
#include <openssl/err.h>
#include <openssl/asn1.h>

namespace wool {
  namespace swool {
    X509Cert::X509Cert( X509 * r, bool owns  ) :Wrapper<X509>(r,owns),  m_subject(NULL), m_issuer(NULL), m_publicKey(NULL) {
      init();
    }
    
    const X509Cert & X509Cert::print ( std::ostream & out ) const {
      wool::swool::streambuf_bio bio(*out.rdbuf());
      if(X509_print(bio.get_wrapped(),const_cast<X509*>(get_wrapped()))!=1) {
	throw swool::exception();
      }
      return *this;
    }  
    
    void X509Cert::free( X509 * t ) {
      X509_free(t);
    }
    
    X509 * X509Cert::generate() {
      X509 * res = X509_new();
      return res;
    }

    X509Cert::~X509Cert() {
      delete m_subject;
      delete m_publicKey;
      delete m_issuer;
      clear();
    }

    void X509Cert::setStartTime( time_t t ) {
      if(t<0)
	::time(&t);
      if(ASN1_TIME_set(X509_get_notBefore(get_wrapped()), t)==0)
	throw swool::exception();
    }
    
    void X509Cert::setEndTime( time_t t ) {
      if(t<0)
	::time(&t);
      if(ASN1_TIME_set(X509_get_notAfter(get_wrapped()), t)==0)
	throw swool::exception();
    }
    

    void X509Cert::setSubject(  Name & name ) {
      X509_NAME * n = name.openssl_name();
      if(!X509_set_subject_name(get_wrapped(), n))
	throw swool::exception();
      update();
    }

    void X509Cert::setIssuer( Name & name ) {
      X509_NAME * n = (name.openssl_name());
      if(!X509_set_issuer_name(get_wrapped(), n))
	throw swool::exception();
      update();
    }

    void X509Cert::sign (Key & pkey) {
      const EVP_MD * digest = NULL;
      switch (pkey.getType()) {
      case Key::dsa:
	digest = EVP_dss1();
      case Key::rsa:
	digest = EVP_sha1();
      default:
	assert(false);
	break;
      };
      if (!(X509_sign (get_wrapped(), pkey.get_wrapped(), digest))) {
	throw swool::exception();
      } 
    }
    void X509Cert::setPublicKey (Key & pkey) {
      if(X509_set_pubkey(get_wrapped(), pkey.get_wrapped())!=1)
	throw swool::exception();;
      update();
    }

    const Key & X509Cert::getPublicKey ( ) const {
      return *m_publicKey;
    }

    void X509Cert::update() {
      delete m_subject;
      m_subject = NULL;
      m_subject = new Name(X509_get_subject_name(get_wrapped()), false);

      delete m_issuer;
      m_issuer = NULL;
      m_issuer = new Name(X509_get_issuer_name(get_wrapped()), false);

      delete m_publicKey;
      m_publicKey = NULL;
      EVP_PKEY * pk;
      if((pk = X509_get_pubkey(get_wrapped()))!=NULL)
	m_publicKey = new Key(pk, true);
    }

    const Name & X509Cert::getIssuer() const {
      return *m_issuer;
    }

    const Name & X509Cert::getSubject() const {
      return *m_subject;
    }
    
  };
};

