/**************************************************************************
*   Copyright (C) 2004 by Vukasin Toroman                                 *
*   vukasin@toroman.org                                                   *
**************************************************************************/
/************************************************************************** 
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifndef SWOOLCIPHER_H
#define SWOOLCIPHER_H

#include "../swool/bio.h"
namespace wool
{
  namespace swool {
      /**
      *  This is a wrapper class for BIO's cipher methods. The cipher routines are a high level
      *  interface to certain symmetric ciphers.
      *  @author Vukasin Toroman <vtoroman@astaro.de> (C) Astaro AG 2005
      */
      class cipher : public swool::bio
      {
      public:
          /**
           *  cipher method definition
           *  @see \ref cipher
           */
          typedef const EVP_CIPHER * (*cipher_type) (void);
          /**
           *  key type definition
           */
          typedef std::string key_type;
          /**
           *  iv_type definition
           */
          typedef std::string iv_type;
          /**
           *  int type definition
           */
          typedef int int_type;
          /**
           *  direction type definition
           */
          typedef int direction_type;
          /**
           *  direction type encryption definition
           */
          static const direction_type encryption;
          /**
           *  direction type decryption definition
           */
          static const direction_type decryption;
          /**
           *  @param ct   cipher method
           *  @param key  key
           *  @param iv   iv
           *  @param enc  direction encryption | decryption
           */
          cipher( cipher_type ct, key_type key, iv_type iv, direction_type enc );
          virtual ~cipher();

      };
  };
};

#endif

/**
  @page cipher Cipher Listing
  @see http://www.openssl.org/docs/crypto/EVP_EncryptInit.html <p>
  All algorithms have a fixed key length unless otherwise stated.

  <b>EVP_enc_null()</b><br>
    Null cipher: does nothing.<p>

  <b>EVP_des_cbc(void), EVP_des_ecb(void), EVP_des_cfb(void), EVP_des_ofb(void)</b><br>
    DES in CBC, ECB, CFB and OFB modes respectively.<p>

  <b>EVP_des_ede_cbc(void), EVP_des_ede(), EVP_des_ede_ofb(void), EVP_des_ede_cfb(void)</b><br>
    Two key triple DES in CBC, ECB, CFB and OFB modes respectively.

  <p><b> EVP_des_ede3_cbc(void), EVP_des_ede3(), EVP_des_ede3_ofb(void), EVP_des_ede3_cfb(void)</b><br>
    Three key triple DES in CBC, ECB, CFB and OFB modes respectively.

  <p><b> EVP_desx_cbc(void)</b><br>
    DESX algorithm in CBC mode.

  <p><b> EVP_rc4(void)</b><br>
    RC4 stream cipher. This is a variable key length cipher with default key length 128 bits.

  <p><b> EVP_rc4_40(void)</b><br>
    RC4 stream cipher with 40 bit key length. This is obsolete and new code should use EVP_rc4() and the
    EVP_CIPHER_CTX_set_key_length() function.

  <p><b> EVP_idea_cbc() EVP_idea_ecb(void), EVP_idea_cfb(void), EVP_idea_ofb(void), EVP_idea_cbc(void)</b><br>
    IDEA encryption algorithm in CBC, ECB, CFB and OFB modes respectively.

  <p><b> EVP_rc2_cbc(void), EVP_rc2_ecb(void), EVP_rc2_cfb(void), EVP_rc2_ofb(void)</b><br>
    RC2 encryption algorithm in CBC, ECB, CFB and OFB modes respectively. This is a variable key length cipher
    with an additional parameter called ``effective key bits'' or ``effective key length''. By default both are
    set to 128 bits.

  <p><b> EVP_rc2_40_cbc(void), EVP_rc2_64_cbc(void)</b><br>
    RC2 algorithm in CBC mode with a default key length and effective key length of 40 and 64 bits.
    These are obsolete and new code should use EVP_rc2_cbc(), EVP_CIPHER_CTX_set_key_length() and
    EVP_CIPHER_CTX_ctrl() to set the key length and effective key length.

  <p><b> EVP_bf_cbc(void), EVP_bf_ecb(void), EVP_bf_cfb(void), EVP_bf_ofb(void)</b><br>
    Blowfish encryption algorithm in CBC, ECB, CFB and OFB modes respectively. This is a variable key length cipher.

  <p><b> EVP_cast5_cbc(void), EVP_cast5_ecb(void), EVP_cast5_cfb(void), EVP_cast5_ofb(void)</b><br>
    CAST encryption algorithm in CBC, ECB, CFB and OFB modes respectively. This is a variable key length cipher.

  <p><b> EVP_rc5_32_12_16_cbc(void), EVP_rc5_32_12_16_ecb(void), EVP_rc5_32_12_16_cfb(void),
  EVP_rc5_32_12_16_ofb(void)</b><br>
  RC5 encryption algorithm in CBC, ECB, CFB and OFB modes respectively. This is a variable key length cipher
  with an additional ``number of rounds'' parameter. By default the key length is set to 128 bits and 12 rounds.


*/
