#include "key.hh"

#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/rsa.h>
#include <openssl/dsa.h>
#include <openssl/dh.h>
#include <openssl/evp.h>

#include "../swool/streambuf_bio.h"

namespace wool {
namespace swool {

  
  Dh::Dh( size_t length, size_t generator  ) :Wrapper<DH>(DH_generate_parameters(length, generator, NULL, NULL),true ) {
    init();
  }
  
  Dh::Dh( DH * dh , bool owns )  :Wrapper<DH>(dh,owns) {    
    init();
  }
  
  Dh::status_code_mask Dh::check() const {
    status_code_mask res;
    return (DH_check(get_wrapped(), &res) == 1)?res:NotAbleToCheck;
  }

  DH * Dh::generate() {
    DH * n;
    if ( (n = DH_new()) == NULL )
      throw swool::exception();
    return n;
  }

  void Dh::update() {  }
  void Dh::free( DH * d) {    DH_free(d);  }
  Dh::~Dh( ) {     
    clear(); 
  }

  Dsa::Dsa( DSA * dsa , bool owns ) : Wrapper<DSA>(dsa, owns) {  
    init();
  }

  DSA * Dsa::generate() {
    DSA * n;
    if( (n = DSA_new()) == NULL )
      throw swool::exception();
    return n;
  }

  void Dsa::free( DSA * d) {    DSA_free(d);  }
  void Dsa::update( ) { }
  Dsa::~Dsa() {  
    clear();  
  }

  Dsa::Dsa( size_t length, const std::string & seed ) 
    : Wrapper<DSA>(DSA_generate_parameters(length, NULL, 0, NULL, NULL, NULL, NULL),true) {
    init();
  }

  Dsa & Dsa::generate_key() {
    if(DSA_generate_key(get_wrapped())!=1)
      throw wool::exception("Could not generate DSA key");
    return *this;
  }

  Rsa::Rsa( size_t length, size_t exponent ) 
    : Wrapper<RSA>(RSA_generate_key( length, exponent, NULL, NULL ),true){
    init();
  }

  Rsa::Rsa( RSA * rsa , bool owns )
    : Wrapper<RSA>(rsa,owns) {
    init();
  }
  
  RSA * Rsa::generate() {
    RSA * n;
    if( (n = RSA_new()) == NULL )
      throw swool::exception();
    return n;
  }

  void Rsa::free( RSA * d) {    RSA_free(d);  }
  void Rsa::update( ) { }
  Rsa::~Rsa() { 
    clear();  
  }
  
 
  Key::Key ( EVP_PKEY * k, bool owns ) : Wrapper<EVP_PKEY>(k,owns) {
    m_dsa = NULL;
    m_rsa = NULL;
    m_dh = NULL;
    init();
  }
  
  EVP_PKEY * Key::generate() {
    EVP_PKEY * i;
    if((i=EVP_PKEY_new())==NULL)
      throw swool::exception();
    return i;
  }
  
  void Key::free( EVP_PKEY * k ) { EVP_PKEY_free(k);}
  Key::~Key ( ) {
    delete m_dsa;
    delete m_rsa;
    delete m_dh;
    clear();
  }
    

  Key::key_type Key::getType() const {
    return (Key::key_type)EVP_PKEY_type(get_wrapped()->type);
  }
  

  Rsa & Key::getRsa() {
    return *m_rsa;
  }

  Dsa & Key::getDsa() {
    return *m_dsa;
  }

  Dh & Key::getDh() {
    return *m_dh;
  }

  Key & Key::setRsa ( const Rsa & k ) {
    if(EVP_PKEY_set1_RSA(get_wrapped(), const_cast<RSA*>(k.get_wrapped()))!=1)
      throw swool::exception();
    assert(getType() == rsa );
    update();
    return *this;
  }

  Key & Key::setDsa ( const Dsa & k ) {
    if(EVP_PKEY_set1_DSA(get_wrapped(), const_cast<DSA*>(k.get_wrapped()))!=1)
      throw swool::exception();
    assert(getType() == dsa );
    update();
    return *this;
  }

  Key & Key::setDh ( const Dh & k ) {
    if(EVP_PKEY_set1_DH(get_wrapped(), const_cast<DH*>(k.get_wrapped()))!=1)
      throw swool::exception();
    assert(getType() == dh );
    update();
    return *this;
  }
  
  const Rsa & Key::getRsa() const {
    return *m_rsa;
  }

  const Dsa & Key::getDsa() const {
    return *m_dsa;
  }

  const Dh & Key::getDh() const {
    return *m_dh;
  }
  
  void Key::update() {
    delete m_dsa;
    delete m_rsa;
    delete m_dh;

    m_dsa = NULL;
    m_rsa = NULL;
    m_dh = NULL;
    
    switch ( getType() ) {
    case EVP_PKEY_DSA :
      m_dsa = new Dsa(EVP_PKEY_get1_DSA(get_wrapped()), true);
      break;
    case EVP_PKEY_RSA :
      m_rsa = new Rsa(EVP_PKEY_get1_RSA(get_wrapped()), true);
      break;
    case EVP_PKEY_DH :
      m_dh = new Dh(EVP_PKEY_get1_DH(get_wrapped()), true);
      break;
    default:
      assert(false); // unsupported key type
      break;
    }
  }

};
};
