#ifndef SWOOL_KEY_HH
#define SWOOL_KEY_HH

#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/rsa.h>
#include <openssl/dsa.h>
#include <openssl/dh.h>
#include <openssl/evp.h>
//#include <openssl/ec.h>
#include <openssl/err.h>

#include <string>
#include <iostream>

#include "../base/exception.h"
#include "../base/wrapper.hh"

namespace wool {
namespace swool {

  class exception : public wool::exception { 
  public:
    long error_id;
    char buffer[256];
    inline exception( ) throw () : error_id(ERR_get_error()) {
      buffer[0] = 0;
      ERR_error_string_n(error_id,buffer,256);
    } 
    inline virtual const char * what () const throw () {
      return buffer;
    }
  };

class Dh : public wool::Wrapper<DH>{
public:
  typedef int status_code;
  typedef int status_code_mask;

  static const size_t DefaultLength = 1024;
  static const size_t DefaultGenerator = 5;
  static const status_code NotSafePrime = DH_CHECK_P_NOT_SAFE_PRIME;
  static const status_code NotSuitableGenerator = DH_NOT_SUITABLE_GENERATOR;
  static const status_code NotAbleToCheck = DH_UNABLE_TO_CHECK_GENERATOR;;

  Dh( size_t length, size_t generator  );
  Dh( DH * dh = NULL , bool owns = false );
  ~Dh( );

  status_code_mask check() const;

protected:
  void update();
  DH * generate();
  void free(DH *);
};


class Dsa : public wool::Wrapper<DSA>{
public:
  Dsa( DSA * dsa = NULL , bool owns = false );
  Dsa( size_t length, const std::string & seed );
  ~Dsa();
  Dsa & generate_key();

protected:
  void update();
  DSA * generate();
  void free(DSA *);
};


class Rsa : public wool::Wrapper<RSA>{
public:
  Rsa( size_t length , size_t exponent );
  Rsa( RSA * rsa = NULL , bool owns = false );
  ~Rsa();  
protected:
  void update();
  RSA * generate();
  void free(RSA *);
};

class Key : public wool::Wrapper<EVP_PKEY>{
public:
  typedef enum {
    dsa = EVP_PKEY_DSA,
    rsa = EVP_PKEY_RSA,
    dh = EVP_PKEY_DH,
    undef = NID_undef 
  } key_type;
  
  Key ( EVP_PKEY * k = NULL, bool owns = false );
  ~Key ( );
  
  key_type getType() const;

  Rsa & getRsa();
  Dsa & getDsa();
  Dh & getDh();

  Key & setRsa ( const Rsa & k );
  Key & setDsa ( const Dsa & k );
  Key & setDh ( const Dh & k );
  
  const Rsa & getRsa() const;
  const Dsa & getDsa() const;
  const Dh & getDh() const;

protected:
  void update();
  EVP_PKEY * generate();
  void free(EVP_PKEY *);
  
private:
  Rsa * m_rsa;
  Dsa * m_dsa;
  Dh * m_dh; 

};

};
};
#endif
