#include "name.hh"
#include "../base/exception.h"

namespace wool {
  namespace swool {
    size_t Name::size() const {
      int i = X509_NAME_entry_count(m_name);
      return (i>0)?i:0;
    }

    Name::Name(X509_NAME * name, bool owns) : m_name(NULL), m_owns(false) {
      m_name = name;
      if(m_name == NULL) {
	m_owns = false;
	m_name = X509_NAME_new();
	if(m_name==NULL)
	  //throw std::runtime_error("cannot create X509 name"); 
	  throw wool::exception("cannot create X509 name"); 
	m_owns = true;
      } else {
	m_owns = owns;
      }
    }
    
    Name::Name ( const Name & n ) : m_name(NULL), m_owns(false) {
      m_name = X509_NAME_dup(n.m_name);
      m_owns = true;
    }

    Name::~Name() {
      if(m_owns)
	X509_NAME_free(m_name);
    }

    Name::iterator Name::find ( int id, Name::iterator start ) {
      return iterator(std::max<int>(0,X509_NAME_get_index_by_NID(m_name, id, start.id - 1)),this);      
    }
    
    Name::const_iterator Name::find ( int id, Name::const_iterator start ) const {
      return const_iterator(std::max<int>(0,X509_NAME_get_index_by_NID(m_name, id, start.id - 1)),this);
    }
    
    Name::iterator Name::find ( const std::string & strid, Name::iterator start ) {
      int id = OBJ_txt2nid(strid.c_str());
      return (id!=NID_undef)?iterator(std::max<int>(0,X509_NAME_get_index_by_NID(m_name, id, start.id - 1)),this):end();
    }

    Name::const_iterator Name::find ( const std::string & strid, Name::const_iterator start ) const {
      int id = OBJ_txt2nid(strid.c_str());
      return (id!=NID_undef)?const_iterator(std::max<int>(0,X509_NAME_get_index_by_NID(m_name, id, start.id - 1)),this):end();
    }

    Name::iterator Name::insert ( const std::string & name, const std::string & value ) {
      return insert(name, value, end());
    }
    
    Name::iterator Name::insert ( const std::string & name, const std::string & value, iterator where  ) {
      X509_NAME_ENTRY *ent;
      int nid;
      if(where==end())
	where.id=size();
	
      if ((nid = OBJ_txt2nid (name.c_str())) == NID_undef)
	throw wool::exception("cannot find NID for "+ name); //!TODO
      
      if (!(ent = X509_NAME_ENTRY_create_by_NID (NULL, nid, MBSTRING_ASC, (unsigned char *)strdup(value.c_str()), -1)))
	throw wool::exception("cannot create entry for NID "+ name); //!TODO;
      int i = X509_NAME_add_entry(m_name, ent, where.id, 0);
      OPENSSL_free(ent);
      if(!i)
	throw wool::exception("could not insert "+ name); //!TODO
      iterator res(where.id, this);
      return res;
    }

    void Name::erase ( Name::iterator what ) {
      X509_NAME_delete_entry(m_name,what.id);
    }

    void Name::erase ( Name::iterator from, Name::iterator to ) {
      int min = std::min<int>(from.id, to.id);
      int max = std::max<int>(from.id, to.id);
      for(int i = max-1; i>=min; i--) {
	X509_NAME_delete_entry(m_name, i);
      }
    }

    std::string Name::Entry::getData() const {
      unsigned char * data;
      ASN1_STRING * asn1_data = X509_NAME_ENTRY_get_data(const_cast<Entry*>(this));      
      int len = ASN1_STRING_to_UTF8(&data, asn1_data);
      std::string res;
      if(len>0) {
	res.append((char *)(data), len);
	OPENSSL_free(data);
      }
      return res;
    }

    std::string Name::Entry::getName() const {
      char data [1024];
      return std::string(data,OBJ_obj2txt(data, 1024,X509_NAME_ENTRY_get_object(const_cast<Entry*>(this)),0));     
    }

    void Name::Entry::setName( int id ) {
      X509_NAME_ENTRY_set_object(this, OBJ_nid2obj(id));
    }
    
    void Name::Entry::setName ( const std::string & n ) {
      setName(OBJ_txt2nid(n.c_str()));
    }
  
    void Name::Entry::setData( const std::string & d) {
      X509_NAME_ENTRY_set_data(this, MBSTRING_ASC, (const unsigned char*)d.c_str(), d.length());
    }

    Name::Entry & Name::operator [] ( size_t id ) {
      return static_cast<Entry &>(*X509_NAME_get_entry(m_name, id));
    }

    const Name::Entry & Name::operator [] ( size_t id ) const {
      return static_cast<const Entry &>(*X509_NAME_get_entry(m_name, id));
    }
    
    X509_NAME * Name::openssl_name() {
      return m_name;
    }
    const X509_NAME * Name::openssl_name() const {
      return m_name;
    }

  };
};
