/**************************************************************************
*   Copyright (C) 2004 by Vukasin Toroman                                 *
*   vukasin@toroman.org                                                   *
**************************************************************************/
/************************************************************************** 
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#ifndef SWOOLSSL_FILTER_H
#define SWOOLSSL_FILTER_H

#include "../base/debug.h"
#include "../swool/bio.h"
#include "../swool/cert.hh"
#include "../swool/key.hh"
#include "../base/wrapper.hh"

#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/x509v3.h>



namespace wool {
  namespace swool {
    class SslContext :public wool::Wrapper<SSL_CTX>{
    public:
      typedef std::string cert_path_type;

      typedef SSL_METHOD * ssl_method;
      typedef SSL_METHOD * (*ssl_method_factory) (void);

      static ssl_method_factory const tls1;
      static ssl_method_factory const ssl2;
      static ssl_method_factory const ssl23;
      static ssl_method_factory const ssl3;
      
      static ssl_method_factory const tls1client;
      static ssl_method_factory const ssl2client;
      static ssl_method_factory const ssl23client;
      static ssl_method_factory const ssl3client;
      
      static ssl_method_factory const tls1server;
      static ssl_method_factory const ssl2server;
      static ssl_method_factory const  ssl23server;
      static ssl_method_factory const  ssl3server;
    

      static void Init();      
      
      SslContext ( SSL_CTX * ctx = NULL, bool owns = true);
      SslContext ( ssl_method_factory method);
      virtual ~SslContext ();
      
      void setCert ( X509Cert & cert );
      void setPrivateKey ( Key & key );

      void setCALocation (cert_path_type path);

      void setTmpRsa ( Rsa & rsa );
      void setTmpDh (  Dh & dh );

    protected:
      virtual SSL_CTX * generate();
      virtual void free( SSL_CTX * );
      virtual void update();
    };


    class Ssl : public wool::Wrapper<SSL> {
    public:   
      typedef int verify_mode;
      typedef long verification_result;
      
      static const verify_mode verify_none = SSL_VERIFY_NONE;
      static const verify_mode verify_peer = SSL_VERIFY_PEER;
      static const verify_mode verify_require_peer_cert = SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
      static const verify_mode verify_client_once =  SSL_VERIFY_CLIENT_ONCE;
      
      Ssl ( SSL * ssl = NULL, bool owns = true);
      Ssl ( SslContext & ctx);
      ~Ssl();
      
      const X509Cert & getPeerCert() const;
      verification_result getVerificationResult () const;

      void setVerificationMode ( verify_mode m );
      void setMethod ( SslContext::ssl_method m );

      const SslContext::ssl_method getMethod ( ) const;
      
      void renegotiate ();
      void doHandshake ();
      
    protected:
      virtual void update();
      virtual SSL * generate( );
      virtual void free( SSL * );
      
    private:
      X509Cert * m_peer_cert;
    };
    
    class SslBio : public bio {
    public:
      SslBio( BIO * b = NULL, bool owns = true );
      SslBio( SslContext & s, bool client_mode = true );

      virtual ~SslBio();

      Ssl & getSsl();
      const Ssl & getSsl() const;
      void setSsl( Ssl & s, int close_flag);
      
    protected:
      virtual void update();
      virtual void free( BIO * b );
    private:
      Ssl * m_ssl;
    };
  };
};

/**
   \page ssl_verify  SSL verification results and error codes
   @see http://www.openssl.org/docs/apps/verify.html#<p>
   
   An exhaustive list of the error codes and messages is shown below. Some of the error codes are defined
   but never returned: these are described as ``unused''.<p>
   
   <b>0 X509_V_OK: ok</b><br>
   the operation was successful.<p>
   
   <b>2 X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT: unable to get issuer certificate</b><br>
   the issuer certificate could not be found: this occurs if the issuer certificate of an untrusted
   certificate cannot be found.<p>
   
   <b>3 X509_V_ERR_UNABLE_TO_GET_CRL unable to get certificate CRL</b><br>
   the CRL of a certificate could not be found. Unused.<p>

   <b>4 X509_V_ERR_UNABLE_TO_DECRYPT_CERT_SIGNATURE: unable to decrypt certificate's signature</b><br>
   the certificate signature could not be decrypted. This means that the actual signature value could not be
   determined rather than it not matching the expected value, this is only meaningful for RSA keys.<p>
   
   <b>5 X509_V_ERR_UNABLE_TO_DECRYPT_CRL_SIGNATURE: unable to decrypt CRL's signature</b><br>
   the CRL signature could not be decrypted: this means that the actual signature value could not be
   determined rather than it not matching the expected value. Unused.<p>
   
   <b>6 X509_V_ERR_UNABLE_TO_DECODE_ISSUER_PUBLIC_KEY: unable to decode issuer public key</b><br>
   the public key in the certificate SubjectPublicKeyInfo could not be read.<p>
   
   <b>7 X509_V_ERR_CERT_SIGNATURE_FAILURE: certificate signature failure</b><br>
   the signature of the certificate is invalid.<p>
   
   <b>8 X509_V_ERR_CRL_SIGNATURE_FAILURE: CRL signature failure</b><br>
   the signature of the certificate is invalid. Unused.<p>
   
   <b>9 X509_V_ERR_CERT_NOT_YET_VALID: certificate is not yet valid</b><br>
   the certificate is not yet valid: the notBefore date is after the current time.<p>
   
   <b>10 X509_V_ERR_CERT_HAS_EXPIRED: certificate has expired</b><br>
   the certificate has expired: that is the notAfter date is before the current time.<p>

   <b>11 X509_V_ERR_CRL_NOT_YET_VALID: CRL is not yet valid</b><br>
   the CRL is not yet valid. Unused.<p>
   
   <b>12 X509_V_ERR_CRL_HAS_EXPIRED: CRL has expired</b><br>
   the CRL has expired. Unused.<p>
   
   <b>13 X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD: format error in certificate's notBefore field</b><br>
   the certificate notBefore field contains an invalid time.<p>
   
   <b>14 X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD: format error in certificate's notAfter field</b><br>
   the certificate notAfter field contains an invalid time.<p>
   
   <b>15 X509_V_ERR_ERROR_IN_CRL_LAST_UPDATE_FIELD: format error in CRL's lastUpdate field</b><br>
   the CRL lastUpdate field contains an invalid time. Unused.<p>
   
   <b>16 X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD: format error in CRL's nextUpdate field</b><br>
   the CRL nextUpdate field contains an invalid time. Unused.<p>

   <b>17 X509_V_ERR_OUT_OF_MEM: out of memory</b><br>
   an error occurred trying to allocate memory. This should never happen.<p>

   <b>18 X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT: self signed certificate</b><br>
   the passed certificate is self signed and the same certificate cannot be found in the list of trusted
   certificates.<p>
   
   <b>19 X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN: self signed certificate in certificate chain</b><br>
   the certificate chain could be built up using the untrusted certificates but the root could not be found locally.<p>
   
   <b>20 X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY: unable to get local issuer certificate</b><br>
   the issuer certificate of a locally looked up certificate could not be found. This normally means the list of
   trusted certificates is not complete.<p>
   
   <b>21 X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE: unable to verify the first certificate</b><br>
   no signatures could be verified because the chain contains only one certificate and it is not self signed.<p>
   
   <b>22 X509_V_ERR_CERT_CHAIN_TOO_LONG: certificate chain too long</b><br>
   the certificate chain length is greater than the supplied maximum depth. Unused.<p>
   
   <b>23 X509_V_ERR_CERT_REVOKED: certificate revoked</b><br>
   the certificate has been revoked. Unused.<p>
   
   <b>24 X509_V_ERR_INVALID_CA: invalid CA certificate</b><br>
   a CA certificate is invalid. Either it is not a CA or its extensions are not consistent with the supplied purpose.<p>
   
   <b>25 X509_V_ERR_PATH_LENGTH_EXCEEDED: path length constraint exceeded</b><br>
   the basicConstraints pathlength parameter has been exceeded.<p>
   
   <b>26 X509_V_ERR_INVALID_PURPOSE: unsupported certificate purpose</b><br>
   the supplied certificate cannot be used for the specified purpose.<p>
   
   <b>27 X509_V_ERR_CERT_UNTRUSTED: certificate not trusted</b><br>
   the root CA is not marked as trusted for the specified purpose.<p>
   
   <b>28 X509_V_ERR_CERT_REJECTED: certificate rejected</b><br>
   the root CA is marked to reject the specified purpose.<p>
   
   <b>29 X509_V_ERR_SUBJECT_ISSUER_MISMATCH: subject issuer mismatch</b><br>
   the current candidate issuer certificate was rejected because its subject name did not match the issuer name of
   the current certificate. Only displayed when the -issuer_checks option is set.<p>
   
   <b>30 X509_V_ERR_AKID_SKID_MISMATCH: authority and subject key identifier mismatch</b><br>
   the current candidate issuer certificate was rejected because its subject key identifier was present and did
   not match the authority key identifier current certificate. Only displayed when the -issuer_checks option is set.<p>
   
   <b>31 X509_V_ERR_AKID_ISSUER_SERIAL_MISMATCH: authority and issuer serial number mismatch</b><br>
   the current candidate issuer certificate was rejected because its issuer name and serial number was present and
   did not match the authority key identifier of the current certificate. Only displayed when the -issuer_checks
   option is set.<p>
   
   <b>32 X509_V_ERR_KEYUSAGE_NO_CERTSIGN:key usage does not include certificate signing</b><br>
   the current candidate issuer certificate was rejected because its keyUsage extension does not permit
   certificate signing.<p>
   
   <b>50 X509_V_ERR_APPLICATION_VERIFICATION: application verification failure</b><br>
   an application specific error. Unused.
   
*/

#endif
