#include "../swool/store.hh"
#include "../base/wrapper.hh"

#include <openssl/x509.h>
#include <openssl/x509_vfy.h>

#include <boost/thread.hpp>

namespace wool {
  namespace swool {  
    

    static boost::thread_specific_ptr<X509Store> current_store;


    X509Store::X509Store( X509_STORE * r, bool owns ) 
      : Wrapper<X509_STORE>(r,owns){
    }

    X509Store::~X509Store() {
      clear();
    }
      
    
    void X509Store::loadLocations( const std::string & ca_file, const std::string & ca_folder) {
      if (X509_STORE_load_locations (get_wrapped(), ca_file.c_str(), ca_folder.c_str()) != 1)
	throw swool::exception();
      if (X509_STORE_set_default_paths (get_wrapped()) != 1)
	throw swool::exception();
    }

    void X509Store::addCrlLookup( const std::string & crl_file, int format ) {
      X509_LOOKUP *lookup;
      if (!(lookup = X509_STORE_add_lookup (get_wrapped(), X509_LOOKUP_file ())))
	throw swool::exception();
      if (X509_load_crl_file (lookup, crl_file.c_str(), format) != 1) 
	throw swool::exception();
    }
    
    void X509Store::setFlags( unsigned long f ) {
      X509_STORE_set_flags (get_wrapped(), f);
    }
    
    /*
      unsigned long X509Store::getFlags( ) const {
      return get_wrapped()->flags;
    }
    */
    
    
    int  X509Store::verify_callback (int ok, X509_STORE_CTX * stor) {     
      return (ok)|| ((current_store.get()) && (current_store->getIgnoredErrors().count(stor->error)>0));
    }

    void X509Store::verify ( X509Cert & cert ) {    
      X509_STORE_CTX *verify_ctx = X509_STORE_CTX_new ();
      if(!verify_ctx) {
	X509_STORE_CTX_free(verify_ctx);
	throw swool::exception();
      }
      if (X509_STORE_CTX_init (verify_ctx, get_wrapped(), cert.get_wrapped(), NULL) != 1) {
	X509_STORE_CTX_free(verify_ctx);
	throw swool::exception();
      }
      current_store.reset(this);
      int res = X509_verify_cert (verify_ctx);
      current_store.reset(NULL);
      X509_STORE_CTX_free(verify_ctx);

      if (res!= 1) {
	throw swool::exception();
      }
    }

    void X509Store::setIgnoredErrors( const X509Store::error_code_set & e ) {
      m_ignored_errors = e;
    }
    
    const X509Store::error_code_set & X509Store::getIgnoredErrors() const {
      return m_ignored_errors;
    }
     
    void X509Store::update() {
    }

    void X509Store::free( X509_STORE * c) {
      X509_STORE_free(c);
    }
    
    X509_STORE * X509Store::generate() {
      X509_STORE * res = X509_STORE_new();
      X509_STORE_set_verify_cb_func (res, &X509Store::verify_callback);
      return res;
    }
  };
};


