/**************************************************************************
*  Copyright (C) 2006 by Jens Finkhaeuser <unwesen@users.sourceforge.net> *
**************************************************************************/
/************************************************************************** 
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

// boost includes
#include <boost/lexical_cast.hpp>

// test incldues
#include <test/test_uri.h>

namespace wool {
namespace test {

// ===========================================================================
// == Testcase implementations

void
test_uri::testCheckURI()
{
    std::map<std::string, bool> test_urls;
    // examples from rfc 3986
    test_urls["ftp://ftp.is.co.za/rfc/rfc1808.txt"] = true;
    test_urls["http://www.ietf.org/rfc/rfc2396.txt"] = true;
    test_urls["ldap://[2001:db8::7]/c=GB?objectClass?one"] = false; // not supported because if the extended host syntax
    test_urls["mailto:John.Doe@example.com"] = true;
    test_urls["news:comp.infosystems.www.servers.unix"] = true;
    test_urls["tel:+1-816-555-1212"] = true;
    test_urls["telnet://192.0.2.16:80/"] = true;
    test_urls["urn:oasis:names:specification:docbook:dtd:xml:4.1.2"] = true;

    // further examples
    test_urls["http://foo@localhost"] = true;
    test_urls["http://foo@localhost/"] = true;
    test_urls["http://foo@localhost/bar/baz"] = true;
    test_urls["http://foo:secret@localhost"] = true;
    test_urls["http://foo:secret@localhost/"] = true;
    test_urls["http://foo:secret@localhost/bar/baz"] = true;
    test_urls["http://localhost:123"] = true;
    test_urls["http://localhost:123/"] = true;
    test_urls["http://localhost:123/bar/baz"] = true;
    test_urls["http://localhost:"] = true;
    test_urls["http://localhost:/"] = true;
    test_urls["http://localhost:/bar/baz"] = true;
    test_urls["http://foo@localhost/bar/baz?test-query=foobar#test-fragment"] = true;
    test_urls["http://foo@localhost/bar/baz?#test-fragment"] = true;
    test_urls["http://foo@localhost/bar/baz?test-query=foobar"] = true;
    test_urls["http://foo@localhost/bar/baz?test-query=foobar#"] = true;
    test_urls["http://foo@localhost/bar/baz?#"] = true;

    std::map<std::string, bool>::const_iterator url_iter = test_urls.begin();
    for ( ; url_iter != test_urls.end() ; ++url_iter) {
        CPPUNIT_ASSERT_EQUAL(url_iter->second, wool::check_uri(url_iter->first, false));
    }
}


void
test_uri::testGenericURI()
{
    try {
        wool::URI uri("");
        CPPUNIT_FAIL("expected wool::parse_error");
    } catch (wool::parse_error const & ex) {
        CPPUNIT_ASSERT_EQUAL(wool::parse_error::uri_empty, ex.code());
    }

    try {
        wool::URI uri("no-scheme");
        CPPUNIT_FAIL("expected wool::parse_error");
    } catch (wool::parse_error const & ex) {
        CPPUNIT_ASSERT_EQUAL(wool::parse_error::uri_no_scheme, ex.code());
    }

    try {
        wool::URI uri("scheme-at-end://");
        CPPUNIT_FAIL("expected wool::parse_error");
    } catch (wool::parse_error const & ex) {
        CPPUNIT_ASSERT_EQUAL(wool::parse_error::uri_empty, ex.code());
    }

    try {
        // the scheme may be invalid, but the URI conforms to the common
        // URI syntax, so at least we can parse it.
        wool::URI uri("invalid-scheme://foobar");
    } catch (wool::parse_error const & ex) {
        CPPUNIT_FAIL("did not expected wool::parse_error");
    }


    std::map<std::string, std::map<std::string, std::string> > test_urls;
    std::map<std::string, std::string> tmp_info;

    // examples from rfc 3986
    tmp_info.clear();
    tmp_info["scheme"]      = "ftp";
    tmp_info["host"]        = "ftp.is.co.za";
    tmp_info["port"]        = "21";
    tmp_info["authority"]   = "ftp.is.co.za";
    tmp_info["path"]        = "/rfc/rfc1808.txt";
    test_urls["ftp://ftp.is.co.za/rfc/rfc1808.txt"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "http";
    tmp_info["host"]        = "www.ietf.org";
    tmp_info["port"]        = "80";
    tmp_info["authority"]   = "www.ietf.org";
    tmp_info["path"]        = "/rfc/rfc2396.txt";
    test_urls["http://www.ietf.org/rfc/rfc2396.txt"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "mailto";
    tmp_info["path"]        = "John.Doe@example.com";
    test_urls["mailto:John.Doe@example.com"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "news";
    tmp_info["path"]        = "comp.infosystems.www.servers.unix";
    test_urls["news:comp.infosystems.www.servers.unix"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "tel";
    tmp_info["path"]        = "+1-816-555-1212";
    test_urls["tel:+1-816-555-1212"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "telnet";
    tmp_info["host"]        = "192.0.2.16";
    tmp_info["port"]        = "80";
    tmp_info["authority"]   = "192.0.2.16:80";
    tmp_info["path"]        = "/";
    test_urls["telnet://192.0.2.16:80/"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "urn";
    tmp_info["path"]        = "oasis:names:specification:docbook:dtd:xml:4.1.2";
    test_urls["urn:oasis:names:specification:docbook:dtd:xml:4.1.2"] = tmp_info;

    // further examples
    tmp_info.clear();
    tmp_info["scheme"]      = "http";
    tmp_info["userinfo"]    = "foo";
    tmp_info["host"]        = "localhost";
    tmp_info["port"]        = "80";
    tmp_info["authority"]   = "foo@localhost";
    test_urls["http://foo@localhost"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "http";
    tmp_info["userinfo"]    = "foo";
    tmp_info["host"]        = "localhost";
    tmp_info["port"]        = "80";
    tmp_info["authority"]   = "foo@localhost";
    tmp_info["path"]        = "/";
    test_urls["http://foo@localhost/"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "http";
    tmp_info["userinfo"]    = "foo";
    tmp_info["host"]        = "localhost";
    tmp_info["port"]        = "80";
    tmp_info["authority"]   = "foo@localhost";
    tmp_info["path"]        = "/bar/baz";
    test_urls["http://foo@localhost/bar/baz"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "http";
    tmp_info["userinfo"]    = "foo";
    tmp_info["host"]        = "localhost";
    tmp_info["port"]        = "80";
    tmp_info["authority"]   = "foo@localhost";
    tmp_info["path"]        = "/bar/baz";
    tmp_info["query"]       = "test-query=foobar";
    test_urls["http://foo@localhost/bar/baz?test-query=foobar"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "http";
    tmp_info["userinfo"]    = "foo";
    tmp_info["host"]        = "localhost";
    tmp_info["port"]        = "80";
    tmp_info["authority"]   = "foo@localhost";
    tmp_info["path"]        = "/bar/baz";
    tmp_info["query"]       = "test-query=foobar";
    tmp_info["fragment"]    = "test-fragment";
    test_urls["http://foo@localhost/bar/baz?test-query=foobar#test-fragment"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "http";
    tmp_info["userinfo"]    = "foo";
    tmp_info["host"]        = "localhost";
    tmp_info["port"]        = "80";
    tmp_info["authority"]   = "foo@localhost";
    tmp_info["path"]        = "/bar/baz";
    tmp_info["fragment"]    = "test-fragment";
    test_urls["http://foo@localhost/bar/baz?#test-fragment"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "http";
    tmp_info["userinfo"]    = "foo";
    tmp_info["host"]        = "localhost";
    tmp_info["port"]        = "80";
    tmp_info["authority"]   = "foo@localhost";
    tmp_info["path"]        = "/bar/baz";
    tmp_info["fragment"]    = "test-fragment";
    test_urls["http://foo@localhost/bar/baz?#test-fragment"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "http";
    tmp_info["userinfo"]    = "foo:secret";
    tmp_info["authority"]   = "foo:secret@localhost";
    tmp_info["host"]        = "localhost";
    tmp_info["port"]        = "80";
    test_urls["http://foo:secret@localhost"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "http";
    tmp_info["host"]        = "localhost";
    tmp_info["authority"]   = "localhost:123";
    tmp_info["port"]        = "123";
    test_urls["http://localhost:123"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "http";
    tmp_info["host"]        = "localhost";
    tmp_info["port"]        = "80";
    tmp_info["authority"]   = "localhost:";
    test_urls["http://localhost:"] = tmp_info;

    tmp_info.clear();
    tmp_info["scheme"]      = "file";
    tmp_info["host"]        = "";
    tmp_info["authority"]   = "";
    tmp_info["path"]        = "/etc/passwd";
    test_urls["file:///etc/passwd"] = tmp_info;

    std::map<std::string, std::map<std::string, std::string> >::iterator url_iter
        = test_urls.begin();
    for ( ; url_iter != test_urls.end() ; ++url_iter) {
        wool::URI uri(url_iter->first);

        CPPUNIT_ASSERT_EQUAL(url_iter->second["scheme"],    uri.getScheme());
        CPPUNIT_ASSERT_EQUAL(url_iter->second["userinfo"],  uri.getUserInfo());
        CPPUNIT_ASSERT_EQUAL(url_iter->second["host"],      uri.getHost());
        CPPUNIT_ASSERT_EQUAL(url_iter->second["authority"], uri.getAuthority());
        CPPUNIT_ASSERT_EQUAL(url_iter->second["path"],      uri.getPath());
        CPPUNIT_ASSERT_EQUAL(url_iter->second["query"],     uri.getQuery());
        CPPUNIT_ASSERT_EQUAL(url_iter->second["fragment"],  uri.getFragment());

        std::string tmp_port;
        if (uri.getPort()) {
            tmp_port = boost::lexical_cast<std::string>(uri.getPort());
        }
        CPPUNIT_ASSERT_EQUAL(url_iter->second["port"], tmp_port);
    }
}



void
test_uri::testHTTPURI()
{
    // specific extensions to the generic uri syntax are
    // - username/password separation
    // - path parsing
    // - query parsing

    std::string uri_string = "http://fo%20o:as%2Fdf@localhost/ba%20r/b%20az?test%20que+ry=foobar&blargh&blargh=%20asdf#test%20fragment";
    wool::URI uri(uri_string);

    // check username/password
    CPPUNIT_ASSERT_EQUAL(std::string("fo o"), uri.getUserName());
    CPPUNIT_ASSERT_EQUAL(std::string("as/df"), uri.getPassword());

    // check path components
    std::vector<std::string> path = uri.getPathParts();
    CPPUNIT_ASSERT_EQUAL(std::vector<std::string>::size_type(2), path.size());
    CPPUNIT_ASSERT_EQUAL(std::string("ba r"), path[0]);
    CPPUNIT_ASSERT_EQUAL(std::string("b az"), path[1]);

    // check fragment
    CPPUNIT_ASSERT_EQUAL(std::string("test fragment"), uri.getFragment());

    // check query components
    std::multimap<std::string, std::string> query = uri.getQueryParts();
    CPPUNIT_ASSERT_EQUAL((std::multimap<std::string, std::string>::size_type(3)), query.size());

    typedef std::multimap<std::string, std::string>::const_iterator query_iter_type;

    query_iter_type iter1 = query.find("test que ry");
    CPPUNIT_ASSERT(query.end() != iter1);
    CPPUNIT_ASSERT_EQUAL(std::string("foobar"), iter1->second);

    std::pair<query_iter_type, query_iter_type> range = query.equal_range("blargh");
    int count = 0;
    bool found_empty = false;
    bool found_asdf = false;
    for (query_iter_type iter = range.first ; iter != range.second ; ++iter) {
        std::string tmp = iter->second;
        if (tmp.empty()) {
            found_empty = true;
        } else if (tmp == " asdf") {
            found_asdf = true;
        }
        ++count;
    }
    CPPUNIT_ASSERT_EQUAL(true, found_empty);
    CPPUNIT_ASSERT_EQUAL(true, found_asdf);
    CPPUNIT_ASSERT_EQUAL(2, count);


    // ctor & comparison tests
    wool::URI uri2(uri);
    CPPUNIT_ASSERT(uri == uri2);
    CPPUNIT_ASSERT(uri == uri_string);
    CPPUNIT_ASSERT(uri2 == uri_string);

    // assignment & comparison tests
    std::string mailto_string = "mailto:jfinkhaeuser@astaro.com";
    uri = mailto_string;
    CPPUNIT_ASSERT(uri == mailto_string);
    CPPUNIT_ASSERT(uri2 != mailto_string);
    CPPUNIT_ASSERT(uri != uri2);

    uri2 = mailto_string;
    CPPUNIT_ASSERT(uri2 == mailto_string);
    CPPUNIT_ASSERT(uri == uri2);

    // comparison of uris with percent-encoded parts
    std::string alternate_mailto_string = "mailto:j%66inkhaeuser@astaro.com";
    uri = alternate_mailto_string;
    CPPUNIT_ASSERT(mailto_string != alternate_mailto_string);
    CPPUNIT_ASSERT(uri == alternate_mailto_string);

    CPPUNIT_ASSERT(uri2 == alternate_mailto_string);

    CPPUNIT_ASSERT(uri == uri2);
    CPPUNIT_ASSERT(uri.getOriginal() != uri2.getOriginal());

}



void
test_uri::testURIDefaultScheme()
{
    // test whether an invalid uri uses the default scheme or not. also test
    // whether valid uris use the default scheme.

    // 1. valid uri with default scheme - default should not be used
    {
        wool::URI uri("ftp://localhost/", "http");
        CPPUNIT_ASSERT_EQUAL(std::string("ftp"), uri.getScheme());
    }

    // 2. invalid uri without default scheme should throw
    try {
        wool::URI uri("localhost");
        CPPUNIT_FAIL("expected wool::parse_error");
    } catch (wool::parse_error const & ex) {
        CPPUNIT_ASSERT_EQUAL(wool::parse_error::uri_no_scheme, ex.code());
    }

    // 3. invalid uri with default scheme should not throw and use the
    //    default
    {
        wool::URI uri("localhost", "http");
        CPPUNIT_ASSERT_EQUAL(std::string("http"), uri.getScheme());
    }

    // 4. just hostname and port... will be recognized as scheme/hierarchy part
    {
        wool::URI uri("localhost:1234");
        CPPUNIT_ASSERT_EQUAL(std::string("localhost"), uri.getScheme());
    }

    // 5. enforce a default scheme, and tell the URI that localhost is _not_ the scheme
    {
        wool::URI uri("localhost:1234", "http", true);
        CPPUNIT_ASSERT_EQUAL(std::string("http"), uri.getScheme());
        CPPUNIT_ASSERT_EQUAL(std::string("localhost"), uri.getHost());
    }
}


}} // namespaces
